/*
    This file is part of Escape.

    Escape is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Escape is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Escape. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.escape;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.xml.bind.annotation.XmlAttribute;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;

/**
 * Transforms the XSD context model to an intermediate format. 
 * 
 * @author Lukasz Juszczyk
 *
 */
public class ContextModelTransformer {

	// types found
	private static HashMap<String,ContextType> types=
		new HashMap<String,ContextType>();
	
	private static final String packageName="tmp.contextmodel";
	private static String namespace=null;
	private static String outputFilename=null;
	
	// static instance of the transformer
	private static ContextModelTransformer staticClass=
		new ContextModelTransformer();
	
	/**
	 * Main.
	 * 
	 * @param args Arguments.
	 * @throws Exception If errors occur.
	 */
	public static void main(String[] args) throws Exception {		
		
		if (args.length<2) {
			System.err.println("Usage: "+
					ContextModelTransformer.class.getSimpleName()+
					" <inputfile> <outputfile>");
			System.exit(1);
		}
		
		SAXBuilder builder=new SAXBuilder();
		Document doc=builder.build(new File(args[0]));
		Element schemaRoot=doc.getRootElement();

		namespace=schemaRoot.getAttributeValue("targetNamespace");
		
		outputFilename=args[1];
		
		Class factoryClass=null;
		
		try {
			factoryClass=Class.forName(packageName+".ObjectFactory");			
		} catch (ClassNotFoundException e) {
			System.err.println("Cannot find ObjectFactory class.");
			System.exit(2);
		}

		// get list of types through create-methods of factory
		for (Method m : factoryClass.getMethods()) {
			String methodName=m.getName();
			if (methodName.startsWith("create") && 
					!methodName.endsWith("ContextID")) {
				
				process(m.getReturnType());
			}
		}
		
		File outFile=new File(outputFilename);
		
		outFile.delete();
		
		// output
		for (ContextType t : types.values()) {
			t.appendToWriter(new FileWriter(outFile,true));
		}
	}
	
	/**
	 * Processes the transformation of a class.
	 * 
	 * @param c Class.
	 */
	private static void process(Class c) {
		
		if (c.getPackage()==null || 
				!c.getPackage().getName().equals(packageName) || 
				c.isMemberClass()) {
			return;
		}

		
		HashMap<String,Field> fields=new HashMap<String,Field>();
		
		Class tmp=c;
		
		// get all fields, also of superclasses
		do  {
			for (Field f : tmp.getDeclaredFields()) {
				fields.put(f.getName(),f);
			}
		} while ((tmp=tmp.getSuperclass())!=null);

		// does it have a contextID? if no, quit 
		Field cid=fields.get("contextID");
		
		if (cid==null) {
			return;
		}

		// retrieve/create the resulting type
		ContextType contextType=getType(c.getSimpleName());

		// check all fields
		for (Field f : fields.values()) {

			// skip if contextID
			if (cid.equals(f)) {
				continue;
			}

			Class fieldType=f.getType();

			boolean isList=false;
			
			// sequences are stored as parametrized lists
			// retrieve parameter type
			if (!fieldType.equals(Object.class) && 
					fieldType.isAssignableFrom(List.class)) {

				isList=true;
				
				ParameterizedType ptype = (ParameterizedType)f.getGenericType();
				StringTokenizer tokens=
					new StringTokenizer(ptype.toString(),"<>");

				// skip first token
				tokens.nextToken();
				
				// second token is class name
				try {
					fieldType=Class.forName(tokens.nextToken());
				} catch (ClassNotFoundException e) {
					continue;
				}
			}

			// is field of a relevant type anyway?
			// this means, whether it is a member of the specified package
			if (fieldType.getPackage()!=null && 
					fieldType.getPackage().getName().equals(packageName) && 
					!fieldType.isMemberClass()) {

				// add parent -> child link
				getType(fieldType.getSimpleName()).parents.add(
						c.getSimpleName()+"@"+namespace);
				
				
				String childString=fieldType.getSimpleName()+"@"+namespace;
				
				if (isList) {
					childString="["+childString+"]";
				}
				
				getType(c.getSimpleName()).children.add(childString);
				
			}
			
		}
		
		// now retrieve all fields of the contextID tag
		Class idClass=cid.getType();
		
		HashSet<Field> idFields=new HashSet<Field>();
		
		Class tmpId=idClass;
		
		// get all fields, also of superclasses
		do  {
			for (Field f : tmpId.getDeclaredFields()) {
				idFields.add(f);
			}
		} while ((tmpId=tmpId.getSuperclass())!=null);
		
		for (Field f : idFields) {

			try {
				XmlAttribute xa=f.getAnnotation(XmlAttribute.class);
				
				String name=f.getName();
				
				// predefine some behavior
				if ("uuid".equals(name)) {
					contextType.fields.put(name, "generate(UUID)");
				} else if ("belongsTo".equals(name)) {
					contextType.fields.put(name, "localParent()");					
				} else {
					
					if (!xa.required()) {
						contextType.fields.put(name, "askUser(String):opt");						
					} else {
						contextType.fields.put(name, "askUser(String)");
					}
				}
			} catch (Exception e) {
				// no annotation? this means not an attribute
			}
		}
	}
	
	/**
	 * Gets/retrieves a type object.
	 * 
	 * @param name Object name.
	 * @return Type object.
	 */
	private static ContextType getType(String name) {
		
		ContextType result=types.get(name);
		
		if (result==null) {
			result=staticClass.new ContextType();
			result.name=name;
			types.put(name, result);
		}

		return result;
	}
	
	/**
	 * 
	 * @author Lukasz Juszczyk
	 *
	 */
	private class ContextType {		

		public String name;
		public Vector<String> children=new Vector<String>();
		public Vector<String> parents=new Vector<String>();
		
		public HashMap<String,String> fields=new HashMap<String,String>();

		protected void appendToWriter(Writer writer) throws IOException {
			
			try {
				Class.forName(packageName+"."+name);
			} catch (ClassNotFoundException e) {
				return;
			}
			
			if (fields.isEmpty()) {
				return;
			}
			
			BufferedWriter bw=new BufferedWriter(writer);
			
			bw.write("["+name+"@"+namespace+"]");
			bw.write("\n");
			
			bw.write("children=");
			for (int i=0;i<children.size();i++) {
				bw.write(children.get(i));
				if (i!=children.size()-1) {
					bw.write(", ");
				}
			}
			bw.write("\n");
			
			bw.write("parents=");
			for (int i=0;i<parents.size();i++) {
				bw.write(parents.get(i));
				if (i!=parents.size()-1) {
					bw.write(", ");
				}
			}
			bw.write("\n");
			
			for (String n: fields.keySet()) {
				bw.write("field."+n+"="+fields.get(n));
				bw.write("\n");
			}
			
			bw.write("\n");
			
			bw.flush();
		}
	}
}
