/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Vector;

import org.xmlpull.v1.XmlPullParserException;

import com.eaio.uuid.UUID;

import edu.wustl.mobilab.sliver.soap.Serialization;
import edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext;

import EDU.oswego.cs.dl.util.concurrent.Semaphore;
import at.ac.tuwien.vitalab.cosine.configuration.CIMSConfiguration;
import at.ac.tuwien.vitalab.cosine.contextrouting.ContextRoutingService;
import at.ac.tuwien.vitalab.cosine.contextrouting.IContextRoutingServiceCallback;
import at.ac.tuwien.vitalab.cosine.initservice.MobileDeviceInitService;
import at.ac.tuwien.vitalab.cosine.interfaces.IDataConsumerWS;
import at.ac.tuwien.vitalab.cosine.interfaces.IDataProviderWS;
import at.ac.tuwien.vitalab.cosine.interfaces.INotificationReceiverWS;
import at.ac.tuwien.vitalab.cosine.processing.notification.AContextQuery;
import at.ac.tuwien.vitalab.cosine.processing.notification.InactiveSubscriptionException;
import at.ac.tuwien.vitalab.cosine.processing.notification.NotificationReceiverHeader;
import at.ac.tuwien.vitalab.cosine.processing.notification.RemoteContextQuery;
import at.ac.tuwien.vitalab.cosine.processing.notification.SubscriptionClient;
import at.ac.tuwien.vitalab.cosine.processing.storage.QueryClient;
import at.ac.tuwien.vitalab.cosine.query.XQuery;
import at.ac.tuwien.vitalab.cosine.rating.ARating;
import at.ac.tuwien.vitalab.cosine.webservice.ASensorWebService;
import at.ac.tuwien.vitalab.cosine.webservice.MarshalContextData;
import at.ac.tuwien.vitalab.cosine.webservice.UpdatableSensorWebService;
import at.ac.tuwien.vitalab.middleware.AWebService;
import at.ac.tuwien.vitalab.middleware.Middleware;
import at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery;
import at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo;

/**
 * Context Management Service.
 * 
 * @author Lukasz Juszczyk
 * @author Florian Skopik
 *
 */
public class CMS extends AWebService 
implements IDataProviderWS, INotificationReceiverWS, IDataConsumerWS
{

	public static final String WSNAME="CMS";
	
	public static final String WSNAMESPACE=
		"http://www.vitalab.tuwien.ac.at/cosine/";
	
	protected Vector operations=new Vector();
	
	// local static CIMS
	private static CMS localCIMS=null;

	private SensorQuery sensorQuery=new SensorQuery();
	
	// routing service to update about sensors found
	private ContextRoutingService contextRouter=null;
	// sensor initialisation service
	private MobileDeviceInitService contextInit=null;

	// subscription map
	// key = AContextQuery object
	// value = SubscriptionCallback object
	private Hashtable subscriptions=new Hashtable();
		
	// subscription map
	// key = subscription ID
	// value = RemoteContextQuery object
	private Hashtable remoteSubscriptions=new Hashtable();
	
	// client managing subscription requests
	private SubscriptionClient subscriptionClient=null;

	// subscription requests
	private Hashtable subscriptionQueries=new Hashtable();

	// locally deployed sensor services
	private Hashtable localSensors=new Hashtable();
	
	static {
		new MarshalContextData().register(Serialization.getParentEnvelope());
	}
	
	/**
	 * Constructor.
	 * 
	 * @throws Exception If the CIMS cannot be initialized.
	 */
	protected CMS() throws Exception {
		
		super(WSNAME,WSNAMESPACE);
		
		operations.add("contextNotification");
		operations.add("contextListNotification");
		operations.add("storeContextInfo");
		operations.add("getContextInfo");
		operations.add("subscribe");
		operations.add("unsubscribe");
		
		contextRouter=ContextRoutingService.getInstance();
		contextInit=MobileDeviceInitService.getInstance();
		
		registerRatings();
		
		subscriptionClient=new SubscriptionClient(this);
	}
	
	/**
	 * Starts the middleware (SOAP and SLP).
	 * 
	 * @param args Arguments are ignored.
	 */
	public static void main(String args[]) {
		
		try {
			
			Middleware.start();
			
			final CMS cims=new CMS();
			
			cims.deploy();
			
			// run  until 'exit' is typed into stdin
			BufferedReader reader=
				new BufferedReader(new InputStreamReader(System.in));
			
			String inputLine=null;
			while ((inputLine=reader.readLine())!=null && 
					!"exit".equals(inputLine.toLowerCase())) {
			}
			
			cims.shutdown();
			
			Middleware.stop();
			
			System.exit(0);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}
	
	/**
	 * Returns the local CIMS instance. If not running yet it will be 
	 * deployed at the local middleware.
	 * 
	 * @return CIMS instance.
	 * @throws Exception If errors occur.
	 */ 
	public synchronized static CMS getLocalCIMS() throws Exception {
		
		if (localCIMS==null) {
			localCIMS=new CMS();
			
			Middleware.start();
			
			localCIMS.deploy();
		}
		
		return localCIMS;
	}
	
	/**
	 * Shuts down the local CIMS instance.
	 * 
	 * @param shutdownMiddleware Whether the middleware has to be shut
	 * down as well.
	 * @throws Exception If errors occur.
	 */
	public synchronized static void shutdownLocalCIMS(
			boolean shutdownMiddleware) throws Exception {
	
		if (localCIMS==null) {
			throw new IllegalStateException("Local CIMS not running.");
		}
		
		localCIMS.shutdown();
		
		localCIMS=null;
		
		if (shutdownMiddleware) {
			Middleware.stop();
		}
	}
	
	/**
	 * Deploys the CIMS at the local middleware.
	 * @throws Exception If the middleware cannot be started.
	 *
	 */
	protected void deploy() throws Exception {
		Middleware.getServiceDatabase().subscribe(sensorQuery, sensorQuery);
		
		Middleware.deployService(this);
		
		logger.info("CIMS has been successfully deployed.");
	}
	
	/**
	 * Undeploys the CIMS from the local middleware.
	 *
	 */
	public void shutdown() {		
		Middleware.undeployService(this);
		
		Middleware.getServiceDatabase().unsubscribe(sensorQuery);
		
		logger.info("CIMS has been successfully undeployed.");
	}	
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#onDeploy()
	 */
	public void onDeploy() {
		startSensors();
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#onUndeploy()
	 */
	public void onUndeploy() {
	}
	
	// *************** MIDDLEWARE FUNCTIONS ******************

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getAttributes()
	 */
	public Hashtable getAttributes() {
		//TODO: put in some attributes
		return null;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getConfigurationRessourceName()
	 */
	public String getConfigurationRessourceName() {
		return null;
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getOperationNames()
	 */
	public Collection getOperationNames() {
		return new Vector(operations);
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getOperationDescription(java.lang.String)
	 */
	public Hashtable getOperationDescription(String operationName) {
		if (operations.contains(operationName)) {
			return new Hashtable();
		} 
		throw new IllegalArgumentException("Invalid/unknown operation.");		
	}
	
// *************** WEB SERVICE OPERATIONS ******************

	/**
	 * Retrieve metrics from a context notification.
	 * 
	 * @param id Subscription ID.
	 * @param context Context Data.
	 */
	private void retrieveMetrics(String id, ContextData context) {
		
		ServiceInfo service=subscriptionClient.getServiceBySubscriptionID(id);
		
		if (service==null) {
			return;
		}
		
		Iterator ratingIter=ARating.getRatingModules().iterator();
		
		while (ratingIter.hasNext()) {
			ARating rating=(ARating)ratingIter.next();
			
			try {
				double value=rating.getRating(service, context);
				
				contextRouter.updateServiceRating(service, rating, value);
			} catch (IllegalArgumentException iae) {
				// ignore
			}
		}
	}
	
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.IDataConsumerWS#contextNotification(java.lang.String, at.ac.tuwien.vitalab.cosine.ContextData)
	 */
	public void contextNotification(String id, ContextData context) {

		subscriptionClient.checkID(id);
		
		subscriptionClient.refreshActivity(id);
		
		AContextQuery query=(AContextQuery)subscriptionQueries.get(id);
		
		if (query!=null) {
			try {
				retrieveMetrics(id,context);
				query.contextQueryNotification(context);				
			} catch (InactiveSubscriptionException e) {
				try {
					if (query instanceof RemoteContextQuery) {
						String subID=
							((RemoteContextQuery)query).getSubscriptionID();
						this.unsubscribe(subID);
					} else {
						this.unsubscribeLocally(query);						
					}
				} catch (Exception e1) {
					// ignore
				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.IDataConsumerWS#contextListNotification(java.lang.String, java.util.Vector)
	 */
	public void contextListNotification(String id, Vector contextList) {
		
		subscriptionClient.checkID(id);
		
		subscriptionClient.refreshActivity(id);
		
		AContextQuery query=(AContextQuery)subscriptionQueries.get(id);
		
		if (query!=null) {
			try {
				Iterator contextIter=contextList.iterator();
				while (contextIter.hasNext()) {
					retrieveMetrics(id,(ContextData)contextIter.next());
				}
				query.contextQueryNotification(contextList);
			} catch (InactiveSubscriptionException e) {
				try {
					if (query instanceof RemoteContextQuery) {
						String subID=
							((RemoteContextQuery)query).getSubscriptionID();
						this.unsubscribe(subID);
					} else {
						this.unsubscribeLocally(query);						
					}
				} catch (Exception e1) {
					// ignore
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.IUpdateReceiverWS#storeContextInfo(at.ac.tuwien.vitalab.cosine.ContextData)
	 */
	public void storeContextInfo(ContextData context) throws Exception {
		
		String belongsTo=
			(String)context.getContextIDAttributes().get("belongsTo");
		
		String name=context.getName();
		String namespace=context.getNamespace();
		
		String key=name+" "+namespace+" "+belongsTo;
		
		UpdatableSensorWebService sensor=
			(UpdatableSensorWebService)localSensors.get(key);
		
		if (sensor==null) {
			
			sensor=new UpdatableSensorWebService(name,namespace,true,true,true);

			contextInit.deploySensor(sensor);
			localSensors.put(name+" "+namespace+" "+belongsTo, sensor);
		}
		
		sensor.storeContextInfo(context);
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.IDataProviderWS#getContextInfo(java.lang.String)
	 */
	public Vector getContextInfo(final String query) throws Exception {
		
		final Object[] dest=contextRouter.findQueryDestination(query);
		
		ServiceInfo[] locations=(ServiceInfo[])dest[1];

		if (locations==null || locations.length==0) {
			throw new Exception("No suitable source for query found.");
		}
		
		final Vector context=new Vector();
		
		// just one location?
		
		if (locations.length==1) {
			ServiceInfo sensor=locations[0];

			String queryString=(String)dest[0];
			context.addAll(QueryClient.getContextInfo(queryString,sensor));									

			return context;
		}
		
		// or multiple locations?
		
		final Semaphore sem=new Semaphore(1);
		
		final Vector errors=new Vector();
		
		for (int i=0;i<locations.length;i++) {
			final ServiceInfo sensor=locations[i];

			sem.acquire();

			new Thread() {

				/* (non-Javadoc)
				 * @see java.lang.Thread#run()
				 */
				public void run() {
					try {
						String queryString=(String)dest[0];
						context.addAll(QueryClient.getContextInfo(
								queryString,sensor));									
					} catch (Exception e) {
						errors.add(e);
					} finally {
						sem.release();
					}
				}
			}.start();
		}

		sem.acquire();
		
		if (context.isEmpty() && !errors.isEmpty()) {
			throw (Exception)errors.firstElement();
		}
		
		return context;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.ISubscriptionProviderWS#subscribe(java.lang.String, java.lang.String, edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext)
	 */
	public boolean subscribe(String query, String id, InvocationContext context) 
	throws Exception {
		
		ContextData tmpCD=new ContextData("a","http://test");
		
		// to filter out XQuery string errors before they are being sent
		// to the sensors
		XQuery.returnMatches(tmpCD, query);
		
		synchronized (remoteSubscriptions) {
			
			if (remoteSubscriptions.containsKey(id)) {
				return true;
			}
			
			try {

				NotificationReceiverHeader receiverHeader=
					new NotificationReceiverHeader(
							context.incomingChannel.getIncomingXMLHeaders());
				
				RemoteContextQuery remoteQuery=
					new RemoteContextQuery(query,id,receiverHeader);
				
				remoteSubscriptions.put(id, remoteQuery);

				this.subscribeLocally(remoteQuery);
				
			} catch (NoSuchElementException nse) {
				throw new Exception(
						"ContextNotificationReceiver-header missing.");
			} catch (XmlPullParserException xppe) {
				throw new Exception("ContextNotificationReceiver-header " +
						"error: "+xppe.getMessage());
			}

			return false;
		}
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.IDataProviderWS#unsubscribe(java.lang.String)
	 */
	public void unsubscribe(String id) throws Exception {
		
		synchronized (remoteSubscriptions) {
			
			RemoteContextQuery remoteQuery=
				(RemoteContextQuery)remoteSubscriptions.remove(id);
			
			if (remoteQuery==null) {
				throw new IllegalArgumentException("Unknown subscription ID.");				
			}
			
			this.unsubscribeLocally(remoteQuery);
			
			remoteSubscriptions.remove(id);
			
		}
	}

//	// *************** LOCAL SUBSRICPTIONS ****************
	

	/**
	 * Places a local subscription.
	 * 
	 * @param query Query object.
	 * @return Whether this susbcription has been registered before.
	 * @throws Exception If the query is invalid.
	 */
	public boolean subscribeLocally(AContextQuery query) throws Exception {
	
		synchronized (subscriptions) {
			if (subscriptions.contains(query)) {
				return true;
			}
			
			SubscriptionCallback callback=new SubscriptionCallback(this, query);
			
			subscriptions.put(query, callback);
			
			contextRouter.findSubscriptionDestination(
					query.getQueryString(), new UUID().toString(), callback);
			
			return false;
		}
	}

	/**
	 * Removes a subscription.
	 * 
	 * @param query Query object.
	 * @throws Exception If the subscription has not been registered before.
	 */
	public void unsubscribeLocally(AContextQuery query) throws Exception {
		
		synchronized (subscriptions) {

			SubscriptionCallback callback=
				(SubscriptionCallback)subscriptions.remove(query);
			
			if (callback==null) {
				throw new Exception(
						"Subscriptions has not been placed before.");
			}
			
			callback.cleanup();
			
			query.stop();
		}
	}
	
	private class SubscriptionCallback 
	implements IContextRoutingServiceCallback {

		private Hashtable sensorSubscriptions=new Hashtable();
		private AContextQuery query=null;
		private HashSet oldSet=null;
		
		/**
		 * Constructor.
		 * 
		 * @param cims CIMS instance.
		 * @param query Query object.
		 */
		public SubscriptionCallback(CMS cims, AContextQuery query) {
			super();
			this.query=query;
		}

		/* (non-Javadoc)
		 * @see at.ac.tuwien.vitalab.cosine.contextrouting.IContextRoutingServiceCallback#queryUpdate(java.lang.Object[], java.lang.String)
		 */
		public synchronized void queryUpdate(Object[] resultArray, 
				String idSubscription) {
			
			
			String queryString=(String)resultArray[0];
			ServiceInfo[] services=(ServiceInfo[])resultArray[1];

			if (services==null) {
				return;
			}

			HashSet set=new HashSet();
			
			for (int i=0;i<services.length;i++) {
				ServiceInfo s=services[i];
				
				if (!sensorSubscriptions.containsValue(s)) {
					String id=subscriptionClient.subscribeTo(s, queryString);		
					sensorSubscriptions.put(id,s);
					subscriptionQueries.put(id, query);
					set.add(id);
				} else {
					set.add(subscriptionClient.getSubscriptionIDByService(s));
				}
				
			}
			
			if (oldSet!=null) {
				oldSet.removeAll(set);
				
				Iterator oldIter=oldSet.iterator();
				
				while (oldIter.hasNext()) {
					String oldId=oldIter.next().toString();
					subscriptionClient.unsubscribeFrom(oldId);
					sensorSubscriptions.remove(oldId);
					subscriptionQueries.remove(oldId);
				}
			}

			oldSet=set;
		}
		
		/**
		 * Closes all subscriptions;
		 *
		 */
		protected void cleanup() {
			Iterator serIter=sensorSubscriptions.keySet().iterator();
			
			while (serIter.hasNext()) {
				String id=(String)serIter.next();
				subscriptionClient.unsubscribeFrom(id);
				subscriptionQueries.remove(id);
			}
			sensorSubscriptions.clear();
		}
	}
	
	// *************** RATING MODULES ******************
	
	private void registerRatings() {
		
//		List classNames=CIMSConfiguration.getRatingClasses();
		List classNames=CIMSConfiguration.getRatingClassesWithLimits();
		
		Iterator classNameIter=classNames.iterator();
		
		while (classNameIter.hasNext()) {
			//final String className=classNameIter.next().toString();
			Vector v = (Vector)classNameIter.next();
			final String className=v.get(0).toString();

			try {

				Class clazz=Thread.currentThread().
				getContextClassLoader().loadClass(className);
				
				ARating rating=(ARating)clazz.newInstance();
				
				if (v.size() ==  4) {
					double min = Double.parseDouble(v.get(1).toString());
					double max = Double.parseDouble(v.get(2).toString());
					double weight = Double.parseDouble(v.get(3).toString());

			
					rating.setMaxRating(max);
					rating.setMinRating(min);
					rating.setWeightRating(weight);					
				}
				
				logger.info("Rating module '"+rating.getClass().getName()+
						"' has been registered.");
				
			} catch (ClassNotFoundException e) {
				logger.warning("Cannot find rating plugin'"+
						className+"'.");
			} catch (Exception e) {
				logger.error("Cannot initialize rating plugin '"+
						className+"': "+e.getMessage());
				e.printStackTrace();
			}			
		}
	}
	
	
	// *************** SENSOR MANAGEMENT ******************
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.INotificationReceiverWS#registerSubscriptionClient(java.lang.String, at.ac.tuwien.vitalab.cosine.processing.notification.SubscriptionClient)
	 */
	public void registerSubscriptionClient(String id, SubscriptionClient sc) {
		// not necessary
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.INotificationReceiverWS#unregisterSubscriptionClient(java.lang.String)
	 */
	public void unregisterSubscriptionClient(String id) {
		// not necessary
	}

	private void startSensors() {
	
		List classNames=CIMSConfiguration.getSensorClasses();
		
		Iterator classNameIter=classNames.iterator();
		
		while (classNameIter.hasNext()) {
			final String className=classNameIter.next().toString();

			try {

				Class clazz=Thread.currentThread().
				getContextClassLoader().loadClass(className);
				
				ASensorWebService sensor=(ASensorWebService)clazz.newInstance();
				
				contextInit.deploySensor(sensor);
				
				if (sensor instanceof UpdatableSensorWebService) {

					String belongsTo=sensor.getBelongsTo();
					String name=sensor.getContextTypeName();
					String namespace=sensor.getContextTypeNamespace();

					localSensors.put(name+" "+namespace+" "+belongsTo, sensor);
				}
				
			} catch (ClassNotFoundException e) {
				logger.warning("Cannot find sensor '"+className+"'.");
			} catch (Exception e) {
				logger.error("Cannot initialize sensor '"+
						className+"': "+e.getMessage());
				e.printStackTrace();
			}			
		}
	}
	
	/**
	 * Query matching all sensors in the network.
	 * 
	 * @author Lukasz Juszczyk
	 *
	 */
	private class SensorQuery extends AServiceQuery 
	implements IServiceSubscriber {

		/**
		 * Constructor.
		 *
		 */
		public SensorQuery() {
		}
		
		/* (non-Javadoc)
		 * @see at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery#matches(at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
		 */
		protected boolean matches(ServiceInfo info) {
			
			// match all sensors which support subscription
			return (ASensorWebService.WSATTR_VALUE_SENSOR.equals(
					info.getAttribute(ASensorWebService.WSATTR_NAME_SENSOR)) &&
					"supported".equalsIgnoreCase(info.getAttribute(
							"subscription")));
		}

		/* (non-Javadoc)
		 * @see at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber#notification(at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery, at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
		 */
		public void notification(AServiceQuery query, ServiceInfo info) {			
			contextRouter.newService(info);
		}
	}
}
