/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine;

import java.io.IOException;
import java.util.HashSet;
import java.util.Hashtable;

import org.kxml2.kdom.Element;
import org.xmlpull.v1.XmlPullParserException;


/**
 * XML Context Data.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class ContextData extends XMLData {

	public static final String NAMESPACE=CMS.WSNAMESPACE+"contextdata";
	
	public static final String NAME="ContextData";
	
	public static final String IDNAME="contextID";
	
	public static final String ATT_CONTEXT_TS="contextTimestamp";
	public static final String ATT_SENDING_TS="sendingTimestamp";
	
	/**
	 * Constructor.
	 * 
	 * @param name Name.
	 * @param namespace Namespace.
	 */
	public ContextData(String name, String namespace) {
		super(name,namespace);
	}

	/**
	 * Constructor.
	 * 
	 * @param xmlString XML string.
	 * @throws XmlPullParserException If the string does contain an invalid 
	 * XML structure.
	 * @throws IOException If IO errors occur while reading.
	 */
	public ContextData(String xmlString) throws XmlPullParserException, 
	IOException {
		super(xmlString);
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.XMLData#clone()
	 */
	public Object clone(){
		return super.clone();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object arg0) {
		XMLData data;
		try {
			data=(XMLData)arg0;			
		} catch (ClassCastException cce) {
			return false;
		}
		
		HashSet ignoredAttribNames=new HashSet();
		
		ignoredAttribNames.add(ATT_CONTEXT_TS);
		ignoredAttribNames.add(ATT_SENDING_TS);
		
		return compare(this,data,ignoredAttribNames,null);
	}
	
	/**
	 * Sets the sending timestamp to the current value.
	 */
	public void refreshSendingTimestamp() {
		setAttribute(null, ATT_SENDING_TS, ""+System.currentTimeMillis());
	}
	
	/**
	 * Returns the sending timestamp.
	 * 
	 * @return Timestamp or -1 if the timestamp is not present or not valid.
	 */
	public long getSendingTimestamp() {
		String tsAtt=this.getAttributeByName(ATT_SENDING_TS);
		if (tsAtt==null) {
			return -1;
		}
		try {
			return Long.parseLong(tsAtt);			
		} catch (NumberFormatException nfe) {
			return -1;
		}
	}
	
	/**
	 * Returns the context timestamp.
	 * 
	 * @return Timestamp or -1 if the timestamp is not present or not valid.
	 */
	public long getContextTimestamp() {
		String tsAtt=this.getAttributeByName(ATT_CONTEXT_TS);
		if (tsAtt==null) {
			return -1;
		}
		try {
			return Long.parseLong(tsAtt);			
		} catch (NumberFormatException nfe) {
			return -1;
		}
	}
	
	/**
	 * Returns the contextID-element.
	 * 
	 * @return ContextID-element or <code>null</code> if none was found.
	 */
	public XMLData getContextIDElement() {
		Element id=getChildByName(IDNAME,NAMESPACE);
		return XMLData.create(id);
	}
		
	/**
	 * Returns the UUID of this context element.
	 * 
	 * @return UUID String or <code>null</code> if none was found.
	 */
	public String getUUID() {
		Element contextID=getContextIDElement();
		if (contextID==null) {
			return null;
		}
		String r=contextID.getAttributeValue(null, "uuid");
		if (r==null) {
			return null;
		}
		return r.intern();
	}
	
	/**
	 * Returns the UUID of the parent it belongs to.
	 * 
	 * @return UUID String or <code>null</code> if none was found.
	 */
	public String getBelongsTo() {
		Element contextID=getContextIDElement();
		if (contextID==null) {
			return null;
		}
		String r=contextID.getAttributeValue(null, "belongsTo");
		if (r==null) {
			return null;
		}
		return r.intern();
	}
	
	/**
	 * Returns a Hashtable containing all contextID attributes.
	 * 
	 * @return Hashtable or <code>null</code> if no contextID was found.
	 */
	public Hashtable getContextIDAttributes() {
		Element contextID=getContextIDElement();
		if (contextID==null) {
			return null;
		}
		
		Hashtable result=new Hashtable();
		for (int i=0;i<contextID.getAttributeCount();i++) {
			String name=contextID.getAttributeName(i).intern();
			String value=contextID.getAttributeValue(i).intern();
			result.put(name, value);
		}
		return result;
	}
	
	/**
	 * Returns a child element which has the passed context ID.
	 * 
	 * @param contextID Context ID.
	 * @return Child element or <code>null</code> if none was found.
	 */
	public XMLData getChildWithContextID(XMLData contextID) {
		
		for(int i=0; i<getChildCount(); i++) {
			
			Object c=getChild(i);
			if (!(c instanceof Element)) {
				continue;
			}
			XMLData child = XMLData.create((Element)c);
			
			Element idElement=child.getChildByName(IDNAME, NAMESPACE);
			
			if (idElement==null) {
				continue;
			}
			
			if (compare(idElement, contextID, null, null)) {
				return child;
			}
		}
		
		return null;
	}
	
	/**
	 * Removes a child element which has the passed context ID.
	 * 
	 * @param contextID Context ID.
	 * @return <code>true</code> if the child could be removed, otherwise
	 * <code>false</code>.
	 */
	public boolean deleteChildWithContextID(XMLData contextID) {
		
		for(int i=0; i<getChildCount(); i++) {
			
			Object c=getChild(i);
			if (!(c instanceof Element)) {
				continue;
			}
			XMLData child = XMLData.create((Element)c);
			
			Element idElement=child.getChildByName(IDNAME, NAMESPACE);
			
			if (idElement==null) {
				continue;
			}
			
			if (compare(idElement, contextID, null, null)) {
				
				this.removeChild(i);
				return true;
			}
		}
		
		return false;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
			
		HashSet ignoredAttribNames=new HashSet();
		
		ignoredAttribNames.add(ATT_CONTEXT_TS);
		ignoredAttribNames.add(ATT_SENDING_TS);
		
		return _hashCode(this,1,ignoredAttribNames,null);		
	}
}
