/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Hashtable;
import java.util.Set;

import org.kxml2.io.KXmlParser;
import org.kxml2.io.KXmlSerializer;
import org.kxml2.kdom.Element;
import org.kxml2.kdom.Node;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlSerializer;


/**
 * XML Data.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class XMLData extends Element implements Cloneable {

	private Element el=null;
	
	/**
	 * Constructor.
	 * 
	 * @param name Name.
	 * @param namespace Namespace.
	 */
	public XMLData(String name, String namespace) {
		
		el=new Element();		
		el.setName(name);
		el.setNamespace(namespace);
	}

	/**
	 * Constructor.
	 * 
	 * @param el XML Element.
	 */
	public XMLData(Element el) {
		this.el = el;
	}

	/**
	 * Constructor.
	 * 
	 * @param xmlString XML string.
	 * @throws XmlPullParserException If the string does contain an invalid 
	 * XML structure.
	 * @throws IOException If IO errors occur while reading.
	 */
	public XMLData(String xmlString) throws XmlPullParserException, 
	IOException {
		
		this(new StringReader(xmlString));
	}
	
	/**
	 * Constructor.
	 * 
	 * @param reader Input reader.
	 * @throws XmlPullParserException If the string does contain an invalid 
	 * XML structure.
	 * @throws IOException If IO errors occur while reading.
	 */
	public XMLData(Reader reader) throws XmlPullParserException, 
	IOException {
		
		KXmlParser parser=new KXmlParser();
		parser.setInput(reader);
		
		parser.setFeature(KXmlParser.FEATURE_PROCESS_NAMESPACES, true);
		
		parser.nextTag();

		el=new Element();
		
		el.setName(parser.getName());
		el.setNamespace(parser.getNamespace());

		if (parser.getEventType()!=XmlPullParser.END_TAG) {
			el.parse(parser);		   
		} 
		
		reader.close();
	}
	
	/**
	 * Wraps an element into an XMLData object.
	 * 
	 * @param element XML element.
	 * @return XMLData object.
	 */
	public static XMLData create(Element element) {
		if (element==null) {
			return null;
		}
		if (element instanceof XMLData) {
			return (XMLData)element;
		}
		return new XMLData(element);
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		KXmlSerializer serializer=new KXmlSerializer();
		StringWriter writer=new StringWriter();
		serializer.setOutput(writer);
		
		try {
			el.write(serializer);
		} catch (IOException e) {
			// shout not happen
			return "";
		}
		
		return writer.toString();
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
			
		return _hashCode(this,1, null, null);
	}
	
	protected static int _hashCode(Element e, int mult, Set ignoredAttribnames, 
			Set ignoredElementNames) {
		
		int h=0;
		
		if (e.getNamespace()!=null) {
			h+=(e.getNamespace()+":"+e.getName()).hashCode()*(mult+0.2);
		} else {
			h+=(e.getName()).hashCode()*(mult+0.2);
		}
		
		for (int i=0;i<e.getAttributeCount();i++) {
			String name=e.getAttributeName(i);
			
			if (ignoredAttribnames==null || !ignoredAttribnames.contains(name)){
				if (e.getAttributeNamespace(i)!=null) {
					h+=(e.getAttributeNamespace(i)+":"+
							name+"="+e.getAttributeValue(i)).hashCode()*mult;
				} else {
					h+=(name+"="+e.getAttributeValue(i)).hashCode()*mult;
				}
			}
		}

		for (int i=0;i<e.getChildCount();i++) {
			
			Object child=e.getChild(i);
			int childType=e.getType(i);
			
			switch (childType) {
			case Node.ELEMENT:
				Element el=(Element)child;
				if (ignoredElementNames==null || 
						!ignoredElementNames.contains(el.getName())) {
					h+=_hashCode(el,mult+1,
							ignoredAttribnames,ignoredElementNames)*mult;
				}
				break;
			case Node.TEXT:
			case Node.COMMENT:
			case Node.CDSECT:
				h+=e.getText(i).hashCode()*mult;
				break;
			default:
			}
		}
		
		return h;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object arg0) {
		XMLData data;
		try {
			data=(XMLData)arg0;			
		} catch (ClassCastException cce) {
			return false;
		}
		
		return compare(this,data,null,null);
	}

	protected boolean compare(Element a, Element b, Set ignoredAttribnames, 
			Set ignoredElementNames) {
		
		if (a==b) {
			return true;
		}
		
		// simple comparison
		if (!a.getName().equals(b.getName()) ||
				!a.getNamespace().equals(b.getNamespace()) ||
				a.getChildCount()!=b.getChildCount() ||
				a.getAttributeCount()!=b.getAttributeCount()) {
			return false;
		}
		
		// compare attributes
		for (int i=0;i<a.getAttributeCount();i++) {
			
			String nameA=a.getAttributeName(i);
			String nsA=a.getAttributeNamespace(i);
			String nameB=b.getAttributeName(i);
			String nsB=b.getAttributeNamespace(i);
			
			if (ignoredAttribnames!=null && 
					ignoredAttribnames.contains(nameA) &&
					ignoredAttribnames.contains(nameB)) {
				continue;
			}
			
			if (!a.getAttributeName(i).equals(b.getAttributeName(i)) ||
					!a.getAttributeValue(i).equals(b.getAttributeValue(i)) ||
					(a.getAttributeNamespace(i)!=null && 
							!a.getAttributeNamespace(i).equals(
									b.getAttributeNamespace(i))) ||
					(b.getAttributeNamespace(i)!=null && 
							!b.getAttributeNamespace(i).equals(
									a.getAttributeNamespace(i)))) {
				return false;
			}
		}
		
		// compare children		
		for (int i=0;i<a.getChildCount();i++) {
			Object cA=a.getChild(i);
			Object cB=b.getChild(i);
			
			if (cA instanceof Element) {
				if (!(cB instanceof Element)) {
					return false;
				}
				
				String nameA=((Element)cA).getName();
				String nameB=((Element)cB).getName();
				
				if (ignoredElementNames!=null && 
						ignoredElementNames.contains(nameA) &&
						ignoredElementNames.contains(nameB)) {
					continue;
				}
				
				if (!compare((Element)cA,(Element)cB,
						ignoredAttribnames,ignoredElementNames)) {
					return false;
				}
			} else if (cB instanceof Element) {
				return false;
			}
		}
		
		return true;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public Object clone() {
		return new XMLData(cloneElement(el));
	}

	/**
	 * Clones an element.
	 * 
	 * @param e Element.
	 * @return Clone of the element.
	 */
	private static Element cloneElement(Element e) {
		Element r=new Element();
		
		r.setName(e.getName());
		r.setNamespace(e.getNamespace());

		for (int i=0;i<e.getAttributeCount();i++) {
			r.setAttribute(
					e.getAttributeNamespace(i), 
					e.getAttributeName(i), 
					e.getAttributeValue(i));
		}

		for (int i=0;i<e.getChildCount();i++) {
			int type=e.getType(i);
			Object child=e.getChild(i);
			switch (type) {
			case Node.ELEMENT:
				r.addChild(type, cloneElement((Element)child));
				break;
			case Node.TEXT:
			case Node.COMMENT:
			case Node.CDSECT:
				r.addChild(type, child.toString());
			default:
			}
		}
		
		return r;
	}
	
	/**
	 * Returns a child element by its name.
	 * 
	 * @param childName Name.
	 * @return Child element or <code>null</code> if nothing was found.
	 */
	public XMLData getChildByName(String childName) {
		
		for(int i=0; i<el.getChildCount(); i++) {
			Object c=el.getChild(i);
			if (!(c instanceof Element)) {
				continue;
			}
			Element child = (Element)c;
			if (childName.equals(child.getName())) {
				return XMLData.create(child);
			}
		}
		return null;
	}	

	/**
	 * Returns a child element by its name.
	 * 
	 * @param childName Name.
	 * @param childNamespace Namespace.
	 * @return Child element or <code>null</code> if nothing was found.
	 */
	public XMLData getChildByName(String childName, String childNamespace) {
		
		for(int i=0; i<el.getChildCount(); i++) {
			Object c=el.getChild(i);
			if (!(c instanceof Element)) {
				continue;
			}
			Element child = (Element)c;
			if (childName.equals(child.getName()) && 
					childNamespace.equals(child.getNamespace())) {
				return XMLData.create(child);
			}
		}
		return null;
	}	

	/**
	 * Returns an attribute value by its name.
	 * 
	 * @param attName Name.
	 * @return Attribute value or <code>null</code> if nothing was found.
	 */
	public String getAttributeByName(String attName) {
		
		for(int i=0; i<el.getAttributeCount(); i++) {
			if (attName.equals(el.getAttributeName(i))) {
				return el.getAttributeValue(i).intern();				
			}
		}
		return null;
	}	

	/**
	 * Returns an attribute value by its name.
	 * 
	 * @param attName Name.
	 * @param attNamespace Namespace.
	 * @return Attribute value or <code>null</code> if nothing was found..
	 */
	public String getAttributeByName(String attName, String attNamespace) {
		
		for(int i=0; i<el.getAttributeCount(); i++) {
			if (attName.equals(el.getAttributeName(i)) &&
					attNamespace.equals(el.getAttributeNamespace(i))) {
				return el.getAttributeValue(i).intern();				
			}
		}
		return null;
	}
	
	/**
	 * Returns all attributes.
	 * 
	 * @return Hashtable.
	 */
	public Hashtable getAttributes() {
		
		Hashtable result=new Hashtable();
		
		for(int i=0; i<el.getAttributeCount(); i++) {
			result.put(
					el.getAttributeName(i).intern(), 
					el.getAttributeValue(i).intern());
		}
		
		return result;
		
	}

	
	// DELEGATION METHODS ++++++++++++++++++++++++++++++++++++++++++++
	
	/**
	 * @param arg0
	 * @param arg1
	 * @param arg2
	 * @see org.kxml2.kdom.Node#addChild(int, int, java.lang.Object)
	 */
	public void addChild(int arg0, int arg1, Object arg2) {
		el.addChild(arg0, arg1, arg2);
	}

	/**
	 * @param arg0
	 * @param arg1
	 * @see org.kxml2.kdom.Node#addChild(int, java.lang.Object)
	 */
	public void addChild(int arg0, Object arg1) {
		el.addChild(arg0, arg1);
	}

	/**
	 * 
	 * @see org.kxml2.kdom.Element#clear()
	 */
	public void clear() {
		el.clear();
	}

	/**
	 * @param arg0
	 * @param arg1
	 * @return
	 * @see org.kxml2.kdom.Element#createElement(java.lang.String, java.lang.String)
	 */
	public Element createElement(String arg0, String arg1) {
		return el.createElement(arg0, arg1);
	}

	/**
	 * @return
	 * @see org.kxml2.kdom.Element#getAttributeCount()
	 */
	public int getAttributeCount() {
		return el.getAttributeCount();
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Element#getAttributeName(int)
	 */
	public String getAttributeName(int arg0) {
		return el.getAttributeName(arg0);
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Element#getAttributeNamespace(int)
	 */
	public String getAttributeNamespace(int arg0) {
		return el.getAttributeNamespace(arg0);
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Element#getAttributeValue(int)
	 */
	public String getAttributeValue(int arg0) {
		return el.getAttributeValue(arg0);
	}

	/**
	 * @param arg0
	 * @param arg1
	 * @return
	 * @see org.kxml2.kdom.Element#getAttributeValue(java.lang.String, java.lang.String)
	 */
	public String getAttributeValue(String arg0, String arg1) {
		return el.getAttributeValue(arg0, arg1);
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Node#getChild(int)
	 */
	public Object getChild(int arg0) {
		return el.getChild(arg0);
	}

	/**
	 * @return
	 * @see org.kxml2.kdom.Node#getChildCount()
	 */
	public int getChildCount() {
		return el.getChildCount();
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Node#getElement(int)
	 */
	public Element getElement(int arg0) {
		return el.getElement(arg0);
	}

	/**
	 * @param arg0
	 * @param arg1
	 * @return
	 * @see org.kxml2.kdom.Node#getElement(java.lang.String, java.lang.String)
	 */
	public Element getElement(String arg0, String arg1) {
		return el.getElement(arg0, arg1);
	}

	/**
	 * @return
	 * @see org.kxml2.kdom.Element#getName()
	 */
	public String getName() {
		return el.getName();
	}

	/**
	 * @return
	 * @see org.kxml2.kdom.Element#getNamespace()
	 */
	public String getNamespace() {
		return el.getNamespace();
	}

	/**
	 * @return
	 * @see org.kxml2.kdom.Element#getNamespaceCount()
	 */
	public int getNamespaceCount() {
		return el.getNamespaceCount();
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Element#getNamespacePrefix(int)
	 */
	public String getNamespacePrefix(int arg0) {
		return el.getNamespacePrefix(arg0);
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Element#getNamespaceUri(int)
	 */
	public String getNamespaceUri(int arg0) {
		return el.getNamespaceUri(arg0);
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Element#getNamespaceUri(java.lang.String)
	 */
	public String getNamespaceUri(String arg0) {
		return el.getNamespaceUri(arg0);
	}

	/**
	 * @return
	 * @see org.kxml2.kdom.Element#getParent()
	 */
	public Node getParent() {
		return el.getParent();
	}

	/**
	 * @return
	 * @see org.kxml2.kdom.Element#getRoot()
	 */
	public Node getRoot() {
		return el.getRoot();
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Node#getText(int)
	 */
	public String getText(int arg0) {
		return el.getText(arg0);
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Node#getType(int)
	 */
	public int getType(int arg0) {
		return el.getType(arg0);
	}

	/**
	 * @param arg0
	 * @param arg1
	 * @param arg2
	 * @return
	 * @see org.kxml2.kdom.Node#indexOf(java.lang.String, java.lang.String, int)
	 */
	public int indexOf(String arg0, String arg1, int arg2) {
		return el.indexOf(arg0, arg1, arg2);
	}

	/**
	 * 
	 * @see org.kxml2.kdom.Element#init()
	 */
	public void init() {
		el.init();
	}

	/**
	 * @param arg0
	 * @return
	 * @see org.kxml2.kdom.Node#isText(int)
	 */
	public boolean isText(int arg0) {
		return el.isText(arg0);
	}

	/**
	 * @param arg0
	 * @throws IOException
	 * @throws XmlPullParserException
	 * @see org.kxml2.kdom.Element#parse(org.xmlpull.v1.XmlPullParser)
	 */
	public void parse(XmlPullParser arg0) throws IOException, XmlPullParserException {
		el.parse(arg0);
	}

	/**
	 * @param arg0
	 * @see org.kxml2.kdom.Node#removeChild(int)
	 */
	public void removeChild(int arg0) {
		el.removeChild(arg0);
	}

	/**
	 * @param arg0
	 * @param arg1
	 * @param arg2
	 * @see org.kxml2.kdom.Element#setAttribute(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void setAttribute(String arg0, String arg1, String arg2) {
		el.setAttribute(arg0, arg1, arg2);
	}

	/**
	 * @param arg0
	 * @see org.kxml2.kdom.Element#setName(java.lang.String)
	 */
	public void setName(String arg0) {
		el.setName(arg0);
	}

	/**
	 * @param arg0
	 * @see org.kxml2.kdom.Element#setNamespace(java.lang.String)
	 */
	public void setNamespace(String arg0) {
		el.setNamespace(arg0);
	}

	/**
	 * @param arg0
	 * @param arg1
	 * @see org.kxml2.kdom.Element#setPrefix(java.lang.String, java.lang.String)
	 */
	public void setPrefix(String arg0, String arg1) {
		el.setPrefix(arg0, arg1);
	}

	/**
	 * @param arg0
	 * @throws IOException
	 * @see org.kxml2.kdom.Element#write(org.xmlpull.v1.XmlSerializer)
	 */
	public void write(XmlSerializer arg0) throws IOException {
		el.write(arg0);
	}

	/**
	 * @param arg0
	 * @throws IOException
	 * @see org.kxml2.kdom.Node#writeChildren(org.xmlpull.v1.XmlSerializer)
	 */
	public void writeChildren(XmlSerializer arg0) throws IOException {
		el.writeChildren(arg0);
	}	
}
