/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.configuration;

import java.io.File;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;

import at.ac.tuwien.vitalab.middleware.util.Configuration;

/**
 * CIMS configuration. 
 * 
 * @author Lukasz Juszczyk
 *
 */
public class CIMSConfiguration {

	private static Logger logger=
		LoggerFactory.getLogger(CIMSConfiguration.class);
	
	private static Configuration config=null;
	
	static {

		// escape.config was the old name of the variable
		// we still keep it for compatibility reasons
		
		String oldConfFileName=
			System.getProperty("escape.config","conf/escape.conf");
		
		File oldConfFile=new File(oldConfFileName);
		if (oldConfFile.exists()) {
			config=new Configuration(System.getProperty(
					"escape.config", "conf/escape.conf"));
		}
			
		if (config==null) {
			config=new Configuration(System.getProperty(
					"cosine.config", "conf/cosine.conf"));
		}
	}
	
	private static final String CONFIGNAME_SENSORSCLASSES="sensors";
	
	private static final String CONFIGNAME_RATINGCLASSES="ratings";
	
	private static final String CONFIGNAME_SUBSCRIPTIONMAXINACTIVITY=
		"subscription.max_inactivity";
	private static final Integer DEFAULT_SUBSCRIPTIONMAXINACTIVITY=
		new Integer(60);	//one minute 

	private static final String CONFIGNAME_NOTIFICATIONMAXWAIT=
		"notification.max_wait";
	private static final Integer DEFAULT_NOTIFICATIONMAXWAIT=
		new Integer(60);	//one minute 
	

	private static final String CONFIGNAME_NOTIFICATIONMAXDELAY=
		"notification.max_delay";
	private static final Integer DEFAULT_NOTIFICATIONMAXDELAY=
		new Integer(5);	//5 seconds 
	
	/**
	 * Resets all configuration to a new source.
	 * 
	 * @param confFile Configuration file name.
	 */
	public static void resetConfigurationFile(String confFile) {
		config=new Configuration(confFile);
	}
	
	/**
	 * Returns a list of class names of sensors.
	 * 
	 * @return List.
	 */
	public static List getSensorClasses() {
		String value=(String)config.getValue(
				CONFIGNAME_SENSORSCLASSES,
				String.class,
				"");
	
		StringTokenizer tokens=new StringTokenizer(value,",:");
		
		Vector result=new Vector();
		
		while (tokens.hasMoreTokens()) {
			result.add(tokens.nextToken());
		}
		
		return result;
	}

	/**
	 * Returns a list of class names of ratings.
	 * 
	 * @return List.
	 */
	public static List getRatingClasses() {
		String value=(String)config.getValue(
				CONFIGNAME_RATINGCLASSES,
				String.class,
				"");
	
		StringTokenizer tokens=new StringTokenizer(value,",:");
		
		Vector result=new Vector();
		
		while (tokens.hasMoreTokens()) {
			result.add(tokens.nextToken());
		}
		
		return result;
	}

	public static List getRatingClassesWithLimits() {

		String value=(String)config.getValue(
				CONFIGNAME_RATINGCLASSES,
				String.class,
				"");
		
	
		StringTokenizer tokens=new StringTokenizer(value,",:");
		
		Vector result=new Vector();
		
		while (tokens.hasMoreTokens()) {
			String token = tokens.nextToken();
			Vector class_limits = new Vector();
			StringTokenizer tokens2 = new StringTokenizer(token, "();");
			while (tokens2.hasMoreTokens())
				class_limits.add(tokens2.nextToken());
			result.add(class_limits);
		}
		
		return result;
	}

	/**
	 * Returns the max inactivity time of subscriptions.
	 * 
	 * @return Max inactivity in seconds.
	 */
	public static int getMaxSubscriptionInactivity() {
		int result=((Integer)config.getValue(
				CONFIGNAME_SUBSCRIPTIONMAXINACTIVITY,
				Integer.class,
				DEFAULT_SUBSCRIPTIONMAXINACTIVITY)).intValue();
		
		if (result<1 || result>60*60) {
			logger.warning("Configuration entry '"+
					CONFIGNAME_SUBSCRIPTIONMAXINACTIVITY+"' must "+
					"be between 1 and 3600. Assuming default value "+
					DEFAULT_SUBSCRIPTIONMAXINACTIVITY+".");
			
			config.saveCorrectedValue(CONFIGNAME_SUBSCRIPTIONMAXINACTIVITY,
					Integer.class,
					DEFAULT_SUBSCRIPTIONMAXINACTIVITY);
			
			return DEFAULT_SUBSCRIPTIONMAXINACTIVITY.intValue();
		}
		
		return result;
	}
	
	/**
	 * Returns the max waiting time of notification.
	 * 
	 * @return Max waiting time.
	 */
	public static int getMaxNotificationWait() {
		int result=((Integer)config.getValue(
				CONFIGNAME_NOTIFICATIONMAXWAIT,
				Integer.class,
				DEFAULT_NOTIFICATIONMAXWAIT)).intValue();
		
		if (result<1 || result>5*60) {
			logger.warning("Configuration entry '"+
					CONFIGNAME_NOTIFICATIONMAXWAIT+"' must "+
					"be between 1 and 300. Assuming default value "+
					DEFAULT_NOTIFICATIONMAXWAIT+".");
			
			config.saveCorrectedValue(CONFIGNAME_NOTIFICATIONMAXWAIT,
					Integer.class,
					DEFAULT_NOTIFICATIONMAXWAIT);
			
			return DEFAULT_NOTIFICATIONMAXWAIT.intValue();
		}
		
		return result;
	}
	
	/**
	 * Returns the max delay of notifications.
	 * 
	 * @return Max delay.
	 */
	public static int getMaxNotificationDelay() {
		int result=((Integer)config.getValue(
				CONFIGNAME_NOTIFICATIONMAXDELAY,
				Integer.class,
				DEFAULT_NOTIFICATIONMAXDELAY)).intValue();
		
		if (result<1 || result>1*60) {
			logger.warning("Configuration entry '"+
					CONFIGNAME_NOTIFICATIONMAXDELAY+"' must "+
					"be between 1 and 60. Assuming default value "+
					DEFAULT_NOTIFICATIONMAXDELAY+".");
			
			config.saveCorrectedValue(CONFIGNAME_NOTIFICATIONMAXDELAY,
					Integer.class,
					DEFAULT_NOTIFICATIONMAXDELAY);
			
			return DEFAULT_NOTIFICATIONMAXDELAY.intValue();
		}
		
		return result;
	}
}
