/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
/**
 *  ContextRoutingAttributeParser
 *  Extracts the attributes of a given xpath in a de.fzi.XPath
 */
package at.ac.tuwien.vitalab.cosine.contextrouting;

import java.util.LinkedHashMap;

import at.ac.tuwien.vitalab.cosine.ContextData;
import at.ac.tuwien.vitalab.cosine.Utils;
import de.fzi.XPath.AndExpr;
import de.fzi.XPath.Axis;
import de.fzi.XPath.EqualityExpr;
import de.fzi.XPath.Expr;
import de.fzi.XPath.FilterExpr;
import de.fzi.XPath.LocationPath;
import de.fzi.XPath.OrderedExpr;
import de.fzi.XPath.RelationalExpr;
import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;


/**
 * @author goldfish
 *
 */
public class ContextRoutingAttributeParser {
	public static final String NAMESPACEINDICATOR = "namespace-uri()";

	/**
	 * logger
	 */
	private static Logger logger=
		LoggerFactory.getLogger(ContextRoutingXPathParser.class);
	/**
	 * Object containing all parsed predicates
	 */


	private boolean unrecognizedAttributes; //fail save 
	private LinkedHashMap attributeList;
	private String nameSpace;

	/**
	 * Carefull! There is a difference between unrecognizedAttributes and
	 * skippedAttributes
	 */
	private String skippedAttributes = "";
	
	public ContextRoutingAttributeParser() {
		reset();
	}
	private void reset() {
		unrecognizedAttributes = true; //fail save 
		attributeList = new LinkedHashMap(3);
		nameSpace = null;		
	}
	
	/**
	 * fill the predicateList with all predicates found at this node 
	 * @param node the node the predicates are gathered from	 
	 */
	public boolean getNodeAttributes(ContextRoutingXPathNode node) {
		// String predicateString = ""; //e.g. attribute::name = ''B-Team''

		Expr[] predicates = node.getAttribute();

		String nodeName = node.getName();
		
		if (predicates.length == 0) {
			reset();
			return false;			
		}

		
		Expr expr = predicates[0];
		
		//will parse the predicates to compare
		//the fact that expr is checked first leads to 
		//the assumption that null is returned at some point
		while (expr != null) {
			expr = checkExpr(expr, nodeName);
		}
		return true;
	}	
	
	/**
	 * break expr down until a equality expression is found.
	 * is called recursively 
	 * @param expr expression to break
	 * @param nodeName attatched to expr
	 * @return null or next piece of expr
	 */
	private Expr checkExpr(Expr expr, String nodeName) {
		if (expr instanceof AndExpr)
			expr = handleAndExpr((AndExpr)expr, nodeName);
		else if (expr instanceof LocationPath) {
			expr = handleLocationPath((LocationPath) expr, nodeName);
		}
		else if (expr instanceof EqualityExpr) {
			expr = handleEqualityExpr((EqualityExpr) expr, nodeName);			
		}
		//still unrecognized -> skipped
		else if (expr instanceof RelationalExpr) {
			expr = handleSkippedExpr((RelationalExpr)expr, nodeName);
			return null;
		}
		else {
			//logger.info("NOT all expr recognized "+expr.getClass().getName());
			//predicateList.clear();
			reset();
			expr = null;

		}
		return expr;
	}
	/**
	 * lowest point of expr. Here expr is finally turned into a matching string
	 * @param andExpr the equality expression
	 * @param nodeName node attatched to expr
	 * @return null. End of breakdown
	 */
	private Expr handleEqualityExpr(EqualityExpr andExpr, String nodeName) {
		Expr ex = (EqualityExpr) andExpr;
		//two possible children
		LocationPath locPath;
		FilterExpr filterExpr;
		String var = "", val = "";

		//is a not equal expr
		if (ex.toString().indexOf("!=") > 0)  {
		
			return handleSkippedExpr((EqualityExpr)ex, nodeName); 
		}
		
		//andExpr.getExprsCount() MUST be 2
		if (andExpr.getExprsCount() != 2) {
			this.unrecognizedAttributes = true;
			return null;
		}

		//... get var
		ex = (Expr)andExpr.getExpr(0);
		if (ex instanceof LocationPath) {
			locPath = (LocationPath) ex;
			for (int i=0; i < locPath.getStepsCount(); i++)
			{
				//idea here: child::Device/child::Battery/attribute::percentLeft
				if (locPath.getStep(i).toString().startsWith("attribute::"))
					var = locPath.getStep(i).getNodeTest().toString();
				else if (locPath.getStep(i).toString().startsWith("child::"))
					//check contextData tag
					if (!ContextData.IDNAME.equals(locPath.getStep(i).getNodeTest().toString())) {
						logger.info("expected "+ContextData.IDNAME+" received "+locPath.getStep(i).getNodeTest().toString());
//						this.unrecognizedAttributes = true;
						return handleSkippedExpr(andExpr, nodeName);
					}
				
			}
		}
		else if (ex instanceof FilterExpr) {
			filterExpr = (FilterExpr) ex;
			var = filterExpr.toString();
		}
		//...and val
		ex = (Expr)andExpr.getExpr(1);
		if (ex instanceof LocationPath) {
			locPath = (LocationPath) ex;
			val = locPath.getStep(0).getNodeTest().toString();
		}
		else if (ex instanceof FilterExpr) {
			filterExpr = (FilterExpr) ex;
			val = filterExpr.toString();
		}

		String type = "numerical";
		if (val.startsWith("'"))
			type = "literal";
		
		val = val.replace('"', ' ');
		val = val.replace('\'', ' ');
		
		if (var.equals(NAMESPACEINDICATOR))
			this.nameSpace = val.trim();
		else {
			String[] predicate = createSearchString(nodeName, var.trim(), val.trim(), type);
			this.attributeList.put(var, predicate);
		}
		this.unrecognizedAttributes = false;
		return null;
	}
	
	/** Collect the skipped attributes
	 * @param skippedExpr the expression to extract
	 * @param nodeName node the expression belongs to 
	 * @return null 
	 */
	private Expr handleSkippedExpr(OrderedExpr skippedExpr, String nodeName) {
		LocationPath locPath;
		FilterExpr filterExpr;
		String var = "", val = "";
		String operator = "";
		String child ="";

		if (skippedExpr.getExprsCount() != 2) {			
			this.unrecognizedAttributes = true;
			return null;
		}

		Expr ex = (Expr)skippedExpr.getExpr(0);


		if (ex instanceof LocationPath) {
			locPath = (LocationPath) ex;

			//we are not interested in the actual var but the original version of the statement

			var = parserAttributeToString(locPath);

		}
		else if (ex instanceof FilterExpr) {
			filterExpr = (FilterExpr) ex;
			var = filterExpr.toString();
		}
		//...and val
		ex = (Expr)skippedExpr.getExpr(1);
		operator = ex.getOperatorAsString();

		if (ex instanceof LocationPath) {
			locPath = (LocationPath) ex;
			val = locPath.getStep(0).getNodeTest().toString();
		}
		else if (ex instanceof FilterExpr) {
			filterExpr = (FilterExpr) ex;
			val = filterExpr.toString();
		}
		String type = "numerical";
		if (val.startsWith("'")) {
			type = "literal";
		}

		val = val.replace('"', ' ');
		val = val.replace('\'', ' ');
		
		addSkippedAttributes(var.trim(), operator, val.trim(), child, type);
		
		this.unrecognizedAttributes = false;
		return null;
	}
	
	/** parse location part of an xpath
	 * @param locExpr the expression 
	 * @param nodeName the node it belongs to
	 * @return an expression
	 */
	private Expr handleLocationPath(LocationPath locExpr, String nodeName) {
		Expr ex = (Expr) locExpr;
		int i=0;
		int j=0;
		
		for (i=0; i<locExpr.getStepsCount(); i++) {
			for (j=0; j<locExpr.getStep(i).getPredicatesCount(); j++) {
				ex = checkExpr((Expr)locExpr.getStep(i).getPredicate(j), nodeName);
			}
		}
		
		return ex;
	}	
	/** parse an and expression
	 * @param expr the expression
	 * @param nodeName the node type
	 * @return parsed expression
	 */
	private Expr handleAndExpr(AndExpr expr, String nodeName) {
		Expr ex = (Expr) expr;
		for (int i=0; i < expr.getExprsCount(); i++) {
			ex = (Expr) checkExpr(expr.getExpr(i), nodeName);		
			if (this.unrecognizedAttributes) //one predicate was not recognized
				break;
		}			
		return ex;
	}


	public boolean hasUnrecognizedAttributes() {
		return unrecognizedAttributes;
	}
	
	public LinkedHashMap getAttributeList() {
		return attributeList;
	}

	public String getNameSpace() {
		return nameSpace;
	}
	/**	concatenates the predicate string to a unique search key
	 * @param nodeType
	 * @param var
	 * @param val
	 * @param type literal or numeric for correct reconstruction (with '' or without)
	 * @return array presentation of all four
	 */
	public static String[] createSearchString(String nodeType, String var, String val, String type) {
		return new String[] {nodeType, var, val, type};
	}	
	public static final short NODE = 1;
	public static final short VAR = 1;
	public static final short VAL = 1;
	public static final short TYPE = 1;
	
	/** capture the skipped attributes in one string @see skippedAttributes
	 * @param var
	 * @param operator
	 * @param val
	 * @param child
	 * @param type
	 */
	public void addSkippedAttributes(String var, String operator, String val, String child, String type) {
		
		if (type.equals("literal"))
			val = "'"+val+"'";
		
		String addAttribute = var+operator+val;
		if (!child.equals(""))
			addAttribute = child+"/@"+addAttribute;
		if (this.skippedAttributes.equals("")) {
			this.skippedAttributes += addAttribute;			
		}
		else {
			this.skippedAttributes += " and "+addAttribute;
		}

	}
	
	public String getSkippedAttributes() {
		return skippedAttributes;
		
	}

	/** de.fzi.XPath does not provide Expr back to xpath so 
	 * here is our own
	 * @param exprs to convert to string
	 * @return string conversion
	 */
	public static String ExprToString(Expr[] exprs) {
		String result = "";
		for (int i=0; i<exprs.length; i++) {
			result += ContextRoutingAttributeParser.parserAttributeToString(exprs[i]); 
		}
		return result;
	}
	
	/** the xpath provided by Expr.toString need further adjustment
	 * @param attribute
	 * @return
	 */
	public static String parserAttributeToString(Expr attribute) {
		
		String the_ugly = "";
		
		
		the_ugly = attribute.toString();
		the_ugly = Utils.replace_string(the_ugly,Axis.getAxisName(Axis.CHILD)+"::","");
		the_ugly = Utils.replace_string(the_ugly,Axis.getAxisName(Axis.DESCENDANT_OR_SELF)+"::","");
		the_ugly = Utils.replace_string(the_ugly,Axis.getAxisName(Axis.ATTRIBUTE)+"::","@");
		the_ugly = Utils.replace_string(the_ugly,"node()","");
		the_ugly = Utils.replace_string(the_ugly,"''","'");					
		the_ugly = Utils.replace_string(the_ugly," = ","=");
		the_ugly = Utils.replace_string(the_ugly," > ",">");
		the_ugly = Utils.replace_string(the_ugly," < ","<");
		the_ugly = Utils.replace_string(the_ugly," <= ","<=");
		the_ugly = Utils.replace_string(the_ugly," >= ",">=");		
		the_ugly = Utils.replace_string(the_ugly,"AND","and");
		the_ugly = Utils.replace_string(the_ugly,"OR","or");

		return the_ugly;
	}


}
