/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
/**
 *  ContextRoutingHandler
 * Just a hidden helper/handler class for the {@link ContextRoutingService}
 * main entry point {@link answerQuery}
 * steps down the approved xpath that was converted into structure:
 * 
 * ContextRoutingXPathNode with main features
 * - name
 * - relations to prev and next
 * - adapted xpath for node
 * - result array for node
 * 
 * calls on any occasion {@link resultIsInLeaves}
 * with 
 *  - loop through all nodes (until next==null)
 *  - path reconstruction  (ContextRoutingXPathNode and XPath)
 *  - call to getMatchingFromLevel
 *  - avoidance of a null result
 *  
 * getMatchingFromLevel focuses on currend node and does
 *  - predicate parsing
 *  - routing table lookup
 *  - intersections
 *  - rating
 *  
 * @author hpsaier
 */
package at.ac.tuwien.vitalab.cosine.contextrouting;

import java.io.File;
import java.net.MalformedURLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import org.ini4j.Ini;
import org.ini4j.Ini.Section;

import at.ac.tuwien.vitalab.cosine.ContextData;
import at.ac.tuwien.vitalab.cosine.Utils;
import at.ac.tuwien.vitalab.cosine.XMLData;
import at.ac.tuwien.vitalab.cosine.rating.ARating;
import at.ac.tuwien.vitalab.cosine.webservice.ASensorWebService;
import at.ac.tuwien.vitalab.middleware.discovery.MissingServiceAttributeException;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;

/**
 * @author goldfish
 *
 */
public class ContextRoutingHandler {
	
	/**
	 * logger
	 */
	private static Logger logger=
		LoggerFactory.getLogger(ContextRoutingHandler.class);
	/**
	 * init file defining the sensor properties
	 */
	private final String iniContextFileName = "conf/contextmodel.ecs";
	/**
	 * contains all data of the contextmodel
	 */
	private Ini iniContextFile = null;

	/**
	 * xpath statement that matches result exactly
	 */
	private String resultingXPath = "";

	/**
	 * holds the node the decission went to
	 */
	private ContextRoutingXPathNode resultNode = null;
	private double currentBestRating = 0;
	
	/**
	 * flags to switch on/off new/old features
	 */
	private final boolean NEW_XPATH_DEBUG = false;	
	private final boolean REQUIRED_CONTEXT_ON = true;
	private final boolean UNDEFINED_PATH_ON = true;	
	
	
	ContextRoutingHandler() {
	}

	/**
	 * prepares the result set to a previously parsed xpath represented by {@link ContextRoutingXPathParser}}
	 * @param parser prepared parser holding a valid xpath statement
	 * @return array of matching ServiceInfo
	 */
	ServiceInfo [] answerQuery(ContextRoutingXPathParser parser) {

		ServiceInfo [] result;
		
		//fail-save return
		if (parser == null) {
			logger.error("parser-tree was 'null'");
			return new ServiceInfo[0];			
		}
	
		//get top node
		ContextRoutingXPathNode nextNode = parser.getFirstNode();

		//fail-save return
		if (nextNode == null) {
			logger.error("parser.getFirstNode was 'null'");
			return new ServiceInfo[0];			
		}

		resetRating();
		

		resultIsInLeaves(nextNode, parser);

		if (resultNode != null)
			nextNode = resultNode;
		else //must never happen!!!
			logger.error("resultNode was null!!! and nextNode: "+nextNode);
		//nextNode is the node returned after considering rating
		result = nextNode.getNodeResult();
		setResultingXPath(nextNode.getXpath());
		
		resetRating();
		
		if (Logger.isDebugAllowed() || NEW_XPATH_DEBUG) {
			String message = "\nresulting xpath: "+getResultingXPath();			
			message += "\nresulting services:\n";
			for (int i=0; i < result.length; i++) {
				try {
					message += "\t"+result[i].getId()+" "+result[i].getHost()+"\n";
				} catch (MissingServiceAttributeException e) {
				} catch (MalformedURLException e) {
				}
			}
			logger.info(message);
		}

		//reset result path if no result
		if (result.length == 0)
			setResultingXPath("");
		return result;
	}

	/**
	 * for each node along the xpath this method is called with appropriate node
	 * and using lastResult a new result is merged by searching the node level
	 * but also taking into account the belongsTo property
	 * @param node the node data is searched for
	 * @param lastResult hash with the service info found at last invocation
	 * @return Hashtable with updated result set
	 */
	public ResultHolder getMatchingFromLevel(ContextRoutingXPathNode node,
			Vector lastResultUuids) {
		//overall result
		ResultHolder result = new ResultHolder();
		//intermediate/intersection result
		HashSet	interResult = new HashSet();

		//rating of the services of this node excluding offline ones
		double nodeRating = 0;
		
		//all ContextRoutingServiceInfo of this node
		HashSet hashSet = null;
		
		//gather attributes of this node including eventual namespace
		ContextRoutingAttributeParser attributeParser = new ContextRoutingAttributeParser();
		attributeParser.getNodeAttributes(node);

		boolean isFirst = true;
		// 1st get all services connected to recognized attributes 
		if (!attributeParser.hasUnrecognizedAttributes() && 
				attributeParser.getAttributeList().size() > 0) {
			
			//get all matching to result and predicate
			for (Iterator it = attributeParser.getAttributeList().values().iterator(); it.hasNext();) {
				String node_var_val[] = (String[])it.next();

				//look-up to context map
				//TODO 0,1,2 make readable				
				hashSet = findContext(node_var_val[0], node_var_val[1], node_var_val[2]);
				
				if (hashSet == null) {
/***************************************** NEW ***********************************/					
					if (REQUIRED_CONTEXT_ON) {
						//1) TODO is a required contextID
						HashSet hashSetNoUse = findContext(node_var_val[0], node_var_val[1], "");

						if (hashSetNoUse == null) {
							attributeParser.addSkippedAttributes(node_var_val[1], "=", node_var_val[2], ContextData.IDNAME, node_var_val[3]);
							hashSet = null;
							continue;						
						} else {
							attributeParser.addSkippedAttributes(node_var_val[1], "=", node_var_val[2], ContextData.IDNAME, node_var_val[3]);
							//get all possible candidates
							hashSet = findContext(node_var_val[0],"", "");
							hashSet.removeAll(hashSetNoUse);
						}
/***************************************** NEW ***********************************/						
					} else {
						//if no result, in a SI-collection case the presented attributes could be hidden from the router
						//therefore the unrecognized ones are collected
						
						attributeParser.addSkippedAttributes(node_var_val[1], "=", node_var_val[2], ContextData.IDNAME, node_var_val[3]);
						hashSet = null;
						continue;
					}						
				}

				//attributes contradicting
				if (!isFirst && result.isEmpty()) {
					break;
				}
				else {
					result.clear();
					isFirst = false;
				}

				for (Iterator it2 = hashSet.iterator(); it2.hasNext(); )
				{
					boolean intersectsWithParent = false;
					boolean intersectsWithLast = false;
					
					ContextRoutingServiceInfo info = (ContextRoutingServiceInfo)it2.next();
					if (lastResultUuids.size() == 0)  {//the top node
						intersectsWithParent = true;
					}
					else if (lastResultUuids.contains(info.getBelongsTo())) {
						intersectsWithParent = true;
					}

					if (interResult.isEmpty())
						intersectsWithLast = true;
					else if (interResult.contains(info.getUuid()))
						intersectsWithLast = true;
					if (intersectsWithParent && intersectsWithLast &&
							checkNamespace(info, attributeParser.getNameSpace())) {
						if (info.isOffline()) {
							logger.debug("OFFLINE: "+info.getNodeType()+" "+info.getUuid());
							result.addResult(info.getUuid(), null);
						} else {
							nodeRating += getRating(info);
							result.addResult(info.getUuid(), info.getServiceInfo());
						}

					} 
				}	
				//interResult is for intersection
				interResult.clear();
				interResult.addAll(result.getUuids());
			}
//			if (hashSet == null)
//				return result;			
		}
		//2nd get all from node
		else if (lastResultUuids.isEmpty())
		{		
			hashSet = (HashSet) findContext(node.getName(), "", "");
			if (hashSet == null)
				return result;

			for (Iterator it = hashSet.iterator(); it.hasNext();) {
				ContextRoutingServiceInfo info = (ContextRoutingServiceInfo) it.next();
				
				if (!checkNamespace(info, attributeParser.getNameSpace()))
					continue;

				if (info.isOffline()) {
					logger.debug("OFFLINE: "+info.getNodeType()+" "+info.getUuid());
					result.addResult(info.getUuid(), null);
				} else {
					nodeRating += getRating(info);
					result.addResult(info.getUuid(), info.getServiceInfo());
				}

			}
		}
		//3rd result depends on previous result, however, NO attributes found
		else {
			hashSet = (HashSet) findContext(node.getName(), "", "");
			if (hashSet == null)
				return result;
			
			for (Iterator it = hashSet.iterator(); it.hasNext();) {
				ContextRoutingServiceInfo info = (ContextRoutingServiceInfo) it.next();
				if (lastResultUuids.contains(info.getBelongsTo()) &&
						checkNamespace(info, attributeParser.getNameSpace())) {
					if (info.isOffline()) {
						logger.debug("OFFLINE: "+info.getNodeType()+" "+info.getUuid());
						result.addResult(info.getUuid(), null);						
					} else {
						nodeRating += getRating(info);
						result.addResult(info.getUuid(), info.getServiceInfo());						
					}
				}

			}
		}
		
		if (result.isEmpty()) {
			return result;
		}

		
		//final rating
		int nrOfServices = result.getResult().length;// NEVER 0 see isEmpty above!!!		
		nodeRating = (nodeRating/nrOfServices);
			
		if (currentBestRating == 0 || nodeRating > currentBestRating) {
			currentBestRating = nodeRating;
			resultNode = node;
		}
		if (attributeParser.hasUnrecognizedAttributes() && node.getAttribute().length > 0) {
			result.setUnrecognizedAttr(true);
		} else if (attributeParser.getSkippedAttributes().length() > 0) {
			result.setSkippedAttr(attributeParser.getSkippedAttributes());
		}

		return result;
	}

	/**
	 * reset rating attributes
	 */
	private void resetRating() {
		this.currentBestRating = 0;
		this.resultNode = null;
	}
	
	/**
	 * return rating for a specific ServiceInfo. Considers also number of ratings applyed
	 * @param info is going to be rated
	 * @return estimated rate
	 */
	private double getRating(ContextRoutingServiceInfo info) {
		double result = 0;
		
		
		ServiceInfo si = info.getServiceInfo();
		
		if (si == null) //failsave but should never happen
			return result;
		
		result = ContextRoutingService.rating.getRating(si);
		int nrOfRatings =  ContextRoutingService.rating.getNumberOfRatings(si);
		result = (nrOfRatings == 0 ? result : result/nrOfRatings);

		return result;
	}
	

	/**
	 * common way of finding result
	 * @return Hashtable as result
	 */
	public void resultIsInLeaves(ContextRoutingXPathNode nextNode, 
													ContextRoutingXPathParser parser) {
		ResultHolder resultHolder = new ResultHolder();

		//don't do this!!!
 		if (nextNode == null) {
 			return;
 		}
 		
		String newXPath = parser.getXPath();
		do {
			//1st resolve a undefined path problem e.g. //Device or //Team//Battery
			if (nextNode.getRealtionToNext().equals("//")
//					&& !isTopNode(nextNode.getName())
					&& (nextNode.getAttribute().length == 0)
					&& UNDEFINED_PATH_ON) {
				
				ContextRoutingPathReconstructor crpr = new ContextRoutingPathReconstructor();
				crpr.setRelations((HashMap)ContextRoutingService.routeToTop);
				//reconstruct path as good as it gets
				LinkedList path = crpr.recreatePath(nextNode.getDownToNext().getName(),nextNode.getName());

				//create dummy nodes
				if (!path.isEmpty()) {
					nextNode = addMissingNodes(nextNode, path);
					newXPath = createMissingXPath(nextNode);
				} else {
					nextNode.setDownToNext(null);
				}					
			}
			else if	(nextNode.getRealtionToNext().equals("//")) {
				nextNode.setDownToNext(null);
			}

			resultHolder = getMatchingFromLevel(nextNode, resultHolder.getUuids());

			nextNode.setNodeResult(resultHolder.getResult());
			nextNode.setUnrecognizedAttributes(resultHolder.getSkippedAttr());

		
			//get xpath for nextNode
			if (!resultHolder.hasUnrecognizedAttr())
				newXPath = adaptXPath(newXPath, nextNode);
			//save current result
			nextNode.setXpath(newXPath);			

			//unrecognized attributes makes no sense to progress
			if (resultHolder.hasUnrecognizedAttr() || !resultHolder.getSkippedAttr().equals(""))
				break;

			//there was no result for the current node -> return result of previous and stop looking
			if (resultHolder.isEmpty())
			{
				//get result form last node
				if (nextNode.getUpToNext() != null)
				{
					nextNode = nextNode.getUpToNext();
				} 
				break; //...and return
			}
			if (nextNode.getDownToNext() != null) //consider next node
				nextNode = nextNode.getDownToNext();
			else //last node reached
				break;
			
		} while (true);

		//fail save
		if(resultNode == null)
			resultNode = nextNode;

	}
	
	/**
	 * Namespace check
	 * @param info ServiceInfo to check
	 * @param nameSpaceUrl the given namespace
	 * @return given namespace and configured namespace match
	 */
	private boolean checkNamespace(ContextRoutingServiceInfo info, String nameSpaceUrl) {
		if (info != null &&
			nameSpaceUrl != null &&	
		!info.getServiceInfo().getAttribute(ASensorWebService.WSATTR_NAME_CONTEXTNS).equals(nameSpaceUrl)) {
			return false;
			
		}
		return true;
	}

	/**adapt result path for current node
	 * @param xpath current xpath
	 * @param node to adapt to
	 * @return adapted XPath
	 */
	private String adaptXPath (String xpath, ContextRoutingXPathNode node) {
		if (node == null)
			return "";

		if (NEW_XPATH_DEBUG)
			System.err.println("ORG: "+node.getName()+" --- "+xpath);		
		
		int start = xpath.indexOf(node.getName()+node.getRealtionToNext());
		int length = (node.getName()+node.getRealtionToNext()).length();
		xpath = xpath.substring(start+length);
		
		if (NEW_XPATH_DEBUG)
			System.err.println("FIRST: "+xpath);		
		if (!node.getRealtionToNext().equals("")) {
			
			if (node.getAttribute().length > 0) {
				xpath = xpath.substring(xpath.indexOf("]"+node.getRealtionToNext())+("]"+node.getRealtionToNext()).length());
			}
			
			if (!node.getUnrecognizedAttributes().equals("")) {
				xpath ="/"+node.getName()+"["+node.getUnrecognizedAttributes()+"]"+node.getRealtionToNext()+xpath;
			}
			else  {
				xpath = "/"+node.getName()+node.getRealtionToNext()+xpath;
				
			}
		} else if (!node.getUnrecognizedAttributes().equals("")){ //last node
			if (NEW_XPATH_DEBUG)
				System.err.println("there was some unrecoginzed/skipped attributes: "+node.getUnrecognizedAttributes());
			xpath ="/"+node.getName()+"["+node.getUnrecognizedAttributes()+"]";			
		} else {
			xpath = "/*";
		}
		if (NEW_XPATH_DEBUG)		
			System.err.println("XPATH: "+xpath);
		return xpath;
	}
	/**
	 * @param node
	 *            to check
	 * @return
	 */
	private boolean isTopNode(String node) {
		Vector vector = (Vector) ContextRoutingService.routeToTop.get(node);
		if (vector != null)
			return (vector.size() == 1 && vector.get(0).equals(""));
		else
			return false;
	}

	/**
	 * add a new {@link ServiceInfo} to the search maps
	 * 
	 * @param aServiceInfo
	 */
	void addContext(ServiceInfo aServiceInfo) {
		try {
			//get data from aServiceInfo 
			String node = (String) aServiceInfo.getId();
			String context = aServiceInfo.getAttribute(ContextData.IDNAME);
			XMLData cd = new XMLData(context);

			// create extended ContextInfo
			ContextRoutingServiceInfo info = ContextRoutingServiceInfo.getInstance(cd.getAttributeByName("uuid"));
			String belongsTo = cd.getAttributeByName("belongsTo");
			info.setBelongsTo((belongsTo == null ? "" : belongsTo));			
//			info.setUuid(cd.getAttributeByName("uuid"));
			info.setServiceInfo(aServiceInfo);
			info.setNodeType(node);
			
			String var, val;
			
			// ...and loop through elements 			
			Hashtable contextTable = cd.getAttributes();			
			for (Iterator it=contextTable.keySet().iterator(); it.hasNext(); ) {
				var = (String) it.next();
				val = (String) contextTable.get(var);
//				val = MobileDeviceInitService.getInstance().returnType(var, aServiceInfo.getId(), aServiceInfo.getAttribute(ASensorWebService.WSATTR_NAME_CONTEXTNS), val);
//				System.err.println(var+"="+val);
				insertContext(node, var, "", info);				
				insertContext(node, var, val, info);				
			}
			//... node itself
			insertContext(node, "", "", info);
//			debug_PrintRoutingTable();

			} catch (Exception e) {
				logger.error(e.getMessage());
			}
		}

	
	/** routing table lookup
	 * @param node 
	 * @param var
	 * @param val
	 * @return HashSet of references (ServiceInfos)
	 */
	private HashSet findContext(String node, String var, String val) {
		
		HashMap attributeMap = (HashMap) ContextRoutingService.contextSearch.get(node);
		if (attributeMap == null) {
			return null;
		}
		HashMap valueMap = (HashMap) attributeMap.get(var);
		if (valueMap == null) {
			return null;
		}		
		HashSet references = (HashSet) valueMap.get(val);
		if (references == null) {
			return null;
		}
		return references;
	}

	/** routing table insert
	 * @param node
	 * @param var
	 * @param val
	 * @param info the ServiceInfo extension ContextRoutingServiceInfo
	 */
	private void insertContext(String node, String var, String val, ContextRoutingServiceInfo info) {
		HashMap attributeMap = (HashMap) ContextRoutingService.contextSearch.get(node);
		if (attributeMap == null) {
			attributeMap = new HashMap();
		}
		HashMap valueMap = (HashMap) attributeMap.get(var);
		if (valueMap == null) {
			valueMap = new HashMap();
		}		
		HashSet references = (HashSet) valueMap.get(val);
		if (references == null) {
			references = new HashSet();
		}
		references.add(info);							 
		
		valueMap.put(val, references);
		
		attributeMap.put(var, valueMap);
		
		ContextRoutingService.contextSearch.put(node, attributeMap);
	}
	/**
	 * @deprecated routing table entries are NEVER deleted.
	 * @param aServiceInfo
	 *            the old ServiceInfo           
	 */
	void removeContext(ServiceInfo aServiceInfo) {
		//TODO decide what further to do!!!
	}

	/** update the ration of a certain service
	 * @param aServiceInfo the service
	 * @param type type of ratio
	 * @param value ratio value
	 */
	void updateRatio(ServiceInfo aServiceInfo, ARating type, double value) {
		ContextRoutingService.rating.setRating(type, aServiceInfo, value);		
	}

	/** creation of context tree from ecs specification file
	 * 
	 */
	boolean createDependencyTree() throws Exception {
		boolean result = true;
		Vector parentVector;
		Iterator it = iniContextFile.keySet().iterator();


		while (it.hasNext()) {
			Section section = (Section) iniContextFile.get(it.next());
			String parents[] = split2((String) section.get("parents"), ",");
			parentVector = new Vector();
			for (int i = 0; i < parents.length; i++) {
				//parse parent part
				parentVector.add(split2(parents[i], "@")[0].trim());
			}
			ContextRoutingService.routeToTop.put(
					split2(section.getName(), "@")[0], parentVector);
		}
		//check if for all nodes routeToTop leads to an end
		for (it = ContextRoutingService.routeToTop.keySet().iterator(); it.hasNext();){
			String childNode = (String) it.next();			
			parentVector = (Vector) ContextRoutingService.routeToTop.get(childNode);
			boolean goodToGo = true;
			for (Iterator it2 = parentVector.iterator(); it2.hasNext();)
				if (ContextRoutingService.routeToTop.get(it2.next()) == null && !isTopNode(childNode)) {					
					goodToGo = false;
					break;
				}
 			if (!goodToGo) {
 				//logger.error("Node :"+childNode+" not all parents defined in context model. Please check...");
 				throw new Exception ("INI-ERROR: For node '"+childNode+"': not all parents defined in context model. Please check...");
			}				
				
		}
		return result;
	}
	
	/**
	 * xpath is syntactically correct, so check if xpath structure
	 * matches the systems context tree
	 * @param node
	 *            the lowest node in the xpath
	 * @return relation ship ok?
	 */
	boolean checkNodeRelationShip(ContextRoutingXPathNode node) 
		throws Exception{
		ContextRoutingXPathNode nextNode = null;

		if (node == null)
			return false;

		while (node != null) {
			nextNode = node.getUpToNext();
						
			if (nextNode == null)
				if (
					 (node.getRealtionToPrev().equals("//")
						|| isTopNode(node.getName()))
					 && ContextRoutingService.routeToTop.containsKey(node.getName())
				    )
				     {

					return true; // reached the top
				} else {	
					return false;
				}

			if (findRouteToNode(node.getName(), nextNode.getName(), node.getRealtionToNext().equals("/")).equals("ok")) {
				node = node.getUpToNext(); // step one up
				//ContextRoutingService.nodeList.add(node.getName());
			} else {
				throw new Exception("PATH-ERROR: NO path from '"+node.getName()+"' to '"+nextNode.getName()+"' with relation "+nextNode.getRealtionToNext());
			}
		}
		return true; // all checks correct
	}
	
	
	/**
	 * recursive route finder
	 * 
	 * @param fromNode
	 *            start route
	 * @param toNode
	 *            end route
	 * @param direct
	 *            connection between previous and next node "//" or "/"
	 * @return error, ok, no_value
	 */
	private String findRouteToNode(String fromNode, String toNode,
			boolean direct) {
		String node = fromNode;
		String result = "no_value";

		if (ContextRoutingXPathParser.isAnyNode(node))
			return "ok";
			
		// end search condition
		if (!result.equals("no_value"))
			return result;
		// get actual level
		Vector next = (Vector) ContextRoutingService.routeToTop.get(node);

		// top of tree reached without success
		if (next == null)
			return "error";

		// check level nodes
		for (int i = 0; i < next.size(); i++) {
			if (result.equals("ok"))
				break;
			
			node = (String) next.get(i);
			if (node == null) {
				return "error";
			} else if (node.equals(toNode)) {
				return "ok";
			} else if (direct && !node.equals(toNode) && (i == next.size() - 1)) {
				return "error";
			} else {
				result = findRouteToNode(node, toNode, direct);
				// if (!result.equals("no_value")) // we got a decision already
				// break;
			}
		}
		return result;
	}

	/**

	 * lists and checks all prerequisite for successful startup
	 */
	void checkPrerequisite() throws Exception {
		File file;

		file = new File(iniContextFileName);
		if (!file.exists()) {
			throw new Exception("INI-ERROR: your context init file does net exist/cannot be found in: "
					+ file.getAbsolutePath());
			//logger.error("INI-ERROR: your context init file does net exist/cannot be found in: "
			//		+ file.getAbsolutePath());
		}
	}
	
	/**
	 * part of path reconstruction 
	 * adds to the ContextRoutingXPathNode chain the missing nodes
	 * @param father top node of reconstruction
	 * @param nodes missing nodes in a list
	 * @return new father
	 */
	private ContextRoutingXPathNode addMissingNodes(ContextRoutingXPathNode father, LinkedList nodes) {
		String fatherNode = father.getName();
		ContextRoutingXPathNode child = father.getDownToNext();
		ContextRoutingXPathNode pNode = null;
		String nextNode = "";
		for (int i=1; i<nodes.size(); i++) {
			nextNode = (String) nodes.get(i);
			if (NEW_XPATH_DEBUG)
				System.err.println("adding Node for missing: "+child.getName()+" "+nextNode);
			
			if (fatherNode.equals(nextNode))
			{
				ContextRoutingXPathNode top = child.getUpToNext();
				top.setDownToNext(child);
				top.setRealtionToNext("/");
				child.setRealtionToPrev("/");
				child.setUpToNext(top);
				pNode = top;
				break;
			}
	
			pNode = new ContextRoutingXPathNode();
			
			//classical features ;)
			pNode.setName(nextNode);
			pNode.setNamespace("");
			pNode.setNodeResult(null);
			pNode.setXpath("");
			pNode.setAttribute(null);
			pNode.setNodeResult(null);
			pNode.setUnrecognizedAttributes("");

			//important relations
			pNode.setDownToNext(child);
			pNode.setRealtionToNext("/");
			//set alibi info
			pNode.setUpToNext(child.getUpToNext());
			pNode.setRealtionToPrev(child.getRealtionToPrev());

			//connect
			child.setUpToNext(pNode);
			child.setRealtionToPrev("/");
			
			//step up
			child = pNode;
		}
		//not a complete reconstruction
		if (!fatherNode.equals(nextNode)) {
			ContextRoutingXPathNode top = child.getUpToNext();
			top.setDownToNext(child);
			top.setRealtionToNext("//");
			child.setRealtionToPrev("//");
			child.setUpToNext(top);			
			pNode = top;
		}
		
		return pNode;			
	}
	/** part of path reconstruction
	 * in order to process the statment correctly also the given
	 * xpath must be adjusted
	 * @param node father node
	 * @return new xpath
	 */
	private String createMissingXPath(ContextRoutingXPathNode node) {
		ContextRoutingXPathNode loopNode = node;
		String xpath = loopNode.getRealtionToPrev();

		while (loopNode != null) {
			xpath+=loopNode.getName()+loopNode.getRealtionToNext();
			if (loopNode.getAttribute().length > 0) {
				xpath+="["+ContextRoutingAttributeParser.ExprToString(loopNode.getAttribute())+"]";
			}
			loopNode = loopNode.getDownToNext();
		}
		if (NEW_XPATH_DEBUG)
			System.err.println("recreated xpath for node: "+node.getName()+" -> "+xpath);

		return xpath;
	}

	/**
	 * wrapper
	 */
	void readIniFiles() throws Exception{
		iniContextFile = Utils.readIniFiles(iniContextFileName);
	}

	/**
	 * wrapper
	 */
	public static String[] split2(String toSplit, String pattern) {
		return Utils.split2(toSplit, pattern);
	}

	public String getResultingXPath() {
		return resultingXPath;
	}

	public void setResultingXPath(String resultingXPath) {
		if (resultingXPath == null)
			resultingXPath = "";
		
		if (!resultingXPath.startsWith("/") && resultingXPath.length() > 0)
			resultingXPath = "/" + resultingXPath;
		
		this.resultingXPath = resultingXPath;
	}
	
	public String stripLeadingSlash(String xpath) {
		if (xpath.startsWith("//"))
			return xpath.substring(1, xpath.length());
		else
			return xpath;
	}
	
	/**
	 * internal class holding the current valid result
	 */
	private class ResultHolder {
		private Vector uuids = new Vector();
		private Vector serviceInfos = new Vector();
		private String skippedAttr = "";
		private boolean unrecognizedAttr = false;
		
		public void setSkippedAttr(String attr) {
			skippedAttr = attr;
		}
		public String getSkippedAttr() {
			return skippedAttr;
		}
		public boolean isEmpty() {
			return serviceInfos.isEmpty();
		}
		
		public ServiceInfo[] getResult() {
			ServiceInfo[] result = new ServiceInfo[0];
			
			if (serviceInfos.size() == 0) {
				return result;				
			}
			try {
				result = (ServiceInfo[]) serviceInfos.toArray(result);
			}catch (Exception e){
				logger.error("could not convert result to ServiceInfo array! Returning empty");
				result = new ServiceInfo[0];
			}
			
			return result;
		}
		
		public Vector getUuids() {
			return uuids;
		}
		public void addResult(String uuid, ServiceInfo si) {
			if (uuid == null)
				return;
			uuids.add(uuid);
			if (si != null)
				serviceInfos.add(si);			
		}
		public void clear() {
			uuids = new Vector();
			serviceInfos = new Vector();
			unrecognizedAttr = false;
			skippedAttr = "";
		}
		public boolean hasUnrecognizedAttr() {
			return unrecognizedAttr;
		}
		public void setUnrecognizedAttr(boolean unrecognizedAttr) {
			this.unrecognizedAttr = unrecognizedAttr;
		}
	}
	
	private void debug_PrintRoutingTable_del() {
		System.out.println("\n ************** Routing Table **********************");
		for (Iterator it = ContextRoutingService.contextSearch.keySet().iterator(); it.hasNext();) {
			String key = (String) it.next();
			System.err.println("For key triple '"+key+"' entries are");
			HashSet hs = findContext(key, "", "");
//			HashSet hs = (HashSet) ContextRoutingService.contextSearch.get(key);
			for (Iterator it2 = hs.iterator(); it2.hasNext(); ) {
				System.err.println("\t"+it2.next());
			}			
		}
		System.out.println("\n ********************  END    **********************");
	}
	private void debug_PrintRouteToTopTable_del() {
		System.out.println("\n ************** Routing Table **********************");
		for (Iterator it = ContextRoutingService.routeToTop.keySet().iterator(); it.hasNext();) {
			String key = (String) it.next();
			System.err.println("For child '"+key+"' entries are");
			Vector v = (Vector) ContextRoutingService.routeToTop.get(key);
//			HashSet hs = (HashSet) ContextRoutingService.contextSearch.get(key);
			for (Iterator it2 = v.iterator(); it2.hasNext(); ) {
				System.err.println("\t"+it2.next());
			}			
		}
		System.out.println("\n ********************  END    **********************");
	}
	
}
