/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.contextrouting;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;


import at.ac.tuwien.vitalab.cosine.rating.ARating;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo;

/**
 * @author hpsaier
 * this class implements a interface in order to route (in a tree)
 * the best matching context information against a xpath request
 * As a prerequisite the class must know a few {@link ServiceInfo}
 * which are collected in nodeSearch and contextSearch. A following xpath 
 * request is parsed for correctness and predicates. Finally a match
 * is gathered from the two search maps
 */
public class ContextRoutingService /*implements IContextRouting*/{


	/**
	 * Class is a singleton and this is the one and only instance 
	 */
	private static ContextRoutingService instance;	
	/**
	 * @deperciated
	 * contains all data of the contextmodel
	 */
	//static Ini iniContextFile = null;
	/**
	 * holds the tree dependencies child -> fathers
	 * [child -> Vector[father, ...]]
	 */
	static Map routeToTop = new HashMap();
	/**
	 * represents the routing table with entries e.g. [[nodetype, var, val] -> [ServiceInfo,...]]
	 */
	static Map contextSearch = Collections.synchronizedMap(new HashMap());
	
	
//	static Map nodeMap = Collections.synchronizedMap(new HashMap(5));
//	static Map valueMap = Collections.synchronizedMap(new HashMap(5));
//	static Map attributeMap = Collections.synchronizedMap(new HashMap());
	/**
	 * map containing all subscriptions 
	 * [idSubscription -> [caller, {@link ContextRoutingXPathParser}]] 
	 */
	static Map subscriptionMap = Collections.synchronizedMap(new HashMap());
	

	/**
	 * to sync operations on routing table
	 */
	private static short accessRoutingTable = 1;
	
	/**
	 * to sync subscriber
	 */
	private static short accessSubscriberMap = 1;
		
	private Object requestResult[] = new Object[2];
	
	static IndividualContextRating rating;
	
	static NotificationDeamonThread ndt = null;

	static int TIMEOUT = 3; //sec
	static boolean NEW_NOTIFICATION = true;

	/**
	 * Constructor
	 */
	private ContextRoutingService() throws Exception{
		ContextRoutingHandler contextRoutingHandler = new ContextRoutingHandler();
		contextRoutingHandler.checkPrerequisite();
		contextRoutingHandler.readIniFiles();
		contextRoutingHandler.createDependencyTree();
		rating = new IndividualContextRating();	
	}
	
	/**
	 * "constructor"
	 * @return the singleton
	 * 
	 */
	public static synchronized ContextRoutingService getInstance () throws Exception{
		
	    if (ContextRoutingService.instance == null)
	    	ContextRoutingService.instance = new ContextRoutingService();
	
	    return ContextRoutingService.instance;
	 }	
	/**
	 * add new sensor information to router
	 * @param aServiceInfo new info
	 */
	public void newService (ServiceInfo aServiceInfo) {
//		System.err.println("NEW "+aServiceInfo);

//		ContextRoutingHandler contextRoutingHandler = new ContextRoutingHandler();
		ContextRoutingHandler contextRoutingHandler = new ContextRoutingHandler();
		
		WAIT_OPERATION(accessRoutingTable);
		contextRoutingHandler.addContext(aServiceInfo);
		SIGNAL_OPERATION(accessRoutingTable);
		if (NEW_NOTIFICATION)
			startNotificationDaemon();
		else {
			try {			
				WAIT_OPERATION(accessSubscriberMap);
				notifySubscriber();
			} catch (Exception e) {
				e.printStackTrace();
			}
			SIGNAL_OPERATION(accessSubscriberMap);
		}
//		System.out.println("+++++++");
	}
	/**
	 * delete all information about a sensor from router
	 * @param aServiceInfo info to delete
	 */
	public void deleteService(ServiceInfo aServiceInfo) {
		//		ContextRoutingHandler contextRoutingHandler = new ContextRoutingHandler();
//		ContextRoutingHandlerNew contextRoutingHandler = new ContextRoutingHandlerNew();
//		WAIT_OPERATION();
//		contextRoutingHandler.removeContext(aServiceInfo);
//		SIGNAL_OPERATION();
		//removeContext has been depreciated so there is no reason in loosing time...		
		//... just notify
		if (NEW_NOTIFICATION)
			startNotificationDaemon();
		else {
			try {
				WAIT_OPERATION(accessSubscriberMap);
				notifySubscriber();
			} catch (Exception e) {
				e.printStackTrace();
			}
			SIGNAL_OPERATION(accessSubscriberMap);
		}
//		System.out.println("------");
	}
	
	private void startNotificationDaemon() {
		if (ndt == null || !ndt.isAlive()) {
			ndt = new NotificationDeamonThread();
			ndt.setDaemon(true);
			try {
				ndt.start();
			} catch (java.lang.IllegalThreadStateException ite) {
				
			}
		}		
	}
	private String prepareXPath(String xpath) {
		String result = xpath;
		if (xpath==null || xpath.equals("") || 
				xpath.equals("/") || xpath.equals("//")) {
			result=" ";
		}		
		return result;
	}
	
	/**
	 * Method allows to update the various values considered for ratio.
	 * @param aServiceInfo the associated service
	 * @param type distinct update value
	 * @param value
	 */
	public void updateServiceRating(ServiceInfo aServiceInfo, ARating type, double value) {
		ContextRoutingHandler contextRoutingHandler = new ContextRoutingHandler();
		WAIT_OPERATION(accessRoutingTable);
		contextRoutingHandler.updateRatio(aServiceInfo, type, value);		
		SIGNAL_OPERATION(accessRoutingTable);
	}
	/**
	 * gets a result for a given xpath-query 
	 * @param xpath query
	 * @return result array
	 */
	private void getResult(String xpath) throws Exception {

		xpath = prepareXPath(xpath);
		
		ServiceInfo [] infoArray = null;
		String xpathResult = "";
		ContextRoutingHandler contextRoutingHandler = new ContextRoutingHandler();
		
		//parsing the stuff
		ContextRoutingXPathParser pathParser = new ContextRoutingXPathParser(xpath);
		String parsermsg = pathParser.parseXPathStatement();
		if (!parsermsg.equals(ContextRoutingXPathParser.GOOD))
			throw new Exception(parsermsg);
			

		//xpath against model check
		if (!(contextRoutingHandler.checkNodeRelationShip(pathParser.getLastNode())))
			throw new Exception("XPATH-Error: Model and received xpath do not match for xpath:'"+xpath+"'!");

		//All checks positive -> enter critical section
		WAIT_OPERATION(accessRoutingTable);
		infoArray = contextRoutingHandler.answerQuery(pathParser);
		xpathResult =  contextRoutingHandler.getResultingXPath();
		SIGNAL_OPERATION(accessRoutingTable);

		requestResult[0] = xpathResult;
		requestResult[1] = infoArray;		
	}
	
	/**
	 * returns a result to the query immediately
	 * @param xpath query to look for
	 * @return map of matching sensor info. Layout [optimized xpath, ServiceInfo array]
	 * @throws Exception exceptions caused by xpath parser
	 */
	public Object[] findQueryDestination(String xpath) throws Exception{
		
		xpath = prepareXPath(xpath);
		
		Object result[] = requestResult;
		getResult(xpath);
		
		return result;
	}
	/**
	 * The xpath is saved and on trigger a callback with idSubscrition is invoked
	 * @param xpath query to look for
	 * @param idSubscription identifier for callback accounting
	 * @throws Exception exceptions caused by xpath parser
	 */
	public void findSubscriptionDestination(String xpath, String idSubscription,
											IContextRoutingServiceCallback caller) throws Exception{
		
		xpath = prepareXPath(xpath);
		
		ContextRoutingHandler contextRoutingHandler = new ContextRoutingHandler();
		ContextRoutingXPathParser pathParser = new ContextRoutingXPathParser(xpath);

		try {
			executeCallback(contextRoutingHandler, pathParser,idSubscription, caller,xpath);
			WAIT_OPERATION(accessSubscriberMap);
			subscriptionMap.put(idSubscription, new Object[] {caller, pathParser});
			SIGNAL_OPERATION(accessSubscriberMap);
		} catch (Exception e){
			//forward
			throw e;
		}
			
	}
	/**
	 * The xpath is saved and on trigger a callback with idSubscrition is invoked
	 * @param xpath query to look for
	 * @param idSubscription identifier for callback accounting
	 */
	public void removeSubscriptionDestination(String idSubscription) {
			WAIT_OPERATION(accessSubscriberMap);
			subscriptionMap.remove(idSubscription);
			SIGNAL_OPERATION(accessSubscriberMap);
	}
	
	private void executeCallback (ContextRoutingHandler contextRoutingHandler,
			ContextRoutingXPathParser pathParser,
			String idSubscription,
			IContextRoutingServiceCallback caller,
			String xpath) throws Exception{
		boolean result = true;
		Object[] requestRes = requestResult;
		
		//causes an exception on error
		getResult(xpath);
		
		if (caller == null)
//			("caller is null");
			;
		else if (requestRes[0] == "")
//			System.err.println("result is null");
			;
		else
			caller.queryUpdate(requestRes, idSubscription);
		
	}
	/**
	 * notify all subscriber with the results of the rerun query
	 */
	private void notifySubscriber() throws Exception {
		ContextRoutingXPathParser pathParser;
		IContextRoutingServiceCallback caller; 		
		String idSubscription;
		ContextRoutingHandler contextRoutingHandler = new ContextRoutingHandler();
		
		for (Iterator it = subscriptionMap.keySet().iterator(); it.hasNext();) {
			idSubscription = (String)it.next();			
			Object[] caller_parser = (Object[]) subscriptionMap.get(idSubscription);
			pathParser = (ContextRoutingXPathParser) caller_parser[1];
			caller = (IContextRoutingServiceCallback) caller_parser[0];
			executeCallback(contextRoutingHandler, pathParser, idSubscription, caller, pathParser.getXPath());
		}
		
	}

	/*********************************** Sync **********************/
	private synchronized void WAIT_OPERATION(short token) {
		
		while (token == 0) {
			try {
				wait();
			} catch (InterruptedException e) {

			}
		}
		token--;
	}
	
	private synchronized void SIGNAL_OPERATION(short token) {
		token++;
		notify();
	}
	
	
	class NotificationDeamonThread extends Thread {
		public void run() {
			try {
				Thread.sleep(TIMEOUT*1000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			try {
				WAIT_OPERATION(accessSubscriberMap);
				notifySubscriber();
			} catch (Exception e) {
				e.printStackTrace();
			}
			SIGNAL_OPERATION(accessSubscriberMap);
		}
	}
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
	}

}
