/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.initservice;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.LinkedHashMap;
import java.util.Vector;

import org.ini4j.Ini;
import org.ini4j.InvalidIniFormatException;
import org.ini4j.Ini.Section;

import at.ac.tuwien.vitalab.cosine.Utils;
import at.ac.tuwien.vitalab.cosine.webservice.ASensorWebService;

import com.eaio.uuid.UUID;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;

public class MobileDeviceInitService {

	public static final short ERROR = -1;
	/**
	 * logger
	 */
	private static Logger logger=
		LoggerFactory.getLogger(MobileDeviceInitService.class);

	/**
	 * UI init file
	 */
	private static final String iniProfileName = "conf/profile.ini";
	/**
	 * init file defining the sensor properties
	 */
	private static final String iniContextFileName = "conf/contextmodel.ecs";
	/**
	 * contains all data of the contexmodel
	 */
	private Ini iniContextFile = null;
	/**
	 * contains all user predefined context data
	 */
	private Ini iniProfileFile = null;

	/**
	 * reference to UI class
	 */
	private MobileDeviceInitUI UI = null;

	/**
	 * collect finished and semi-finished context sensors Format: name ->
	 * [uuid/deployed?/ASenserWebService]
	 */
	private Map deploymentList;
	/**
	 * List that accounts dependencies of context sensors Format: dependency ->
	 * [sensorA, sensorB]
	 */
	private Map onHoldList;
		
	/**
	 * Class is a singleton and this is the one and only instance 
	 */
	private static MobileDeviceInitService instance;
	
	private static final boolean FAST_SETUP = false;

	/**
	 * Singleton: PRIVATE constructor necessary 
	 */
	private MobileDeviceInitService() throws Exception{
		// check for all requirements first
		checkPrerequisite();

		// setup all ini information for easy access
		readIniFiles();

		// init deployment maps
		deploymentList = new LinkedHashMap();
		onHoldList = new LinkedHashMap();	
	}	
	/**
	 * "constructor"
	 * @return the singleton
	 * 
	 */
	public static synchronized MobileDeviceInitService getInstance() throws Exception {
	    if (MobileDeviceInitService.instance == null)
	    	MobileDeviceInitService.instance = new MobileDeviceInitService();
	    
	    return MobileDeviceInitService.instance;
	  }

	/**
	 * Main entrance to this object. The beloved service is send to the
	 * initService which check all requirements to deploy it. Another option is
	 * to put it on a hold list
	 * 
	 * @param aWebService
	 * @return success of deployOperation (not yet in use -> always true)
	 */
	public boolean deploySensor(ASensorWebService aWebService) {

		generateContextInfo(aWebService);

		return true;
	}

	/**
	 * build and invoke the UI
	 */
	private void showUI(String title, String contextSection) {
		
		UI.getUIAccess(); //<- blocking
		UI.setFieldsPerPage(2);
		UI.addFieldsToUI();
		// UI.showUI();
		thinlet.FrameLauncher fl = null;
		fl = new thinlet.FrameLauncher(title, UI, 208,
				256);
		UI.waitForResult(); //<-blocking
		fl.dispose();
		//save ui fields
		//TODO: ...
		
		saveUIValues(contextSection);
//		saveProfile(contextSection); 
	}

	private void showAlibiUI(String title, String contextSection) {
		
		UI.getUIAccess(); //<- blocking
		UI.setFieldsPerPage(2);
		UI.addFieldsToUI();
		// UI.showUI();
		thinlet.FrameLauncher fl = null;
		fl = new thinlet.FrameLauncher(title, UI, 208,
				256);
		UI.waitForResult(); //<-blocking
		fl.dispose();
		//save ui fields
		saveUIValues(contextSection); 
	}


	/**
	 * @depreciate
	 * SHOULD be called for eventual error handling
	 * 
	 * @return information of remaining services
	 *//*
	public String allServicesStarted() {
		if (deploymentList == null)
			return "IS_EMPTY";
		
		Iterator it = deploymentList.values().iterator();
		Vector v = null;
		while (it.hasNext()) {
			v = (Vector) it.next();
			if ((Boolean)v.get(1) == Boolean.FALSE)
				return "NOT_ALL_STARTED";
		}
		
		return "ALL_STARTED";
	}*/
	private boolean loadIniProfile() {
		boolean result = true;
		try {
			iniProfileFile.load(new FileReader(new File(System
					.getProperty("user.dir")
					+ System.getProperty("file.separator")
					+ MobileDeviceInitService.iniProfileName)));
		} catch (InvalidIniFormatException e1) {
			result = false;
			logger.error(e1.getMessage());
		} catch (FileNotFoundException e1) {
			result = false;
			logger.error(e1.getMessage());
		} catch (IOException e1) {
			result = false;			
			logger.error(e1.getMessage());
		}
		return result;		
	}
	private boolean saveIniProfile() {
		boolean result = true;
		try {
			iniProfileFile.store(new FileWriter(new File(System
					.getProperty("user.dir")
					+ System.getProperty("file.separator")
					+ MobileDeviceInitService.iniProfileName)));
		} catch (IOException e) {
			result = false;
			logger.error("Could not write to Profile-File!!! Message: "+e.getMessage());
		}
		return result;
	}

	private void saveUIValues(String contextSection) {
		Map UiMap = UI.getUiMap();

		Section sec = (Section) iniProfileFile.get(contextSection);
		if (sec == null)
			sec = iniProfileFile.add(contextSection);
		
		if (UiMap != null) {
			Iterator it = UiMap.keySet().iterator();
			while (it.hasNext()) {
				String key = (String) it.next();
				String[] type_label = MobileDeviceInitService.split2(key, "_");
				Object[] o = (Object[]) UiMap.get(key);
				sec.put(type_label[1], UI.getValueFromField(o[1]));
				// update also context of this section
				this.updateContextValue(contextSection, "field." + type_label[1], UI
						.getValueFromField(o[1]), false);
			}
		}
		
	}

	/**
	 * update the profile - file for next startup and set new gathered 
	 * values to context information also
	 * @param contextSection section to update
	 */
	private void saveProfile(String contextSection) {

//		Ini profileIni = new Ini();
		// if profile- file does not exist -> create
		if (iniProfileFile == null) {
			createProfileFile();
			loadIniProfile();
		}

		Section sec = (Section) iniProfileFile.get(contextSection);
		if (sec == null)
			sec = iniProfileFile.add(contextSection);
		
		//must have!!!
		sec.put("uuid", this.getContextValue(contextSection, "field.uuid"));
		//could have!!!
		if (this.getContextValue(contextSection, "field.belongsTo") != null)
			sec.put("belongsTo", this.getContextValue(contextSection, "field.belongsTo"));
		
	}
	/**
	 * creates a fresh profile-file
	 * 
	 * @return success of creation
	 */
	private boolean createProfileFile() {
		File file = new File(System.getProperty("user.dir")
				+ System.getProperty("file.separator")
				+ MobileDeviceInitService.iniProfileName);

		try {
			file.createNewFile();
			iniProfileFile = Utils.readIniFiles(MobileDeviceInitService.iniProfileName);
		} catch (IOException e) {
			// e.printStackTrace();
			logger.error("Could not write to Profile-File!!! Message: "+e.getMessage());
			return false;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return true;
	}

	/**
	 * quick update the context map
	 * 
	 * @param sec
	 *            update this section
	 * @param key
	 *            ... at this key
	 * @param value
	 *            ... with this value
	 */	
	private void updateContextValue(String sec, String key, String value,
			boolean append) {
		String oldValue = "";
		if (!append /* = override */)
			((Section) iniContextFile.get(sec)).put(key, value);
		else {
			//System.out.println((String) ((Section) iniContextFile.get("Worker@http://www.workpad-project.eu/contextmodel")).get(key));
			oldValue = (String) ((Section) iniContextFile.get(sec)).get(key);
			if (oldValue.equals("") || oldValue.equals("")) //override empty value
				((Section) iniContextFile.get(sec)).put(key, value);
			else
				((Section) iniContextFile.get(sec)).put(key, oldValue + ", "
					+ value);
		}

	}
	/**
	 * get a value form the actual context information
	 * 
	 * @param sec
	 *            which context
	 * @param key
	 * @return value
	 */
	private String getContextValue(String sec, String key) {
		return (String) ((Section) iniContextFile.get(sec)).get(key);
	}
	/**
	 *  generate all context: autogenerated, UI, sensors
	 *  ----------------> THIS provides the main function <--------------
	 * @param aWebService sensor service for which info is gathered
	 */
	private void generateContextInfo(ASensorWebService aWebService) {

		//is required to find the section in the context init file
		String contextSectionName = aWebService.getContextTypeName()+"@"+aWebService.getContextTypeNamespace();
		Section sec = null;
		Section profileSec = null;
		if (iniProfileFile != null) 
			profileSec = (Section) (iniProfileFile.get(contextSectionName));
			
		
		String propertykey, propertyvalue, parents = "";

		// init UI object
		UI = new MobileDeviceInitUI();

		sec = (Section) iniContextFile.get(contextSectionName);
		if (sec == null)
			return;
		
		// 1st generate UUID		
		propertykey = "field.uuid";
		if (sec.get(propertykey) != null) {
			if (profileSec != null && profileSec.get("uuid") != null) { // from profile
				updateContextValue(contextSectionName, propertykey, (String)profileSec.get("uuid"), false);
			}
			else {
				propertyvalue = (String) sec.get(propertykey);
				autoGen(parsePropertyType(propertyvalue), propertykey,
					contextSectionName);
			}
		}
		
		// 2nd collect all context data
		Iterator itSection = sec.keySet().iterator();
		while (itSection.hasNext()) {
			propertykey = (String) itSection.next();
			propertyvalue = (String) sec.fetch(propertykey);

			if (propertyvalue.startsWith("askUser")) {

				createUI(sec.getName(), propertykey, propertyvalue,"");				
			} else if (propertyvalue.startsWith("fromList")) {
				// TODO enable option ui
				
			}
		}

		// call UI
		if (!UI.getUiMap().isEmpty()) // there is something to display
			//showUI(sectionContext[0]/* title = section */, contextSectionName);
			if (!FAST_SETUP)
				showUI(aWebService.getContextTypeName()/* title = section */, contextSectionName);
			else {
				if (aWebService.getContextTypeName().startsWith("Team"))
					showUI(aWebService.getContextTypeName()/* title = section */, contextSectionName);
				else
					showAlibiUI(aWebService.getContextTypeName()/* title = section */, contextSectionName);
			}
		// 3) now all contextdata MUST be ready -> try to deploy if parent link
		// exists
		propertykey = "field.belongsTo";
		if (sec.get(propertykey) != null) {
			propertyvalue = (String) sec.get(propertykey);

			if (propertyvalue.startsWith("localParent")) {
//				if (profileSec != null && profileSec.get("belongsTo") != null && !profileSec.get("belongsTo").equals("")) { // from profile
//					updateContextValue(contextSectionName, propertykey, (String)profileSec.get("belongsTo"), false);
//				}
				parents = sec.fetch("parents");
				localParent(propertykey, contextSectionName, parents, aWebService);				
			} else if (propertyvalue.startsWith("remoteParent")) {
				parents = sec.fetch("parents");
				remoteParent(propertykey, contextSectionName, parents, aWebService);				
			}
		} else { // must be root node
			Vector uuid_deployed = new Vector();
			uuid_deployed.add(getContextValue(contextSectionName, "field.uuid"));
			uuid_deployed.add(Boolean.TRUE);
			uuid_deployed.add(aWebService);
			deploymentList.put(contextSectionName, uuid_deployed);
			deploy(contextSectionName, aWebService);
			notifyChildren(contextSectionName);
		}
		//save all data
//TODO: ...		
//		saveProfile(contextSectionName);
//TODO:	
//		saveIniProfile();
	}
	/** deploy on remote parent
	 * @param key belongsTo
	 * @param contextSection
	 * @param parents list of parents
	 * @param aWebService sensor of interest
	 */
	private void remoteParent(String key, String contextSection, String parents,
			ASensorWebService aWebService) {

		String[] remoteParents = MobileDeviceInitService.split2(parents, ",");
		
//		//init context info
		updateContextValue(contextSection, key,
				"", false);
		// parents=Worker@http://www.workpad-project.eu/contextmodel,
		// Leader@http://www.workpad-project.eu/contextmodel
		for (int i = 0; i < remoteParents.length; i++) {
			//ask user for selection
			String parent = remoteParents[i];
			String section = contextSection;
			UI = new MobileDeviceInitUI();
		
			String[] parent_namespace = split2(parent, "@");
			String xpath = "//"+parent_namespace[0];

			if (parent_namespace.length == 2) {
				xpath +="[@*[namespace-uri()='"+parent_namespace[1]+"']]";
			}

			createUI(section, key,"askUser(combobox)",xpath);
			showUI(aWebService.getContextTypeName()/* title = section */, section);
			Vector uuid_deployed = new Vector();
			String uuid = getContextValue(section, "field.belongsTo");
//			String uuid = "3e9981f0-aca2-11dd-9099-0015c5568d5b";
			uuid_deployed.add(uuid);
			if (uuid.equals(""))
				uuid_deployed.add(Boolean.FALSE);
			else {
				uuid_deployed.add(Boolean.TRUE);
				deploymentList.put(parent, uuid_deployed);
			}
			//add to deploymentList
			
			if (findParents(remoteParents[i], contextSection, aWebService))
				break;
			
		}
	}
	/** 
	 * find context info for local parent
	 * @param key variable to set
	 * @param contextSection
	 * @param parents list of parents
	 * @param aWebService sensor of interest
	 */
	private void localParent(String key/* belongsTo */, String contextSection,
			String parents, ASensorWebService aWebService) {

		String[] localParents = MobileDeviceInitService.split2(parents, ",");
		
		//init context info
		updateContextValue(contextSection, key,
				"", false);
		// parents=Worker@http://www.workpad-project.eu/contextmodel,
		// Leader@http://www.workpad-project.eu/contextmodel
		for (int i = 0; i < localParents.length; i++) {
			if (findParents(localParents[i], contextSection, aWebService))
				break;
		}                    
	}
	/**
	 * 1st tries to deploy actual (aWebService) by looking for parent (parent)
	 * 2nd adds a new entry to deployment list (deploymentList)
	 * 3rd adds a new entry to hold list if not deployed
	 * 4th notifies the children of presence  
	 * @param parent sensors parent
	 * @param section sensor describer
	 * @param aWebService the sensor
	 * @return success of deploy for section
	 */
	private boolean findParents(String parent, String section,
			ASensorWebService aWebService) {

		Vector uuid_deployed = (Vector) deploymentList.get(parent);
		boolean wasDeployed = tryToDeploy(parent, section, uuid_deployed,
				aWebService);

		uuid_deployed = new Vector();
		uuid_deployed.add(getContextValue(section, "field.uuid"));
		uuid_deployed.add(wasDeployed ? Boolean.TRUE : Boolean.FALSE);
		uuid_deployed.add(aWebService);


		if (!wasDeployed)
			enterOnHold(parent, section);
		deploymentList.put(section, uuid_deployed);
		notifyChildren(section);
		return wasDeployed;
	}
	/**
	 * enter a sensor to the hold list. This happenes if parent is not yet
	 * available
	 * @param parent parent child is waiting for
	 * @param child waiting sensor
	 */
	private void enterOnHold(String parent, String child) {
		//System.err.println("->"+parent+" - "+child);
		Vector v = (Vector) onHoldList.get(parent);
		if (v != null)
			v.add(child);
		else {
			v = new Vector();
			v.add(child);
			onHoldList.put(parent.trim(), v);
		}
	}
	/**
	 * loop on hold list children matching the parent
	 * @param parent sensor parent processed at the moment
	 */
	private void notifyChildren(String parent) {
		Vector v = (Vector) onHoldList.get(parent);
		Vector uuid_deployedParent, uuid_deployedChild;
		String child;
		
		
		// no children waiting
		if (v == null)
			return;

		uuid_deployedParent = (Vector) deploymentList.get(parent);
		for (int i = 0; i < v.size(); i++) {
			child = (String) v.get(i);
			uuid_deployedChild = (Vector) deploymentList.get(child);
			// if child not already deployed
			if ((Boolean) uuid_deployedChild.get(1) == Boolean.FALSE) {
				tryToDeploy(parent, child, uuid_deployedParent,
						(ASensorWebService) uuid_deployedChild.get(2));
				uuid_deployedChild.set(1, Boolean.TRUE);
				deploymentList.put(child, uuid_deployedChild);
			}

		}
		onHoldList.remove(parent);
	}
	/**
	 * try to deploy sensor defined by section
	 * @param parentName sensor's parent
	 * @param section sensor's id
	 * @param uuid_deployed parent's properties (uuid,...)
	 * @param aWebService the sensor to deploy
	 * @return is sensor deployable
	 */
	private boolean tryToDeploy(String parentName, String section,
			Vector uuid_deployed, ASensorWebService aWebService) {
		boolean deployable = false;

		// is parent already in list?
		if (uuid_deployed != null) {
			// parent exists -> let us make an entry and deploy
			// update context info "parents" to "belongsTo":
			updateContextValue(section, "field.belongsTo",
					(String) uuid_deployed.get(0), false);
			
			deployable = true;
//			System.out.print(parentName + " is now deploying ");
//			logger.info(parentName + " is now deploying ");
			deploy(section, aWebService);
			//save belongsTo
			//TODO: ...			
//			saveProfile(section);
		}

		return deployable;
	}
	/**
	 * deploy a sensor
	 * @param section sensor's id
	 * @param aWebService sensor as a service
	 */
	private void deploy(String section/* = sensor context */,
			ASensorWebService aWebService) {
		Section sec = (Section) iniContextFile.get(section);
		Iterator it = sec.keySet().iterator();
		String contextKey;
		Hashtable contextMap = new Hashtable();

		saveProfile(section);
		saveIniProfile();
		
		while (it.hasNext()) {
			contextKey = (String) it.next();
			if (contextKey.startsWith("field")) { // found a context field
//				System.err.println(contextKey+" - "+sec.get(contextKey));
				// = key with no "field." and value 
				//contextMap.put(split2(contextKey,".")[1], sec.get(contextKey));
				contextMap.put(parsePropertyKey(contextKey), sec.get(contextKey));
			}
		}
		aWebService.deploy(contextMap);
		//logger.info(" '" + section + "'");
	}
	/**
	 * Autogenerate a property
	 * @param type defines the method of autogeneration
	 * @param key of section
	 * @param sec variable to set
	 */
	private void autoGen(String type, String key, String sec) {
		// could be done by middleware
		//System.out.println("generate for: " + key + ": on " + sec);
		if (type.equals("UUID"))
			updateContextValue(sec, key, new UUID().toString(), false);
	}
	/**
	 * create the UI input fields
	 * @param sensor the field belongs to
	 * @param field variable to ask
	 * @param value value to check
	 */
	private void createUI(String sensor, String field, String value, String xpath) {

		Section sec = null;
		String profileKey, profileType, profileValue = "";
		boolean isOptional = false;
		boolean isRemote = (xpath.equals("") ? false : true);

		profileKey = parsePropertyKey(field);
		profileType = parsePropertyType(value);

		if (value.indexOf(":opt") > 0)
			isOptional = true;
		
		if (iniProfileFile != null) {
			sec = (Section) iniProfileFile.get(sensor);
			if (sec != null)
				profileValue = (String) sec.get(profileKey);
//			System.err.println("Using values: "+sec.getName()+", "+profileKey+", "+profileValue);				
//			this.updateContextValue(sec.getName(), profileKey, profileValue, false);
			
		}
		//TODO: what about :opt:selectionA, selectionB
		if (profileType.equals("combobox") && value.indexOf(":") > 0) {
			profileValue = value.substring(value.indexOf(":")+1);
		}

		
		if (profileValue == null)
			profileValue = "";
		
//		if (!profileValue.equals(""))
//			return;

		
		if (profileType.equals("String") || profileType.equals("Int")
				|| profileType.equals("Float") // || and so on
		)
			UI.generateFields("text_" + profileKey, profileValue, isOptional, isRemote, xpath);
		
		if (profileType.equals("combobox"))
			UI.generateFields("combobox_" + profileKey, profileValue, isOptional, isRemote, xpath);
			
		// TODO options ...
	}

	/**
	 * lists and checks all prerequisite for successful startup
	 */
	private void checkPrerequisite() throws Exception{
//		boolean fail = false;
		File file;

		//
		// 1st... context ini file is required
		file = new File(MobileDeviceInitService.iniContextFileName);
		if (!file.exists()) {
			/*logger.error("INI-ERROR: your context init file does net exist/cannot be found in: "
					+ file.getAbsolutePath());*/
			throw new Exception("INI-ERROR: your context init file does net exist/cannot be found in: "
					+ file.getAbsolutePath());
			// file = new
			// File(System.getProperty("user.dir")+System.getProperty("file.separator")+"conf"+System.getProperty("file.separator")+MobileDeviceInitService.iniContextFileName);
			/*
			 * TODO: if (!file.exists()) { System.err .println("... still no
			 * luck in "+file.getAbsolutePath()+"!\nCannot continue ->
			 * EXITing");
			 * 
			 *  }
			 */
			//fail = true;
		}
		// 2nd... is user.dir writeabel
		file = new File(System.getProperty("user.dir")
				+ System.getProperty("file.separator") + "dummyfile");
		try {
			file.createNewFile();
			file.delete();
		} catch (Exception e) {
			logger.error("INI-WARNING: cannot write to your home dir... Ini values are lost");
		}

/*		if (fail)
			System.exit();*/
	}
	/**
	 * wrapper
	 */
	private void readIniFiles() throws Exception{
		try {
			iniContextFile = Utils.readIniFiles(MobileDeviceInitService.iniContextFileName);
		} catch (Exception e) {
			throw new Exception ("INI-ERROR: '"+e.getMessage()+"' contextmodel could not be read");
		}
		try {
			iniProfileFile = Utils.readIniFiles(MobileDeviceInitService.iniProfileName);
		}catch (Exception e){
			logger.warning("INI-WARNING: '"+e.getMessage()+"'... continuing!");
			//System.exit(ERROR);
		}
	}

	/** **************** helper methods **************************** */
	/**
	 * wrapper
	 */
	public static String[] split2(String toSplit, String pattern) {
		return Utils.split2(toSplit, pattern);
	}
	/**
	 * whole keys look like field.thisIsTheKey
	 * 
	 * @param wholeKey
	 * @return key
	 */
	private String parsePropertyKey(String wholeKey) {
		return wholeKey.substring(wholeKey.indexOf('.') + 1);
	}

	/**
	 * whole properties look like function(thisIsTheType)
	 * 
	 * @param wholeProperty
	 * @return
	 */
	private String parsePropertyType(String wholeProperty) {
		return wholeProperty.substring(wholeProperty.indexOf('(') + 1,
				wholeProperty.indexOf(')'));
	}

	/*call me just on debug!!!*/
//	public void DEBUG() {
//		Iterator it = onHoldList.keySet().iterator();
//		String Id;
//		while (it.hasNext()) {
//			Vector v = (Vector) onHoldList.get(it.next());
//			System.out.println(v.toArray());
//		}
//		System.out.println("-----------------------");
//		it = deploymentList.keySet().iterator();
//		while (it.hasNext()) {
//			Id = (String) it.next();
//			Vector v = (Vector) deploymentList.get(Id);
//			System.out.println("Sensor: " + Id + " " + (String) v.get(0)
//					+ " -> " + v.get(1));
//		}
//	}

/*	public static void main(String[] args) {
		MobileDeviceInitService mdis = new MobileDeviceInitService();

		/***********************************************************************
		 * test
		 */
		// mdis.test();
		/*mdis.deploySensor(new Object(), "A"); mdis.deploySensor(new Object(),
		"G"); mdis.deploySensor(new Object(), "E"); mdis.deploySensor(new
		 Object(), "D"); mdis.deploySensor(new Object(), "C");
		mdis.deploySensor(new Object(), "B"); mdis.deploySensor(new Object(),
		"H"); mdis.DEBUG();
		*/
		/** ********************************************************** */

		/*
		 * mdis.deploySensor(new Object(),
		 * "Leader@http://www.workpad-project.eu/contextmodel");
		 * mdis.deploySensor(new Object(),
		 * "Team@http://www.workpad-project.eu/contextmodel");
		
		System.exit(0);
	}*/

}
