/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.processing.notification;

import java.util.Hashtable;
import java.util.NoSuchElementException;

import org.kxml2.kdom.Element;
import org.xmlpull.v1.XmlPullParserException;

import at.ac.tuwien.vitalab.cosine.CMS;
import at.ac.tuwien.vitalab.middleware.MiddlewareConfiguration;
import at.ac.tuwien.vitalab.middleware.soap.header.AHeaderElement;

/**
 * Notification receiver header element.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class NotificationReceiverHeader extends AHeaderElement {

	public static final String ALIEN_CLIENT_SOAPACTION=
		CMS.WSNAMESPACE+"client/";
	
	protected String address=null;
	protected boolean receiverIsAlien=true;
	protected String hostUUID=null;
	protected String namespace=null;
	protected String name=null;
	
	protected static Element receiverElement=new Element();
	protected static Element typeElement=new Element();
	protected static Element addressElement=new Element();
	protected static Element hostUUIDElement=new Element();
	protected static Element receiverNameElement=new Element();
	protected static Element receiverNSElement=new Element();
	
	protected final static int FOUND_ADDRESS=1;
	protected final static int FOUND_HOSTID=2;
	protected final static int FOUND_NAMESPACE=4;
	protected final static int FOUND_NAME=8;
	protected final static int FOUND_TYPE=16;
		
	protected final static int FOUND_ALL= FOUND_HOSTID | 
		FOUND_NAMESPACE | FOUND_NAME;

	protected final static int FOUND_ALL_ALIEN=FOUND_ADDRESS;

	
	static {
		receiverElement.setName("ContextNotificationReceiver");
		receiverElement.setNamespace(CMS.WSNAMESPACE);
		
		typeElement.setName("ReceiverType");
		typeElement.setNamespace(CMS.WSNAMESPACE);
		
		addressElement.setName("Address");
		addressElement.setNamespace(CMS.WSNAMESPACE);
		
		hostUUIDElement.setName("HostID");
		hostUUIDElement.setNamespace(CMS.WSNAMESPACE);
		
		receiverNSElement.setName("ReceiverNamespace");
		receiverNSElement.setNamespace(CMS.WSNAMESPACE);
		
		receiverNameElement.setName("ReceiverName");
		receiverNameElement.setNamespace(CMS.WSNAMESPACE);
	}
	
	/**
	 * Constructor.
	 *
	 */
	public NotificationReceiverHeader() {
		super(receiverElement.getName(),receiverElement.getNamespace());
		
		hostUUID=MiddlewareConfiguration.getHostUUID();
		receiverIsAlien=false;
	}

	/**
	 * Constructor, reading in the element from the SOAP headers.
	 * 
	 * @param headers XML header elements.
	 * @throws NoSuchElementException If the SOAP headers do not contain it.
	 * @throws XmlPullParserException If the header is malformed.
	 */
	public NotificationReceiverHeader(Element[] headers) 
	throws NoSuchElementException, XmlPullParserException {
		this();
		receiverIsAlien=true;
		this.load(headers);
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.soap.header.AHeaderElement#load(org.kxml2.kdom.Element)
	 */
	public void load(Element el) throws XmlPullParserException {
		if (!this.hasSameName(receiverElement, el)) {
			throw new IllegalArgumentException(
					"Not a valid From element.");
		}

		int found=0;
		hostUUID=null;
		
		for (int i=0;i<el.getChildCount();i++) {
			Object child=el.getChild(i);
			if (!(child instanceof Element)) {
				continue;
			}
			Element tmp=(Element)child;
			if (this.hasSameName(typeElement, tmp)) {
				receiverIsAlien=!"noAlien".equals(tmp.getText(0));
				found|=FOUND_ADDRESS;
			} else if (this.hasSameName(addressElement, tmp)) {
				address=tmp.getText(0);
				if (address==null) {
					throw new XmlPullParserException("Address missing");
				}
				found|=FOUND_ADDRESS;
			} else if (this.hasSameName(hostUUIDElement, tmp)) {
				hostUUID=tmp.getText(0);
				if (hostUUID==null) {
					throw new XmlPullParserException("Host ID missing");
				}
				if (hostUUID.startsWith("uuid:")) {
					hostUUID=hostUUID.substring(5);
				}
				found|=FOUND_HOSTID;
			} else if (this.hasSameName(receiverNSElement, tmp)) {
				namespace=tmp.getText(0);
				if (namespace==null) {
					throw new XmlPullParserException("Namespace missing");
				}
				found|=FOUND_NAMESPACE;
			} else if (this.hasSameName(receiverNameElement, tmp)) {
				name=tmp.getText(0);
				if (name==null) {
					throw new XmlPullParserException("Name missing");
				}
				found|=FOUND_NAME;
			}
		}
		
		if (receiverIsAlien) {
			if ((found & FOUND_ALL_ALIEN) != FOUND_ALL_ALIEN) {
				throw new XmlPullParserException("Child elements are missing");	
			}
		} else {
			if ((found & FOUND_ALL) != FOUND_ALL) {
				throw new XmlPullParserException("Child elements are missing");	
			}
		}
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.soap.header.AHeaderElement#toXML()
	 */
	public Element toXML() {

		Element receiverEl=this.createWithSameName(receiverElement);

		if (receiverIsAlien) {
			
			Element fromAddressEl=this.createWithSameName(addressElement);
			fromAddressEl.addChild(Element.TEXT, address);
			receiverEl.addChild(Element.ELEMENT, fromAddressEl);
			
		} else {
			
			if (hostUUID!=null) {
				Element fromHostIDEl=this.createWithSameName(hostUUIDElement);
				if (hostUUID.startsWith("uuid:")) {
					fromHostIDEl.addChild(Element.TEXT,hostUUID);				
				} else {
					fromHostIDEl.addChild(Element.TEXT,"uuid:"+hostUUID);
				}
				receiverEl.addChild(Element.ELEMENT, fromHostIDEl);			
			}
			
			Element nameEl=this.createWithSameName(receiverNameElement);
			nameEl.addChild(Element.TEXT, name);				
			receiverEl.addChild(Element.ELEMENT, nameEl);			
			
		}
		
		Element namespaceEl=this.createWithSameName(receiverNSElement);
		namespaceEl.addChild(Element.TEXT, namespace);				
		receiverEl.addChild(Element.ELEMENT, namespaceEl);			

		Element typeEl=this.createWithSameName(typeElement);
		typeEl.addChild(Element.TEXT, receiverIsAlien ? "alien" : "noAlien");				
		receiverEl.addChild(Element.ELEMENT, typeEl);			
		
//		fromEl.setAttribute("http://www.w3.org/2001/12/soap-envelope", 
//				"mustUnderstand", 
//				"1");
		
		return receiverEl;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.soap.header.AHeaderElement#getNamespaceMap()
	 */
	public Hashtable getNamespaceMap() {
		Hashtable result=new Hashtable();
		result.put("esc", CMS.WSNAMESPACE);
		return result;
	}

	/**
	 * Set the address.
	 * 
	 * @param address Address.
	 */
	public void setAddress(String address) {
		this.address=address;
	}
	
	/**
	 * Returns the address.
	 * 
	 * @return address.
	 */
	public String getAddress() {
		return address;
	}

	/**
	 * Returns the Host ID.
	 * 
	 * @return Host ID.
	 */
	public String getHostID() {
		return hostUUID;
	}

	/**
	 * Sets the namespace.
	 * 
	 * @param namespace Receiver service namespace.
	 */
	public void setReceiverNamespace(String namespace) {
		this.namespace=namespace;
	}
	
	/**
	 * Returns the namespace.
	 * 
	 * @return Receiver service namespace.
	 */
	public String getReceiverNamespace() {
		return namespace;
	}
	
	/**
	 * Sets the name.
	 * 
	 * @param name Receiver service name.
	 */
	public void setReceiverName(String name) {
		this.name=name;
	}
	
	/**
	 * Returns the name.
	 * 
	 * @return Receiver service name.
	 */
	public String getReceiverName() {
		return name;
	}

	/**
	 * Returns whether the receiver is an alien node.
	 * 
	 * @return <code>true</code> if it is an alien node, otherwise
	 * <code>false</code>.
	 */
	public boolean isReceiverAlien() {
		return receiverIsAlien;
	}

	
	
}
