/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.processing.notification;

import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Vector;

import org.ksoap2.serialization.SoapObject;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;
import edu.wustl.mobilab.sliver.soap.Channel;
import edu.wustl.mobilab.sliver.soap.j2se.HTTPChannel;

import at.ac.tuwien.vitalab.cosine.ContextData;
import at.ac.tuwien.vitalab.cosine.XMLData;
import at.ac.tuwien.vitalab.cosine.configuration.CIMSConfiguration;
import at.ac.tuwien.vitalab.middleware.Middleware;
import at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery;
import at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceDB;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo;
import at.ac.tuwien.vitalab.middleware.soap.SoapFault;

/**
 * Remote context query. Context notifications will be forwarded via
 * Web service calls.
 * 
 * @author Lukasz Juszczyk
 *
 */
public final class RemoteContextQuery extends AContextQuery 
implements IServiceSubscriber {

	private static Logger logger=
		LoggerFactory.getLogger(RemoteContextQuery.class);
	
	private String queryString=null;
	
	// service query for notification receiver
	private NotificationClientQuery query=null;
	
	private List notificationData=Collections.synchronizedList(new Vector());
	private HashMap tmpDataQueue=new LinkedHashMap();
	
	private ServiceDB serviceDB=Middleware.getServiceDatabase();
	
	private long lastNotificationTS=-1;
	private final long maxInactivity=
		CIMSConfiguration.getMaxNotificationWait()*1000;
	
	private boolean faultReceived=false;
	
	private String subscriptionID=null;
	
	private boolean clientIsAlien=false;
	
	private NotificationReceiverHeader receiverHeader=null;
	
	/**
	 * Constructor.
	 * 
	 * @param queryString Query string.
	 * @param subscriptionID Subscription ID.
	 * @param receiverHeader Receiver SOAP Header.
	 */
	public RemoteContextQuery(String queryString, String subscriptionID, 
			NotificationReceiverHeader receiverHeader) {
		
		super();
		
		this.queryString = queryString;
		this.subscriptionID=subscriptionID;
		this.receiverHeader=receiverHeader;
		
		this.clientIsAlien=receiverHeader.isReceiverAlien();
		
		if (!clientIsAlien) {
			query=new NotificationClientQuery(subscriptionID, receiverHeader);
			
			serviceDB.subscribe(query, this);
		} 

		lastNotificationTS=System.currentTimeMillis();
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.processing.notification.AContextQuery#stop()
	 */
	public void stop() {	
		
		if (!clientIsAlien) {
			serviceDB.unsubscribe(query);
		}
		
		notificationData.clear();
	}
	
	/**
	 * Returns the subscription ID.
	 * 
	 * @return Subscription ID.
	 */
	public String getSubscriptionID() {
		return subscriptionID;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.processing.notification.AContextQuery#getQueryString()
	 */
	public String getQueryString() {
		return queryString;
	}
	
	/**
	 * Sends the context notification to an alien client.
	 */
	private synchronized void notifyAlienClient() {
		
		try {
			
			shiftContextToTmpQueue();

			if (tmpDataQueue.isEmpty()) {
				return;
			}
			
			String url=this.receiverHeader.getAddress();
			String ns=NotificationReceiverHeader.ALIEN_CLIENT_SOAPACTION;
			
			HttpURLConnection connection=
				(HttpURLConnection)new URL(url).openConnection();
			
			final Channel channel=new HTTPChannel(connection,ns);
			
			final SoapObject soapObject=
				new SoapObject(ns, "contextListNotification");
			soapObject.addProperty("subscriptionID", this.subscriptionID);
			soapObject.addProperty(
					"contextDataList", new Vector(tmpDataQueue.values()));
			

			new Thread() {
				public void run() {
					try {
						channel.call(soapObject);
						
						tmpDataQueue.clear();
						lastNotificationTS=System.currentTimeMillis();						
					} catch (Exception e) {
						logger.debug("Cannot propagate context notification: "+
								e.getMessage());
					}
				}
			}.start();
			
		} catch (Exception e) {
			logger.debug("Cannot establish HTTP connection: "+
					e.getMessage());
		}
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.processing.notification.AContextQuery#contextQueryNotification(at.ac.tuwien.vitalab.cosine.ContextData)
	 */
	public void contextQueryNotification(ContextData context) 
	throws InactiveSubscriptionException {
		
		if (faultReceived || (lastNotificationTS>0 && 
				System.currentTimeMillis()-lastNotificationTS>maxInactivity)) {
			throw new InactiveSubscriptionException();
		}

		notificationData.add(context);

		if (!clientIsAlien) {
			serviceDB.checkSubscription(query);
		} else {
			notifyAlienClient();
		}
		
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.processing.notification.AContextQuery#contextQueryNotification(java.util.Vector)
	 */
	public void contextQueryNotification(Vector contextList) 
	throws InactiveSubscriptionException {
		
		if (faultReceived || (lastNotificationTS>0 && 
				System.currentTimeMillis()-lastNotificationTS>maxInactivity)) {
			
			throw new InactiveSubscriptionException();
		}
		
		notificationData.addAll(contextList);
		
		if (!clientIsAlien) {
			serviceDB.checkSubscription(query);			
		} else {
			notifyAlienClient();
		}
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber#notification(at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery, at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
	 */
	public synchronized void notification(AServiceQuery query, 
			ServiceInfo info) {
		
		if ((query!=this.query)) {
			return;
		}
		
		shiftContextToTmpQueue();
		
		if (tmpDataQueue.isEmpty()) {
			return;
		}
		
		NotificationClientQuery clientQuery=(NotificationClientQuery)query;
		
		String subId=clientQuery.subscriptionID;		
		
		try {		
			
			Vector data=new Vector(tmpDataQueue.values());
			
			Iterator dataIter=data.iterator();
			
			while (dataIter.hasNext()) {
				ContextData cd=(ContextData)dataIter.next();
				cd.refreshSendingTimestamp();
			}
			
			info.invokeOperation("contextListNotification", 
					new Object[]{subId,data},null);

			tmpDataQueue.clear();
			lastNotificationTS=System.currentTimeMillis();

		} catch (Exception e) {
			if (e instanceof SoapFault) {
				faultReceived=true;
			}
			logger.debug("Cannot propagate context notification: "+
					e.getMessage());
		}
	}

	/**
	 * Move context data to temporary quere from where it is sent.
	 */
	private void shiftContextToTmpQueue() {
		
		synchronized (notificationData) {

			Iterator ci=notificationData.iterator();

			while (ci.hasNext()) {
				ContextData c=(ContextData)ci.next();
				XMLData id=c.getContextIDElement();
				if (id==null) {
					tmpDataQueue.put(c, c);
				} else {
					tmpDataQueue.put(id, c);
				}
			}
			
			notificationData.clear();
		
//			if (tmpDataQueue.isEmpty()) {
//				return;
//			}
		}
	}
	
	private class NotificationClientQuery extends AServiceQuery {

		protected String subscriptionID;
		protected NotificationReceiverHeader receiverHeader;

		/**
		 * Constructor.
		 * 
		 * @param subscriptionID Subscruption ID.
		 * @param receiverHeader NotificationReceiver-header.
		 */
		public NotificationClientQuery(String subscriptionID, 
				NotificationReceiverHeader receiverHeader) {
			super();
			this.subscriptionID = subscriptionID;
			this.receiverHeader = receiverHeader;
			this.skipDuplicates = false;
		}


		/* (non-Javadoc)
		 * @see at.ac.tuwien.vitalab.middleware.discovery.ServiceQuery#matches(at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
		 */
		public boolean matches(ServiceInfo info) {
			
			try {
				
				// host ID must match
				if (!info.getHostId().equals(receiverHeader.getHostID())) {
					return false;
				}
				
				// service ID must match
				if (!receiverHeader.getReceiverName().equals(info.getId())) {
					return false;
				}

				// service namespace must match
				if (!receiverHeader.getReceiverNamespace().equals(
								info.getNamespace())) {
					return false;
				}

				return true;
				
				
			} catch (Exception e) {
				// just return false
			}
			return false;
		}
	}
}
