/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.processing.notification;

import java.util.Hashtable;

import com.eaio.uuid.UUID;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;

import at.ac.tuwien.vitalab.cosine.configuration.CIMSConfiguration;
import at.ac.tuwien.vitalab.cosine.interfaces.INotificationReceiverWS;
import at.ac.tuwien.vitalab.middleware.AWebService;
import at.ac.tuwien.vitalab.middleware.Middleware;
import at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery;
import at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber;
import at.ac.tuwien.vitalab.middleware.discovery.InactiveServiceException;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceDB;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo;
import at.ac.tuwien.vitalab.middleware.soap.header.SoapHeaders;

/**
 * Context subscription client.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class SubscriptionClient {

	// key = subscription ID
	// value = SubscriptionQuery object
	private Hashtable queries=new Hashtable();

	// key = ServiceInfo object
	// value = subscription ID
	private Hashtable ids=new Hashtable();
	
	private ServiceDB serviceDB=Middleware.getServiceDatabase();
	
	// service which is supposed to receive the updates
	private AWebService receiverService=null;
	
	private static Logger logger=
		LoggerFactory.getLogger(SubscriptionClient.class);
		
	// synchronization locks
	// key = subscription ID
	// value = lock object
	private static Hashtable locks=new Hashtable();
	
	// maximum allowed time of inactivity after which the subscription
	// gets removed
	private static int maxInactivityTime=
		CIMSConfiguration.getMaxSubscriptionInactivity();
	
	/**
	 * Constructor.
	 * 
	 * @param receiver Receiver service.
	 */
	public SubscriptionClient(AWebService receiver) {
		if (!(receiver instanceof INotificationReceiverWS)) {
			throw new IllegalArgumentException("Service must implement " +
					"INotificationReceiverWS interface.");
		}
		
		this.receiverService=receiver;
	}
	
	/**
	 * Returns the service to which the subscription with the passed ID
	 * has been sent.
	 * 
	 * @param id Subscription ID.
	 * @return Service or <code>null</code> if nothing found.
	 */
	public ServiceInfo getServiceBySubscriptionID(String id) {
		try {
			return ((SubscriptionQuery)queries.get(id)).service;
		} catch (Exception e) {
			return null;
		}
	}
	
	/**
	 * Returns the subscription ID of the service.
	 * 
	 * @param s Service.
	 * @return Subscription ID or <code>null</code> if nothing found.
	 */
	public String getSubscriptionIDByService(ServiceInfo s) {
		try {
			return (String)ids.get(s);
		} catch (Exception e) {
			return null;
		}
	}	
	
	/**
	 * Subscribes to a sensor service.
	 * 
	 * @param service Sensor service.
	 * @param queryString Query string.
	 * 
	 * @return Subscription ID.
	 */
	public String subscribeTo(ServiceInfo service, String queryString) {

		// unique ID for each subscription
		String id=new UUID().toString();

		synchronized (getSubscriptionLock(id)) {

			SubscriptionQuery q=new SubscriptionQuery(id,service,queryString);
			
			queries.put(id, q);
			ids.put(service,id);
			
			serviceDB.subscribe(q, q);
			// query matching will perform subscription
		}
		
		return id;
	}

	/**
	 * Performs the subscription communication.
	 * 
	 * @param subscriptionID Subscription ID.
	 */
	private void performSubscription(String subscriptionID) {
		synchronized (getSubscriptionLock(subscriptionID)) {

			SubscriptionQuery q=(SubscriptionQuery)queries.get(subscriptionID);
			
			if (q==null) {
				return;
			}
			
			ServiceInfo sensor=q.service;

			// SOAP headers
			NotificationReceiverHeader receiverHeader=
				new NotificationReceiverHeader();
			receiverHeader.setReceiverName(receiverService.getServiceID());
			receiverHeader.setReceiverNamespace(
					receiverService.getServiceNamespace());

			SoapHeaders headers=new SoapHeaders();

			headers.addHeader(receiverHeader);

			try {

				boolean refreshed=((Boolean)sensor.invokeOperation(
						"subscribe", 
						new String[]{q.queryString,subscriptionID},
						null, 
						headers)).booleanValue();

				q.lastActivityTS=System.currentTimeMillis();

				if (!refreshed) {

					logger.info("Subscribed to sensor '"+sensor.getId()+
							"' running at '"+sensor.getUrl()+"'.");							
				} else {

					logger.debug("Refreshed subscription at sensor '"+
							sensor.getId()+"' running at '"+
							sensor.getUrl()+"'.");
				}

			} catch (Exception e) {
				try {
					logger.debug("Cannot subscribe to sensor '"+
							sensor.getId()+"' running at '"+
							sensor.getUrl()+"': "+e.getMessage());
				} catch (Exception e1) {
					// ignore
				}
			}
		}
	}

	/**
	 * Refreshes the last activity timestamp of a subscription.
	 * 
	 * @param subscriptionID Subscription ID.
	 */
	public void refreshActivity(String subscriptionID) {
		SubscriptionQuery q=(SubscriptionQuery)queries.get(subscriptionID);
		
		if (q==null) {
			return;
		}
		
		q.lastActivityTS=System.currentTimeMillis();
	}
	
	/**
	 * Checks whether the passed ID is known to the client.
	 * 
	 * @param subscriptionID ID.
	 * @return <code>true</code> if known, otherwise <code>false</code>.
	 */
	public boolean hasID(String subscriptionID) {
		return queries.containsKey(subscriptionID);
	}
	
	/**
	 * Checks whether the passed ID is known to the client and throws
	 * an exception if it is unknown.
	 * 
	 * @param subscrioptionID ID.
	 * 
	 * @throws IllegalArgumentException If the ID is unknown.
	 */
	public void checkID(String subscrioptionID) throws IllegalArgumentException{
		if (!queries.containsKey(subscrioptionID)) {
			throw new IllegalArgumentException("Unknown ID");
		}
	}
	
	/**
	 * Performs the unsubscription communication.
	 * 
	 * @param subscriptionID Subscription ID.
	 */
	public void unsubscribeFrom(String subscriptionID) {
		synchronized (getSubscriptionLock(subscriptionID)) {
			
			SubscriptionQuery q=(SubscriptionQuery)queries.get(subscriptionID);
			
			if (q==null) {
				throw new IllegalArgumentException("Unknown/invalid " +
						"subscription ID.");
			}

			serviceDB.unsubscribe(q);
			
			ServiceInfo sensor=q.service;
			
			ids.remove(sensor);
			queries.remove(subscriptionID);
			
			try {
				sensor.invokeOperation("unsubscribe", 
						new Object[]{subscriptionID},
						null);
				logger.info("Unsubscribed from sensor '"+
						sensor.getId()+"' running at '"+
						sensor.getUrl()+"'.");
				
			} catch (InactiveServiceException nse) {
				// everything is fine, ignore
			} catch (Exception e) {
				try {
					logger.debug("Cannot unsubscribe from sensor '"+
							sensor.getId()+"' running at '"+
							sensor.getUrl()+"': "+e.getMessage());
				} catch (Exception e1) {
					// ignore
				}
			}
		}
	}
	
	/**
	 * Query class.
	 * 
	 * @author Lukasz Juszczyk
	 *
	 */
	protected class SubscriptionQuery extends AServiceQuery 
	implements IServiceSubscriber {

		protected String id=null;
		protected ServiceInfo service=null;
		protected String queryString=null;
		protected long lastActivityTS=0;
		
		/**
		 * Constructor.
		 * 
		 * @param id Subscription ID.
		 * @param service Sensor service.
		 * @param queryString Query String.
		 */
		public SubscriptionQuery(String id, ServiceInfo service, 
				String queryString) {
			
			this.id=id;
			this.service=service;
			this.queryString=queryString;
			
			this.skipDuplicates=false;
			
		}
		
		/* (non-Javadoc)
		 * @see at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery#matches(at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
		 */
		protected boolean matches(ServiceInfo info) {			
			return info.equals(service);
		}

		/* (non-Javadoc)
		 * @see at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber#notification(at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery, at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
		 */
		public void notification(AServiceQuery query, ServiceInfo info) {
			
			synchronized (getSubscriptionLock(id)) {
			
				if ((System.currentTimeMillis()-lastActivityTS)>
				maxInactivityTime*1000) {
					
					performSubscription(id);
				}
			}
		}
	}
	
	/**
	 * Returns a lock for a sensor.
	 * 
	 * @param id Subscription ID.
	 * @return Lock object.
	 */
	private static Object getSubscriptionLock(String id) {
		synchronized (locks) {
			Object o=locks.get(id);
			if (o==null) {
				o=new Object();
				locks.put(id, o);
			}
			return o;
		}
	}

}
