/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.processing.notification;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Vector;

import org.xmlpull.v1.XmlPullParserException;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;
import edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext;

import at.ac.tuwien.vitalab.cosine.ContextData;
import at.ac.tuwien.vitalab.cosine.interfaces.IContextHandler;
import at.ac.tuwien.vitalab.cosine.interfaces.ILocalSubscriptionProvider;
import at.ac.tuwien.vitalab.cosine.interfaces.ISubscriptionProviderWS;
import at.ac.tuwien.vitalab.cosine.query.XQuery;

/**
 * Handles context subscriptions.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class SubscriptionHandler implements ISubscriptionProviderWS, 
ILocalSubscriptionProvider, IContextHandler {

	private static Logger logger=
		LoggerFactory.getLogger(SubscriptionHandler.class);

	// key = subscription id
	// value = AContextQuery object
	private Hashtable remoteQueries=new Hashtable();
	
	// key = AContextQuery object
	// value = subscription id
	private Hashtable remoteQueryIDs=new Hashtable();
	
	// values = AContextQuery objects
	private HashSet allQueries=new HashSet();
	
	// key = AContextQuery object
	// value = last context data hash
	private Hashtable lastContextHash=new Hashtable();
	
	private ContextData currentContext=null;
	
	/**
	 * Constructor.
	 * 
	 */
	public SubscriptionHandler() {
		super();
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.processing.notification.RequestHandler#updateContext(at.ac.tuwien.vitalab.cosine.ContextData)
	 */
	public void updateContext(ContextData context) {
		
		if (context==null) {
			return;
		}
				
		synchronized (allQueries) {
			
			currentContext=context;
			
			Iterator queryIter=new HashSet(allQueries).iterator();
			
			while (queryIter.hasNext()) {
				AContextQuery query=(AContextQuery)queryIter.next();
				try {					
					processContext(query,currentContext);
				} catch (InactiveSubscriptionException ise) {
					
					try {
						if (query instanceof RemoteContextQuery) {
							this.unsubscribe((String)remoteQueryIDs.get(query));
						} else {
							this.unsubscribe(query);
						}
						logger.debug("Removing inactive context query.");
						
					} catch (Exception e1) {
						logger.debug("Cannot remove inactive context query: "+
								e1.getMessage());
					}
				} catch (Exception e) {
					logger.debug("Cannot process context query: "+
							e.getMessage());
				}

			}
		}
	}
	
	/**
	 * Processes new Context.
	 * 
	 * @param query Query object.
	 * @param context Context data.
	 * 
	 * @throws Exception If the query cannot be processed.
	 */
	private void processContext(AContextQuery query, ContextData context) 
	throws Exception {
		
		Vector matchingContext=null;
		
		try {
			matchingContext=
				XQuery.returnMatches(context,query.getQueryString());
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		if (matchingContext.isEmpty()) {
			return;
		}
		
		Vector contextList=new Vector();
				
		// duplicate suppression
		Iterator matchIter=matchingContext.iterator();
		while (matchIter.hasNext()) {
			ContextData match=(ContextData)matchIter.next();
			
			if (ContextData.IDNAME.equals(match.getName()) && 
					ContextData.NAMESPACE.equals(match.getNamespace())) {
				continue;
			}
			
			Integer matchHash=new Integer(match.hashCode());
			Integer lastHash=getLastHash(query,match);
			
			if (lastHash==null || !lastHash.equals(matchHash)) {
				
				contextList.add(match);
				setLastHash(query,match);
			}
		}
		
		query.contextQueryNotification(contextList);
	}
	
	/**
	 * Returns the last hash sum seen for this context data.
	 * 
	 * @param query Related query.
	 * @param context Context data.
	 * @return Hash sum or <code>null</code> if none was found.
	 */
	private Integer getLastHash(AContextQuery query, ContextData context) {
		
		HashMap h=(HashMap)lastContextHash.get(query);
		
		if (h==null) {
			return null;
		}
		
		return (Integer)h.get(context.getContextIDElement());
	}

	/**
	 * Sets the last hash sum seen for this context data.
	 * 
	 * @param query Related query.
	 * @param context Context data.
	 */
	private void setLastHash(AContextQuery query, ContextData context) {

		HashMap h=(HashMap)lastContextHash.get(query);
		
		if (h==null) {
			h=new HashMap();
			lastContextHash.put(query, h);
		}
		
		h.put(context.getContextIDElement(), new Integer(context.hashCode()));
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.ILocalSubscriptionProvider#subscribe(at.ac.tuwien.vitalab.cosine.processing.notification.AContextQuery)
	 */
	public boolean subscribe(AContextQuery query) throws Exception {
		
		synchronized (allQueries) {

			if (allQueries.contains(query)) {
				return true;
			}
			
			allQueries.add(query);
			
			// in order to process notifications immediately
			if (currentContext!=null) {
				processContext(query, currentContext);					
			}
		}

		return false;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.ILocalSubscriptionProvider#unsubscribe(at.ac.tuwien.vitalab.cosine.processing.notification.AContextQuery)
	 */
	public void unsubscribe(AContextQuery query) throws Exception {
		
		synchronized (allQueries) {
			
			if (!allQueries.contains(query)) {
				throw new IllegalArgumentException("Query is not subscribed.");
			}
			
			query.stop();
			
			allQueries.remove(query);			
			lastContextHash.remove(query);
		}
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.ISubscriptionProviderWS#subscribe(java.lang.String, java.lang.String, edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext)
	 */
	public boolean subscribe(String query, String id, 
			InvocationContext context) throws Exception {
		
		try {
			
			synchronized (allQueries) {
				
				if (remoteQueries.containsKey(id)) {
					return true;
				}
				
				NotificationReceiverHeader receiverHeader=
					new NotificationReceiverHeader(
							context.incomingChannel.getIncomingXMLHeaders());
				
				RemoteContextQuery remoteQuery=
					new RemoteContextQuery(query,id,receiverHeader);
								
				remoteQueries.put(id, remoteQuery);
				remoteQueryIDs.put(remoteQuery, id);
				
				this.subscribe(remoteQuery);
			}
			
			return false;
			
		} catch (NoSuchElementException nse) {
			throw new Exception("ContextNotificationReceiver-header missing.");
		} catch (XmlPullParserException xppe) {
			throw new Exception("ContextNotificationReceiver-header error: "+
					xppe.getMessage());
		}
	}

	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.ISubscriptionProviderWS#unsubscribe(java.lang.String)
	 */
	public void unsubscribe(String id) throws Exception {
		
		synchronized (allQueries) {
			
			AContextQuery remoteQuery=(AContextQuery)remoteQueries.remove(id);
			
			if (remoteQuery==null) {
				throw new IllegalArgumentException("Unknown subscription ID.");
			}

			this.unsubscribe(remoteQuery);
			
			remoteQueryIDs.remove(remoteQuery);
		}
	}
}