/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.webservice;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import org.kxml2.kdom.Element;

import at.ac.tuwien.vitalab.cosine.ContextData;
import at.ac.tuwien.vitalab.cosine.XMLData;
import at.ac.tuwien.vitalab.cosine.configuration.CIMSConfiguration;
import at.ac.tuwien.vitalab.cosine.interfaces.INotificationReceiverWS;
import at.ac.tuwien.vitalab.cosine.processing.notification.SubscriptionClient;
import at.ac.tuwien.vitalab.middleware.Middleware;
import at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery;
import at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceDB;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo;

/**
 * Abstract context aggregator Web service.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class AContextAggregationWebService extends ASensorWebService 
implements IServiceSubscriber, INotificationReceiverWS {

	private ServiceDB serviceDB=Middleware.getServiceDatabase();
	
	private ContextSensorQuery query=new ContextSensorQuery();
	private SubscriptionClient subscriptionClient=new SubscriptionClient(this);
	
	private Hashtable uuidContextMap=new Hashtable();
	
	private ContextData contextData = new ContextData(
			this.getContextTypeName(), 
			this.getContextTypeNamespace()); 
	
	private long lastUpdateTS=0;
	private long contextChangeTS=0;
	
	private Timer timer;
	
	private int maxDelay=CIMSConfiguration.getMaxNotificationDelay();
	
	/**
	 * Constructor.
	 * 
	 * @param contextTypeName Name of the context type provided.
	 * @param contextTypeNamespace Namespace of the context type provided.
	 * @param supportsQuery Whether it supports querying.
	 * @param supportsSubscription Whether it supports subscriptions.
	 */
	protected AContextAggregationWebService(String contextTypeName, 
			String contextTypeNamespace, boolean supportsQuery, 
			boolean supportsSubscription) {
		
		super(contextTypeName, contextTypeNamespace, supportsQuery,
				supportsSubscription);
		
		operations.add("contextListNotification");
		operations.add("contextNotification");
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getConfigurationRessourceName()
	 */
	public String getConfigurationRessourceName() {
		return null;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#onDeploy()
	 */
	public void onDeploy() {
		super.onDeploy();
		timer=new Timer();
		timer.scheduleAtFixedRate(
				new UpdateTask(), 
				new Random().nextInt(1000*maxDelay), 
				1000);
		serviceDB.subscribe(query, this);				
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#onUndeploy()
	 */
	public void onUndeploy() {
		serviceDB.unsubscribe(query);
		timer.cancel();
		super.onUndeploy();
	}

	private class UpdateTask extends TimerTask {

		
		/* (non-Javadoc)
		 * @see java.util.TimerTask#run()
		 */
		public void run() {
						
			synchronized (contextData) {
				
				if (contextChangeTS!=0 &&
						maxDelay<=Math.round(((double)
						(System.currentTimeMillis()-lastUpdateTS))/1000)) {
					
					updateContext(contextData,contextChangeTS);
					
					lastUpdateTS=System.currentTimeMillis();
					
					contextChangeTS=0;
				}
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber#notification(at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery, at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
	 */
	public void notification(AServiceQuery query, ServiceInfo info) {
		// subscribe to all data
		subscriptionClient.subscribeTo(info, "");
	}

	private class ContextSensorQuery extends AServiceQuery {

		/* (non-Javadoc)
		 * @see at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery#matches(at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
		 */
		protected boolean matches(ServiceInfo info) {
			
			String myUUID=null;
			
			if (contextID!=null) {
				
				for (int i=0;i<contextID.getAttributeCount();i++) {
					String attName=contextID.getAttributeName(i);
					if ("uuid".endsWith(attName)) {
						myUUID=contextID.getAttributeValue(i);
						break;
					}
				}
			}

			try {				
				if (info.getAttribute(WSATTR_NAME_SENSOR).equals(
						WSATTR_VALUE_SENSOR) &&
						info.getAttribute("subscription").equalsIgnoreCase(
						"supported")) {

					Element idElement=new XMLData(
							info.getAttribute(ContextData.IDNAME));
					
					for (int i=0;i<idElement.getAttributeCount();i++) {
						String attName=idElement.getAttributeName(i);
						if (attName.equals("belongsTo")) {
							String attValue=idElement.getAttributeValue(i);
							if (myUUID.equals(attValue)) {
								return true;
							}
						}
					}
				}
			} catch (Exception e) {
			}
			
			return false;
		}
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.INotificationReceiverWS#contextListNotification(java.lang.String, java.util.Vector)
	 */
	public synchronized void contextListNotification(String id, 
			Vector contextList) throws Exception {

		subscriptionClient.checkID(id);
		
		subscriptionClient.refreshActivity(id);

		Iterator contextIter=contextList.iterator();
		
		while (contextIter.hasNext()) {
			processContext((ContextData)contextIter.next());
		}
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.INotificationReceiverWS#contextNotification(java.lang.String, at.ac.tuwien.vitalab.cosine.ContextData)
	 */
	public synchronized void contextNotification(String id, ContextData context) 
	throws Exception {

		subscriptionClient.checkID(id);
		
		subscriptionClient.refreshActivity(id);

		processContext(context);
	}
	
	/**
	 * Process new incoming context.
	 * 
	 * @param context Context data.
	 */
	protected void processContext(ContextData context) {
		
		synchronized (contextData) {
			
			XMLData contextID=context.getContextIDElement();

			if (contextID==null) {
				return;
			}
			
			String tsStr=null;		
			long ts=0;
			
			tsStr=context.getAttributeByName("contextTimestamp");
			ts=tsStr==null ? 0 : Long.valueOf(tsStr).longValue();
					
			ContextData oldContext=(ContextData)uuidContextMap.get(contextID);
			
			if (oldContext!=null) {
				for (int i=0;i<contextData.getChildCount();i++) {
					ContextData tmp=(ContextData)contextData.getChild(i);
					if (tmp.equals(oldContext)) {
						
						if (tsStr!=null) {

							String oldTsStr=
								tmp.getAttributeByName("contextTimestamp");
							
							if (oldTsStr!=null) {
								
								long oldTs=Long.valueOf(oldTsStr).longValue();

								if (ts<=oldTs) {
									return;
								}
							}
						}
						contextData.removeChild(i);
						break;
					}
				}
			} 
			
			contextData.addChild(Element.ELEMENT, context);
			uuidContextMap.put(contextID, context);
			
			// mark context as changed
			contextChangeTS=System.currentTimeMillis();
		}
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.INotificationReceiverWS#registerSubscriptionClient(java.lang.String, at.ac.tuwien.vitalab.cosine.processing.notification.SubscriptionClient)
	 */
	public void registerSubscriptionClient(String id, SubscriptionClient sc) {
		// not necessary
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.INotificationReceiverWS#unregisterSubscriptionClient(java.lang.String)
	 */
	public void unregisterSubscriptionClient(String id) {
		// not necessary
	}

}
