/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.webservice;

import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.kxml2.kdom.Element;

import edu.wustl.mobilab.sliver.soap.Serialization;
import edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext;

import at.ac.tuwien.vitalab.cosine.CMS;
import at.ac.tuwien.vitalab.cosine.ContextData;
import at.ac.tuwien.vitalab.cosine.XMLData;
import at.ac.tuwien.vitalab.cosine.interfaces.IDataProviderWS;
import at.ac.tuwien.vitalab.cosine.processing.notification.SubscriptionHandler;
import at.ac.tuwien.vitalab.cosine.processing.storage.QueryHandler;
import at.ac.tuwien.vitalab.middleware.AWebService;
import at.ac.tuwien.vitalab.middleware.Middleware;

/**
 * Abstract sensor Web service.
 * 
 * @author Lukasz Juszczyk
 *
 */
public abstract class ASensorWebService extends AWebService 
implements IDataProviderWS {

	protected SubscriptionHandler subscriptionHandler=null;
	protected QueryHandler queryHandler=null;

	protected boolean supportsQuery=true;
	protected boolean supportsSubscription=true;
	
	public static final String WSATTR_NAME_SENSOR="sensorservice";
	public static final String WSATTR_NAME_CONTEXTNS="contextnamespace";
	public static final String WSATTR_NAME_CONTEXTID=ContextData.IDNAME;
	public static final String WSATTR_VALUE_SENSOR="true";
	public static final String WSATTR_NAME_QUERY="query";
	public static final String WSATTR_NAME_SUBSCRIPTION="subscription";
	
	protected String contextTypeName=null;
	protected String contextTypeNamespace=null;
	
	protected XMLData contextID=null;
	
	protected Vector operations=new Vector();
	
	static {
		new MarshalContextData().register(Serialization.getParentEnvelope());
	}

	/**
	 * Constructor.
	 * 
	 * @param contextTypeName Name of the context type provided.
	 * @param contextTypeNamespace Namespace of the context type provided.
	 * @param supportsQuery Whether it supports querying.
	 * @param supportsSubscription Whether it supports subscriptions.
	 */
	protected ASensorWebService(
			String contextTypeName, String contextTypeNamespace, 
			boolean supportsQuery, boolean supportsSubscription) {
		super(contextTypeName,CMS.WSNAMESPACE+contextTypeName);
		
		this.contextTypeName=contextTypeName;
		this.contextTypeNamespace=contextTypeNamespace;
		
		if (!(supportsQuery || supportsSubscription)) {
			throw new IllegalArgumentException(
					"At least one interaction type must be chosen.");
		}
		
		this.supportsQuery=supportsQuery;
		if (supportsQuery) {
			queryHandler=new QueryHandler();
			operations.add("getContextInfo");
		}
		
		this.supportsSubscription=supportsSubscription;
		if (supportsSubscription) {
			subscriptionHandler=new SubscriptionHandler();
			operations.add("getConfiguration");
			operations.add("setConfiguration");	
			operations.add("subscribe");
			operations.add("unsubscribe");			
		}
	}

	/**
	 * Deploys a sensor service.
	 * 
	 * @param contextID ContextID element.
	 */
	private void deploy(XMLData contextID) {

		if (contextID==null) {
			throw new NullPointerException("Context ID is null.");
		}
		
		if (!contextID.getName().equals(ContextData.IDNAME) || 
				!contextID.getNamespace().equals(ContextData.NAMESPACE)) {
			throw new IllegalArgumentException(
					"Not a valid contextID element.");
		}
		
		this.contextID=contextID;

		updateContext(new ContextData(contextTypeName,contextTypeNamespace));

		Middleware.deployService(this);
	}
	
	/**
	 * Deploys a sensor service.
	 * 
	 * @param idAttributes ContextID attributes.
	 */
	public void deploy(Hashtable idAttributes) {
		XMLData id=new XMLData(ContextData.IDNAME,ContextData.NAMESPACE);
		
		Enumeration names=idAttributes.keys();
		
		while (names.hasMoreElements()) {
			String name=names.nextElement().toString();
			String value=idAttributes.get(name).toString();
			id.setAttribute(null, name, value);
		}
		
		deploy(id);
	}
	
	/**
	 * Returns the context type name.
	 * 
	 * @return Name.
	 */
	public String getContextTypeName() {
		return this.contextTypeName;
	}

	/**
	 * Returns the context type namespace.
	 * 
	 * @return Namespace.
	 */
	public String getContextTypeNamespace() {
		return this.contextTypeNamespace;
	}

	/**
	 * Returns the uuid of this sensor.
	 * 
	 * @return UUID.
	 */
	public String getUUID() {
		return contextID.getAttributeByName("uuid");
	}

	/**
	 * Returns the uuid of the sensors parent.
	 * 
	 * @return UUID.
	 */
	public String getBelongsTo() {
		return contextID.getAttributeByName("belongsTo");
	}

	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getAttributes()
	 */
	public Hashtable getAttributes() {
		Hashtable result=new Hashtable();
		
		result.put(WSATTR_NAME_SENSOR, WSATTR_VALUE_SENSOR);
		result.put(WSATTR_NAME_CONTEXTNS, this.contextTypeNamespace);
		result.put(WSATTR_NAME_CONTEXTID, this.contextID.toString());			
		
		if (supportsQuery) {
			result.put(WSATTR_NAME_QUERY, "supported");
		}
		
		if (supportsSubscription) {
			result.put(WSATTR_NAME_SUBSCRIPTION, "supported");
		}
		return result;
	}

	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getOperationNames()
	 */
	public Collection getOperationNames() {
		return new Vector(operations);
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getOperationDescription(java.lang.String)
	 */
	public Hashtable getOperationDescription(String operationName) {
		if (operations.contains(operationName)) {
			return new Hashtable();
		} 
		throw new IllegalArgumentException("Invalid/unknown operation.");		
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.ISubscriptionProviderWS#subscribe(java.lang.String, java.lang.String, edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext)
	 */
	public boolean subscribe(String query, String id, InvocationContext context) 
	throws Exception {
		if (!supportsSubscription) {
			throw new UnsupportedOperationException(
					"Subscription not supported.");
		}
		return subscriptionHandler.subscribe(query,id,context);
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.ISubscriptionProviderWS#unsubscribe(java.lang.String)
	 */
	public void unsubscribe(String id) throws Exception {
		if (!supportsSubscription) {
			throw new UnsupportedOperationException(
					"Subscription not supported.");
		}
		subscriptionHandler.unsubscribe(id);
	}

	/*
	 * (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.IQueryProviderWS#getContextInfo(java.lang.String)
	 */
	public Vector getContextInfo(String query) throws Exception {
		if (!supportsQuery) {
			throw new UnsupportedOperationException(
					"Query not supported.");
		}
		return queryHandler.getContextInfo(query);
	}
	
	/**
	 * Notifies about changed context.
	 * 
	 * @param context Context data.
	 */
	protected void updateContext(ContextData context) {
		updateContext(context,System.currentTimeMillis());
	}
	
	/**
	 * Notifies about changed context.
	 * 
	 * @param context Context data.
	 * @param timestamp Context timestamp.
	 */
	protected void updateContext(ContextData context, long timestamp) {
		
		int indexOfID=context.indexOf(
				ContextData.NAMESPACE, 
				ContextData.IDNAME, 
				0);
		
		if (indexOfID!=-1) {
			context.removeChild(indexOfID);
		}
		
		context.addChild(Element.ELEMENT, contextID);			
		
		if (timestamp>0) {
			context.setAttribute(null, "contextTimestamp", ""+timestamp);			
		}
		
		context.refreshSendingTimestamp();
		
		if (supportsSubscription) {
			subscriptionHandler.updateContext(context);			
		}
		if (supportsQuery) {
			queryHandler.updateContext(context);			
		}
	}
}
