/*
    This file is part of COSINE.

    COSINE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    COSINE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with COSINE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.cosine.webservice;

import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.kxml2.kdom.Element;
import org.kxml2.kdom.Node;

import edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext;

import at.ac.tuwien.vitalab.cosine.ContextData;
import at.ac.tuwien.vitalab.cosine.XMLData;
import at.ac.tuwien.vitalab.cosine.interfaces.IUpdateReceiverWS;

/**
 * Updatable sensor Web service.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class UpdatableSensorWebService extends ASensorWebService 
implements IUpdateReceiverWS{

	private boolean appending=false;
	
	private final ContextData parentElement= 
		new ContextData("TMP",ContextData.NAMESPACE);
	
	/**
	 * Constructor.
	 * 
	 * @param contextTypeName Name of the context type provided.
	 * @param contextTypeNamespace Namespace of the context type provided.
	 * @param supportsQuery Whether it supports querying.
	 * @param supportsSubscription Whether it supports subscriptions.
	 * @param appending Whether context data should be appended or overwritten.
	 */
	public UpdatableSensorWebService(String contextTypeName,
			String contextTypeNamespace, boolean supportsQuery,
			boolean supportsSubscription, boolean appending) {
		super(contextTypeName, contextTypeNamespace, supportsQuery,
				supportsSubscription);
		
		this.appending=appending;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getConfigurationRessourceName()
	 */
	public String getConfigurationRessourceName() {
		return null;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.webservice.ASensorWebService#getOperationNames()
	 */
	public Collection getOperationNames() {
		Vector names=new Vector(super.getOperationNames());
		names.add("storeContextInfo");
		return names;
	}

	/*
	 * (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.interfaces.IUpdateReceiverWS#storeContextInfo(at.ac.tuwien.vitalab.cosine.ContextData)
	 */
	public void storeContextInfo(ContextData context) 
	throws Exception {
		if (appending) {
			
			XMLData newContextID=(XMLData)(contextID.clone());
			
			Hashtable attributes=context.getContextIDAttributes();
			
			if (attributes==null) {
				throw new IllegalArgumentException("No contextID.");
			}
			
			int indexOfID=context.indexOf(
					ContextData.NAMESPACE, 
					ContextData.IDNAME, 
					0);
			
			if (indexOfID!=-1) {
				context.removeChild(indexOfID);
			}
							
			Enumeration nameEnum=attributes.keys();
			while (nameEnum.hasMoreElements()) {
				
				String name=nameEnum.nextElement().toString();
				
				if (name.equals("xmlns") || 
						name.startsWith("xmlns:") ||
						name.equals("uuid")) {
					continue;
				}
				
				String value=attributes.get(name).toString();
				
				newContextID.setAttribute(null, name, value);
			}
			
			context.addChild(Element.ELEMENT, newContextID);			

			context.setAttribute(null, "contextTimestamp", 
					""+System.currentTimeMillis());			
			
			context.refreshSendingTimestamp();
			
			parentElement.deleteChildWithContextID(newContextID);
			
			parentElement.addChild(Node.ELEMENT, context);

			if (supportsSubscription) {
				subscriptionHandler.updateContext(parentElement);			
			}
			if (supportsQuery) {
				queryHandler.updateContext(parentElement);			
			}
			
		} else {
			super.updateContext(context);			
		}

	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.webservice.ASensorWebService#getContextInfo(java.lang.String)
	 */
	public Vector getContextInfo(String query) throws Exception {
		
		if (appending) {
			query=query.trim();
			if (query.length()!=0) {
				return super.getContextInfo("/"+parentElement.getName()+query);				
			} else {
				return super.getContextInfo("/"+parentElement.getName()+"/*");
			}
		} else {
			return super.getContextInfo(query);
		}
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.cosine.webservice.ASensorWebService#subscribe(java.lang.String, java.lang.String, edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext)
	 */
	public boolean subscribe(String query, String id, InvocationContext context)
			throws Exception {
		
		if (appending) {
			query=query.trim();
			if (query.length()!=0) {
				return super.subscribe(
						"/"+parentElement.getName()+query, id, context);			
			} else {
				return super.subscribe(
						"/"+parentElement.getName()+"/*", id, context);			
			}
		} else {
			return super.subscribe(query, id, context);
		}
	}
}
