/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Properties;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;



/**
 * Web Service which is able to deliver some description about itself.
 * 
 * @author Lukasz Juszczyk
 *
 */
public abstract class AWebService {

	protected Properties configuration=new Properties();
	
	private ClassLoader loader=null;
	
	protected Logger logger=LoggerFactory.getLogger(this.getClass());
	
	protected boolean published=true;
	
	private String serviceID;
	private String serviceNamespace;
	
	/**
	 * Constructor.
	 * 
	 * @param serviceID Service name/ID. 
	 * @param serviceNamespace Service namespace.
	 *
	 */
	public AWebService(String serviceID, String serviceNamespace) {
		
		if (serviceID==null || serviceNamespace==null) {
			throw new NullPointerException("Service name and namespace " +
					"cannot be null.");
		}
		
		this.serviceID=serviceID;
		this.serviceNamespace=serviceNamespace;
	}
	
	/**
	 * Returns the name/ID of this Web service.
	 * 
	 * @return Service ID.
	 */
	public final String getServiceID() {
		return serviceID;
	}
	
	/**
	 * Returns the namespace of this Web service.
	 * 
	 * @return Namespace.
	 */
	public final String getServiceNamespace() {
		return serviceNamespace;
	}
	
	/**
	 * Returns the attributes/description of this Web service.
	 * 
	 * @return Hashtable with attributes.
	 */
	public abstract Hashtable getAttributes();
	
	/**
	 * Hook, invoked after deployment of Web service.
	 *
	 */
	public void onDeploy() {
		// empty
	}
	
	/**
	 * Hook, invoked before undeployment of Web service.
	 *
	 */
	public void onUndeploy() {
		// empty
	}
	
	/**
	 * Returns a collection of operation names which should be made
	 * public via the Web service.
	 * 
	 * @return Collection of operation names.
	 */
	public abstract Collection getOperationNames();

	/**
	 * Returns a mapping of operation names. This way it is possible to
	 * assign names of operations which do not really exist inside the 
	 * classes. Useful for automatically generated Web services.
	 * 
	 * @return Operation mapping.
	 */
	public Hashtable getOperationNameMapping() {
		return new Hashtable();
	}
	
	/**
	 * Returns the description attributes of a particular operation.
	 * 
	 * @param operationName Operation name.
	 * 
	 * @return Attributes.
	 */
	public abstract Hashtable getOperationDescription(String operationName);
	
	/**
	 * Returns the ressource name of the configuration file. 
	 * 
	 * @return Ressource name.
	 */
	public abstract String getConfigurationRessourceName();
		
	/**
	 * Sets the class laoder if the service was read from a Jar file.
	 * 
	 * @param loader Class laoder.
	 */
	public final void setClassLoader(ClassLoader loader) {
		this.loader=loader;
	}
	
	/**
	 * Reads in the configuration of the service.
	 * 
	 * @return <code>true</code> if the configuration could be found,
	 * otherwise <code>false</code>.
	 * 
	 * @throws FileNotFoundException If the config file cannot be found. 
	 * @throws IOException If IO errors occur while reading.
	 */
	protected final boolean readConfiguration() 
	throws FileNotFoundException, IOException {
		
		String ressourceName=getConfigurationRessourceName();
		
		if (ressourceName==null) {
			return false;
		}
		
		if (loader==null) {
			loader=Thread.currentThread().getContextClassLoader();
		}

		File configFile=new File(ressourceName);

		InputStream configStream=null;

		if (configFile.exists() && configFile.canRead() && configFile.isFile()){
			configStream=new FileInputStream(configFile);
			
		} else {
			configStream=loader.getResourceAsStream(ressourceName);
		}
		
		
		if (configStream==null) {
			logger.warning("Cannot read configuration '"+ressourceName+
					"' of service '"+getServiceID()+".'");
			return false;
		}
		
		configuration.load(configStream);
		configStream.close();
		
		return true;
	}
	
	/**
	 * Whether this Web service should be published in the network.
	 * 
	 * @return <code>true</code> if it should be published, otherwise 
	 *  <code>false</code>.
	 */
	public boolean isPublished() {
		return published;
	}

	/**
	 * Sets whether this Web service should be published in the network.
	 * 
	 * @param published <code>true</code> if it should be published, otherwise 
	 *  <code>false</code>.
	 */
	public void setPublished(boolean published) {
		this.published = published;
	}		
}
