/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.discovery;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;


import at.ac.tuwien.vitalab.middleware.AWebService;
import at.ac.tuwien.vitalab.middleware.MiddlewareConfiguration;
import at.ac.tuwien.vitalab.middleware.soap.services.InfoService;

/**
 * Service Database. Contains all services found in the network.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class ServiceDB {

	// key = host uuid
	// value = string containing the serial number and instance id
	private Hashtable hostStatus=new Hashtable();
	
	// key = host uuid
	// value = interpolated hop value
	private Hashtable hops=new Hashtable();
	
	// currently running status updates from hosts are placed into this set
	// values = host uuids
	private HashSet currentHostUpdates=new HashSet();
	
	// key = host uuid
	// value = set of service namespaces
	private Hashtable hostedServices=new Hashtable();
	
	// all services, remote and local
	// key = key string containing service name and host uuid
	// value = ServiceInfo instance
	private Map services=Collections.synchronizedMap(new Hashtable());

	// only local services
	// key = namespace
	// value = ServiceInfo instance
	private Hashtable localServices=new Hashtable();
	
	// only local services which have already been undeployed
	// key = namespace
	// value = ServiceInfo instance
	private Hashtable oldLocalServices=new Hashtable();
	
	// key = query object
	// value = subscriber
	private Map subscriptions=Collections.synchronizedMap(new Hashtable());
	
	private static Logger logger=LoggerFactory.getLogger(ServiceDB.class);

	/**
	 * Update all services of a given host.
	 * 
	 * @param curInstanceID Current instance ID.
	 * @param curSerial Current serial number of the host. 
	 * @param uuid Host uuid.
	 * @param url Host url.
	 * @param updateString Service update String, or <code>null</code> if no
	 * update has been sent.
	 */
	protected void updateFromHost(final String curInstanceID, 
			final String curSerial, final String uuid, final String url,
			final String updateString) {

		synchronized (currentHostUpdates) {
			if (currentHostUpdates.contains(uuid)) {
				return;
			}
			currentHostUpdates.add(uuid);			
		}
		
		try {
			// do we know the last status? 
			String statusStr=curInstanceID+" "+curSerial;		
			String oldStatusStr=(String)hostStatus.get(uuid);

			String oldInstanceID=null;
			String oldSerial=null;
			
			boolean changedInstance=true;
			
			if (oldStatusStr!=null) {				
			
				StringTokenizer tokens=new StringTokenizer(oldStatusStr);
				
				oldInstanceID=tokens.nextToken();
				oldSerial=tokens.nextToken();
				
				changedInstance=!oldInstanceID.equals(curInstanceID);
				
				if (oldSerial.equals(curSerial) && 
						oldInstanceID.equals(curInstanceID)) {
					
					// nothing changed? refresh the services
					HashSet hostedSet=((HashSet)hostedServices.get(uuid));
					
					if (hostedSet!=null) {
						Iterator hostedIter=hostedSet.iterator();
						
						while (hostedIter.hasNext()) {
							ServiceInfo h=(ServiceInfo)hostedIter.next();
							
							h.updateUrl(url);
							h.refresh();
						}

						hostedIter=hostedSet.iterator();
						
						while (hostedIter.hasNext()) {
							ServiceInfo h=(ServiceInfo)hostedIter.next();
							
							checkServiceForSubscriptions(h);							
						}
					}

					return;	
				}			
			}
			
			// otherwise save the last status string and checks what has changed
			hostStatus.put(uuid, statusStr);
			
			String fromSerial=changedInstance ? "0" : oldSerial;

			// create/get reference to the info service
			ServiceInfo info;

			synchronized (services) {
				
				String infoKey=tableKey(InfoService.NAMESPACE,uuid);
				info=(ServiceInfo)services.get(infoKey);

				if (info==null) {
					info=new ServiceInfo(url,uuid,this);
					services.put(infoKey, info);			
				} else {
					info.updateUrl(url);
					info.refresh();
				}						
			}

			// notify subscribers
			checkServiceForSubscriptions(info);

			Collection addedServicesNS=new HashSet();
			Collection removedServicesNS=new HashSet();

			try {
				int from=Integer.parseInt(fromSerial);
				int to=Integer.parseInt(curSerial);
				
				if (from==to-1 && updateString!=null) {
					
					char plusOrMinus=updateString.charAt(0);
					String updateUrl=updateString.substring(1);
					
					switch (plusOrMinus) {
					case '+':
						addedServicesNS.add(updateUrl);
						break;
					case '-':
						removedServicesNS.add(updateUrl);
						break;
					default:
					}
				}
			} catch (Exception nfe) {
				// if update string cannot be parsed, retrieve all data
			}
			
			if (addedServicesNS.isEmpty() && removedServicesNS.isEmpty()) {
				// try via udp
				try {
					ServiceRegistry.retrieveChangesViaUDP(
							info, fromSerial, curSerial,
							addedServicesNS, removedServicesNS);

				} catch (Exception e1) {
					logger.debug("Cannot retrieve status via udp from "+
							info+": "+e1.getMessage());

					// alternatively try via a soap call
					try {
						Vector result=(Vector)info.invokeOperation(
								"getStatusChanges", 
								new Object[]{
									new Integer(fromSerial),
									new Integer(curSerial)}, 
									null);

						addedServicesNS=(Vector)result.get(0);
						removedServicesNS=(Vector)result.get(1);

					} catch (Exception e) {
						logger.debug("Cannot retrieve status via WS from "+
								info+": "+e.getMessage());
						
						return;
					}
				}
			}

			HashSet oldServiceSet=((HashSet)hostedServices.get(uuid));
			HashSet newServiceSet=new HashSet();

			// refresh old services and deactivate removed ones
			if (oldServiceSet!=null) {
				Iterator serviceIter=oldServiceSet.iterator();
				while (serviceIter.hasNext()) {
					ServiceInfo service=(ServiceInfo)serviceIter.next();

					try {
						String serviceNS=service.getNamespace();

						if (removedServicesNS.contains(serviceNS)) {

							service.setInactive();

						} else {

							if (changedInstance && 
									addedServicesNS.contains(serviceNS)) {
								addedServicesNS.remove(serviceNS);
							}

							service.updateUrl(url);
							service.refresh();

							// notify subscribers
							checkServiceForSubscriptions(service);

							newServiceSet.add(service);
						}

					} catch (MissingServiceAttributeException e) {
						// should not happen
					}
				}
			}

			// create new services
			Iterator nsIter=addedServicesNS.iterator();

			while (nsIter.hasNext()) {

				String ns=nsIter.next().toString();

				if (ns.equals(InfoService.NAMESPACE)) {
					continue;
				}
				
				String serviceKey=tableKey(ns,uuid);
				
				ServiceInfo service=(ServiceInfo)services.get(serviceKey);
				if (service==null) {
					service=new ServiceInfo(info,ns,uuid,this);					
				}

				services.put(serviceKey, service);							

				// notify subscribers
				checkServiceForSubscriptions(service);

				newServiceSet.add(service);
			}

			newServiceSet.add(info);
			
			hostedServices.put(uuid, newServiceSet);
	
		} finally {
			currentHostUpdates.remove(uuid);
		}		
	}

	/**
	 * Returns the number of hops to the host with the passed UUID.
	 * 
	 * @param hostUUID Host UUID.
	 * @return Number of hops.
	 * @throws IllegalArgumentException If the UUID is unknown.
	 */
	protected int getHopsToHost(String hostUUID) 
	throws IllegalArgumentException {
		Float f=(Float)hops.get(hostUUID);
		if (f==null) {
			throw new IllegalArgumentException("Unknown host UUID.");
		}
		return Math.round(f.floatValue());
	}
	
	/**
	 * Adds a local service to the database.
	 * 
	 * @param ws Web service.
	 */
	protected void addLocalService(AWebService ws) {
		
		synchronized (localServices) {
			
			if (localServices.containsKey(ws.getServiceNamespace())) {
				return;
			}
			
			String uuid=MiddlewareConfiguration.getHostUUID();
			String key=tableKey(ws.getServiceNamespace(),uuid);
			
			ServiceInfo info=(ServiceInfo)oldLocalServices.get(
					ws.getServiceNamespace());
			
			if (info==null) {
				info=new ServiceInfo(ws,this);				
			} else {
				info.refresh();
			}
			
			services.put(key, info);
			localServices.put(ws.getServiceNamespace(), info);
			oldLocalServices.remove(ws.getServiceNamespace());
			
			// notify subscribers
			checkServiceForSubscriptions(info);
		}
	}
	
	/**
	 * Delete a local service from the database.
	 * 
	 * @param namespace Service namespace.
	 * @return Whether it existed.
	 */
	protected boolean deleteLocalService(String namespace) {
		
		synchronized (localServices) {
			
			ServiceInfo removedService=
				(ServiceInfo)localServices.remove(namespace);
			
			if (removedService!=null) {
				
				String uuid=MiddlewareConfiguration.getHostUUID();
				String key=tableKey(namespace,uuid);
				
				ServiceInfo old=(ServiceInfo)services.remove(key);
				if (old!=null) {
					old.setInactive();					
					oldLocalServices.put(namespace,old);
				}
				return true;
			}
			return false;
		}
	}
	
	/**
	 * Delete a local service from the database.
	 * 
	 * @param ws Web service.
	 * @return Whether it existed.
	 */
	protected boolean deleteLocalService(AWebService ws) {
		
		synchronized (localServices) {
			
			if (!localServices.containsKey(ws.getServiceNamespace())) {
				return false;
			}
			
			String uuid=MiddlewareConfiguration.getHostUUID();
			String key=tableKey(ws.getServiceNamespace(),uuid);
						
			ServiceInfo old=(ServiceInfo)services.remove(key);
			
			if (old!=null) {
				old.setInactive();				
				oldLocalServices.put(ws.getServiceNamespace(),old);
			}
			return localServices.remove(ws.getServiceNamespace())!=null;
		}
	}
	
	/**
	 * Checks whether the service exists.
	 * 
	 * @param namespace Service namespace.
	 * @return <code>true</code> if it exists, otherwise <code>false</code>.
	 */
	protected boolean hasLocalService(String namespace) {
		return localServices.containsKey(namespace);
	}

	/**
	 * Checks whether the service exists.
	 * 
	 * @param ws Web service.
	 * @return <code>true</code> if it exists, otherwise <code>false</code>.
	 */
	protected boolean hasLocalService(AWebService ws) {
		return localServices.containsKey(ws.getServiceNamespace());
	}
	
	/**
	 * Creates a unique key for the hashtable.
	 * 
	 * @param ns Namespace.
	 * @param id UUID.
	 * @return Key String.
	 */
	private String tableKey(String ns, String id) {
		return ns+" @ "+id;
	}
	
	/**
	 * Gets all services.
	 * 
	 * @return Services.
	 */
	public HashSet getServices() {
		
		HashSet result=new HashSet();
		Collection serviceSet=new HashSet();
		synchronized (services) {
			serviceSet.addAll(services.values());			
		}
			
		for (Iterator serviceIter=serviceSet.iterator();
		serviceIter.hasNext();) {
				
			ServiceInfo info=(ServiceInfo)serviceIter.next();
			if (!info.isInactive()) {
				result.add(info);
			}
		}
		
		return result;			
	}
	
	/**
	 * Gets all services, even the inactive ones.
	 * 
	 * @return Services.
	 */
	protected HashSet getAllServices() {
		HashSet result=new HashSet();
		synchronized (services) {
			result.addAll(services.values());			
		}
		result.addAll(oldLocalServices.values());
		return result;
	}
	
	/**
	 * Gets all services which match the following query.
	 * 
	 * @param query AServiceQuery.
	 * @return Hashset with resulting services.
	 */
	public HashSet getServices(AServiceQuery query) {

		HashSet result=new HashSet();
		Collection serviceSet=new HashSet();
		synchronized (services) {
			serviceSet.addAll(services.values());			
		}
		
		for (Iterator serviceIter=serviceSet.iterator();
		serviceIter.hasNext();) {
			
			ServiceInfo info=(ServiceInfo)serviceIter.next();
			
			if (!info.isInactive() && query.matches(info)) {
				result.add(info);
			}
		}
		
		return result;
	}
	
	/**
	 * Subscribe to monitor the changes of a service.
	 * 
	 * @param query AServiceQuery object.
	 * @param subscriber IServiceSubscriber.
	 */
	public void subscribe(AServiceQuery query, IServiceSubscriber subscriber) {
		
		subscriptions.put(query,subscriber);
			
		checkSubscription(query);
	}
	
	/**
	 * Checks a subscription query.
	 * 
	 * @param query Query.
	 */
	public void checkSubscription(AServiceQuery query) {
		
		IServiceSubscriber subscriber=
			(IServiceSubscriber)subscriptions.get(query);
		
		if (subscriber==null) {
			return;
		}
		
		Iterator serviceIter=null;
		synchronized (services) {
			serviceIter=new HashSet(services.values()).iterator();
		}
		while (serviceIter.hasNext()) {
			ServiceInfo service=(ServiceInfo)serviceIter.next();

			runQuery(query,service);
		}	
	}
	
	
	/**
	 * Runs a query on a service.
	 * 
	 * @param query Query.
	 * @param service Service.
	 */
	private void runQuery(AServiceQuery query, ServiceInfo service) {

		if (service.isInactive()) {
			return;
		}
		
		if (!query.check(service)) {
			return;
		}
		
		IServiceSubscriber subscriber=
			(IServiceSubscriber)subscriptions.get(query);
		
		if (subscriber!=null) {
			notify(subscriber, query, service);					
		}
	}
	
	/**
	 * Unsubscribe from a service.
	 * 
	 * @param query AServiceQuery object.
	 */
	public void unsubscribe(AServiceQuery query) {
		
		subscriptions.remove(query);						
	}

	/**
	 * Checks whether a query has been subscribed.
	 * 
	 * @param query AServiceQuery object.
	 * @return <code>true</code> if subscribed, otherwise <code>false</code>.
	 */
	public boolean isSubscribed(AServiceQuery query) {
		
		return subscriptions.containsKey(query);
	}
	
	/**
	 * Check whether a service has changed and eventually notify all 
	 * subscribers.
	 * 
	 * @param info ServiceInfo instance.
	 */
	protected void checkServiceForSubscriptions(ServiceInfo info) {
		
		Iterator queryIter=new Vector(subscriptions.keySet()).iterator();

		// check all queries
		while (queryIter.hasNext()) {
			AServiceQuery query=(AServiceQuery)queryIter.next();

			runQuery(query,info);
		}			
	}

	/**
	 * Checks whether local services match subscriptions.
	 *
	 */
	protected void checkLocalServicesForSubscriptions() {

		HashSet serviceSet=new HashSet(localServices.values());
		
		Iterator serviceIter=serviceSet.iterator();
		
		while (serviceIter.hasNext()) {
			checkServiceForSubscriptions((ServiceInfo)serviceIter.next());
		}
	}
	
	/**
	 * Notify the subscriber in a background thread.
	 * 
	 * @param sub Subscriber.
	 * @param query Query.
	 * @param info Service.
	 */
	private void notify(final IServiceSubscriber sub, 
			final AServiceQuery query, final ServiceInfo info) {
		
		// create a thread and run it
		new Thread() {
			public void run() {
				sub.notification(query, info);
			}
		}.start();
	}
	
}
