/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.discovery;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;


import at.ac.tuwien.vitalab.middleware.MiddlewareConfiguration;
import at.ac.tuwien.vitalab.middleware.AWebService;
import at.ac.tuwien.vitalab.middleware.soap.ServiceContainer;
import at.ac.tuwien.vitalab.middleware.soap.services.InfoService;
import at.ac.tuwien.vitalab.middleware.util.Base64;

/**
 * Service Publisher/Locator.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class ServiceRegistry {

	private static Logger logger=LoggerFactory.getLogger(ServiceRegistry.class);
	
	private int interval;
	
	private Timer timer;
	private static RegistryTask registryTask=null;
	private static AvailabilityUpdateTask updateTask=null;
	
	private static ServiceRegistry localRegistry=null;
		
	private InfoService infoService;
	
	private static String ip=MiddlewareConfiguration.getIP();	
	private static int avdPort=MiddlewareConfiguration.getAdvertisementPort();
	private static int cmdPort=MiddlewareConfiguration.getSoapPort()-1;
	private InetAddress mcastAddr=null;
	
	private ServiceDB serviceDB=null;
	
	private UDPReceiverThread advReceiverThread=null;
	private UDPReceiverThread cmdReceiverThread=null;
	
	private Boolean doAdvertisment=Boolean.FALSE;
	private long lastAdvSentTS=-1;
	private long lastAdvRcvdTS=-1;
	private int lastAdvSerial=-1;
	
	private static final String CMD_ADVERTISE="advertise";
	private static final String CMD_GETATTRIBUTES="getAttrib";
	private static final String CMD_GETSTATUS="getStatus";
	private static final String CMD_IFUUID="ifUUID";
	
	private static final int MAX_UDP_SIZE=8*1024;	// 8KB should be fine
	
	/**
	 * Constructor.
	 * 
	 * @param infoService InfoService which shall be published.
	 * @throws IOException If the sockets cannot be instantiated; 
	 */
	protected ServiceRegistry(InfoService infoService) throws IOException {
		super();		
		
		mcastAddr = InetAddress.getByName("239.255.255.253");
		
		InetSocketAddress advAddr=new InetSocketAddress(ip,avdPort); 
		
		MulticastSocket advRecvSocket=null;
		
		advRecvSocket=new MulticastSocket(advAddr);
		advRecvSocket.joinGroup(mcastAddr);
		advRecvSocket.setTimeToLive(
				MiddlewareConfiguration.getAdvertisementTTL());
		
		advReceiverThread=new UDPReceiverThread(advRecvSocket);

		InetSocketAddress cmdAddr=new InetSocketAddress(ip,cmdPort); 
		
		MulticastSocket cmdRecvSocket=null;
		
		cmdRecvSocket=new MulticastSocket(cmdAddr);
		cmdRecvSocket.joinGroup(mcastAddr);
		cmdRecvSocket.setTimeToLive(
				MiddlewareConfiguration.getAdvertisementTTL());
		
		cmdReceiverThread=new UDPReceiverThread(cmdRecvSocket);
		
		serviceDB = new ServiceDB();
		
		interval=MiddlewareConfiguration.getAdvertisementInterval();
		
		this.infoService=infoService;
		
		advReceiverThread.start();
		cmdReceiverThread.start();
	}

	/**
	 * Publishes a Web service.
	 * 
	 * @param ws Web service.
	 */
	public void publish(AWebService ws) {

		if (!ws.isPublished()) {
			return;
		}
		
		// put a new advertisement string into the queue
		synchronized (doAdvertisment) {
			serviceDB.addLocalService(ws);
			
			infoService.registerService(ws);
						
			doAdvertisment=Boolean.TRUE;
		}
	}
	
	/**
	 * Unpublishes a Web services.
	 * 
	 * @param ws Web service.
	 */
	public void unpublish(AWebService ws) {
		
		// put a new advertisement string into the queue
		synchronized (doAdvertisment) {
			serviceDB.deleteLocalService(ws);
			
			if (!infoService.unregisterService(ws)) {
				return;
			}
			
			doAdvertisment=Boolean.TRUE;
		}
	}
	
	/**
	 * Serializes the status changes into a string.
	 * 
	 * @param fromSerial Last serial number retrieved by the client.
	 * @param toSerial Current serial number received by the client.
	 * @return String representation.
	 */
	private String serializeStatusChanges(int fromSerial, int toSerial) {

		String result="";
		
		Vector plusAndMinus=infoService.getStatusChanges(fromSerial, toSerial);
		Vector plus=(Vector)plusAndMinus.get(0);
		Vector minus=(Vector)plusAndMinus.get(1);
		
		if (!plus.isEmpty()) {
			result+="+ ";
			Iterator plusIter=plus.iterator();
			while (plusIter.hasNext()) {
				result+=plusIter.next()+" ";
			}			
		}

		if (!minus.isEmpty()) {
			result+="- ";
			Iterator minusIter=minus.iterator();
			while (minusIter.hasNext()) {
				result+=minusIter.next()+" ";
			}
		}
		
		return result;
	}
	
	/**
	 * Retrieves status changes of a host via UDP.
	 * 
	 * @param service Infoservice of host which shall be refreshed.
	 * @param fromSerial Last serial number retrieved by the client.
	 * @param toSerial Current serial number received by the client.
	 * @param added Collection where added hosts shall be put.
	 * @param removed Collection where removed hosts shall be put.
	 * @throws IOException If communication errors occured.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If the url attribute has not
	 * been published 
	 */
	protected static void retrieveChangesViaUDP(
			ServiceInfo service, String fromSerial, String toSerial,
			Collection added, Collection removed) throws IOException, 
			InactiveServiceException, MissingServiceAttributeException {
		
		
		String msg=CMD_GETSTATUS+" "+fromSerial+" "+toSerial;
		
		String data=sendAndReceive(service,msg);
		
		// correct response?
		if (data.startsWith("OK ")) {
			
			service.refresh();
			
			if (data.length()<=3 ) {
				return;
			}
			
			//deserialize
			
			StringTokenizer tokens=new StringTokenizer(data);
			tokens.nextToken(); // get rid of status "OK"
			
			Collection current=new HashSet();
		
			while (tokens.hasMoreTokens()) {
				String s=tokens.nextToken();
				
				if ("+".equals(s)) {
					current=added;
				} else if ("-".equals(s)) {
					current=removed;
				} else {
					current.add(s);
				}
			}
		} else {
			throw new IOException(data);
		}
	}
	
	/**
	 * Retrieve service attributes via UDP.
	 * 
	 * @param service Service..
	 * @param namespace Namespace of service.
	 * @return Attribute hashtable.
	 * @throws IOException If errors occur.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If the url attribute has not
	 * been published. 
	 */
	protected static Hashtable retrieveAttributesViaUDP(
			ServiceInfo service, String namespace) throws IOException, 
			InactiveServiceException, MissingServiceAttributeException {
	
		String msg=CMD_GETATTRIBUTES+" "+namespace;
		
		String data=sendAndReceive(service,msg);
		
		// correct response?
		if (data.startsWith("OK ")) {
			service.refresh();
			return deserializeAttributes(data.substring(3));			
		} else {
			throw new IOException(data);
		}
		
	}
	
	/**
	 * Send and receive via UDP.
	 * 
	 * @param host Host.
	 * @param msg Request message.
	 * @return Response message.
	 * @throws IOException If communication errors occur.
	 * @throws MissingServiceAttributeException If the url attribute has not
	 * been published. 
	 */
	private static String sendAndReceive(ServiceInfo service, String msg) 
	throws IOException, MissingServiceAttributeException {

		msg=CMD_IFUUID+" "+service.getHostId()+" "+msg;
		
		DatagramSocket socket=new DatagramSocket();
		DatagramPacket packetOut=new DatagramPacket(
				msg.getBytes(),msg.length(),
				InetAddress.getByName(
						service.getHost()),service.getUDPCommandPort());
		
		socket.setSoTimeout(
				MiddlewareConfiguration.getAdvertisementUDPTimeout()*1000);
		
		// send
		socket.send(packetOut);
		
		DatagramPacket packetIn;
		byte[] bytes = new byte[MAX_UDP_SIZE];
		packetIn = new DatagramPacket(bytes, bytes.length);
		
		// and receive
		socket.receive(packetIn);
		
		return new String(packetIn.getData(),0,packetIn.getLength());
	}
	
	/**
	 * Deserializes an Hashtable from a String. 
	 * 
	 * @param attStr String.
	 * @return Hashtable.
	 */
	private static Hashtable deserializeAttributes(String attStr) {
		StringTokenizer tokens=new StringTokenizer(attStr);
		
		Hashtable result=new Hashtable();
		
		while (tokens.hasMoreTokens()) {
			String att=tokens.nextToken();
			
			int posOfHash=att.indexOf('#');
			int posOfEquals=att.indexOf('=');
			
			String key=null;
			String value=null;
			
			if (posOfHash!=-1 && (posOfEquals==-1 || posOfEquals>posOfHash)) {
				key=att.substring(0, posOfHash);
				value=new String(Base64.decode(att.substring(posOfHash+1)));
				
			} else if (posOfEquals!=-1 && 
					(posOfHash==-1 || posOfHash>posOfEquals)) {
				key=att.substring(0, posOfEquals);
				
				value=att.substring(posOfEquals+1);
			} else {
				throw new IllegalArgumentException(
						"Invalid attribute '"+att+"'");
			}
			
			result.put(key, value);
		}
		
		return result;
	}
	
	/**
	 * Serializes a Hashtable to a String.
	 * 
	 * @param atts Attributes.
	 * @return String.
	 */
	private static String serializeAttributes(Hashtable atts) {
		Enumeration keyEnum=atts.keys();
		
		String result="";
		
		while (keyEnum.hasMoreElements()) {
			String key=keyEnum.nextElement().toString();
			String value=atts.get(key).toString();
			if (value.indexOf(' ')!=-1 ||
					value.indexOf('\t')!=-1 ||
					value.indexOf('\n')!=-1 ||
					value.indexOf('\r')!=-1 ||
					value.indexOf('\f')!=-1) {
				result+=key+"#"+Base64.encodeBytes(
						value.getBytes(), Base64.DONT_BREAK_LINES)+" ";
			} else {
				result+=key+"="+value+" ";
			}
		}
		
		return result;
	}
	
	private class AvailabilityUpdateTask extends TimerTask {

		/* (non-Javadoc)
		 * @see java.util.TimerTask#run()
		 */
		public void run() {
			
			Collection services=serviceDB.getAllServices();
			Iterator serviceIter=services.iterator();

			while (serviceIter.hasNext()) {
				
				ServiceInfo service=(ServiceInfo)serviceIter.next();

				service.checkAvailabilityUpdates();
			}
		}
	}
	
	private class RegistryTask extends TimerTask {

		/* (non-Javadoc)
		 * @see java.util.TimerTask#run()
		 */
		public void run() {
			
			String msg=null;
			
			synchronized (doAdvertisment) {
				if (doAdvertisment.booleanValue()) {
					// fetch the last advertisment from the queue
					msg=buildAdvertisementString(true);
					doAdvertisment=Boolean.FALSE;				
				} else {

					// or 
					if (interval<=Math.round(((double)
							(System.currentTimeMillis()-lastAdvSentTS))/1000) ||
						interval<=Math.round(((double)
							(System.currentTimeMillis()-lastAdvRcvdTS))/1000)) {

						msg=buildAdvertisementString(false);
					}
				}
			}
			
			if (msg!=null) {				
				msg=CMD_ADVERTISE+" "+msg;
				DatagramPacket packet=new DatagramPacket(
						msg.getBytes(), msg.length(), mcastAddr, avdPort);
				
				// periodically we should also check the subscriptions which
				// match local services
				serviceDB.checkLocalServicesForSubscriptions();
				
				DatagramSocket socket=null;
				try {
					socket=new DatagramSocket();
					socket.send(packet);
					lastAdvSentTS=System.currentTimeMillis();
				} catch (IOException e) {
					logger.debug("Cannot send advertisment: "+
							e.getMessage());
				} finally {
					if (socket!=null && !socket.isClosed()) {
						socket.close();						
					}
				}
			}			
		}
	}
	
	/**
	 * Builds the published String.
	 * 
	 * @param appendLastChange Whether the last change in the state should
	 * be propagated with the advertisment message.
	 * 
	 * @return String.
	 */
	private String buildAdvertisementString(boolean appendLastChange) {
		
		ServiceContainer container=ServiceContainer.getLocalServiceContainer();

		if (container==null) {
			throw new IllegalStateException(
					"Service container not initialized.");
		}
		
		String hostID=MiddlewareConfiguration.getHostUUID();
		
		int serialNr=infoService.getSerial();
		
		String msg=hostID+" "+container.getServiceURL()+
		" "+infoService.getInstanceID()+" "+serialNr;
		
		if (appendLastChange) {
			if (lastAdvSerial==serialNr-1) {
				msg+=" "+infoService.getLastStatusChange();				
			}
		}
		
		lastAdvSerial=serialNr;
		
		return msg;
	}
	
	/**
	 * Starts and returns the local service registry.
	 * 
	 * @param infoService InfoService which will be published.
	 * @return Local service registry instance.
	 * @throws IOException If the sockets cannot be instantiated;
	 */
	public synchronized static ServiceRegistry startLocalServiceRegistry(
			InfoService infoService) throws IOException { 
		
		if (localRegistry==null) {
			localRegistry=new ServiceRegistry(infoService);			
			localRegistry.timer=new Timer();
			
			registryTask=localRegistry.new RegistryTask();
			localRegistry.timer.scheduleAtFixedRate(registryTask, 3000, 1000);
			
			updateTask=localRegistry.new AvailabilityUpdateTask();
			localRegistry.timer.scheduleAtFixedRate(updateTask, 0, 1000);
		}	

		return localRegistry;
	}
	
	/**
	 * Closes the local service registry.
	 */
	public synchronized static void closeLocalServiceRegistry() {

		if (localRegistry!=null) {			
			localRegistry.timer.cancel();
			registryTask.run();
			try {
				// allow other peers to update status
				Thread.sleep(3000);
			} catch (InterruptedException e) {
				// ignore
			}
			localRegistry.advReceiverThread.shutdown();
			localRegistry.cmdReceiverThread.shutdown();
//			localRegistry.advSocket.close();
//			localRegistry.cmdSocket.close();
			localRegistry=null;
		}
	}
	
	/**
	 * Returns the local service registry.
	 * 
	 * @return The local service registry or <code>null</code> if none 
	 * is currently running.
	 */
	public synchronized static ServiceRegistry getLocalServiceRegistry() {

		return localRegistry;
	}

	/**
	 * Returns the service database.
	 * 
	 * @return Service database.
	 */
	public ServiceDB getServiceDataBase() {
		return serviceDB;
	}
	
	/**
	 * Thread which is responsible for receiving udp requests.
	 * 
	 * @author Lukasz Juszczyk
	 *
	 */
	private class UDPReceiverThread extends Thread {
		
		DatagramSocket socket=null;
		
		public UDPReceiverThread(DatagramSocket socket) {
			super();
			this.socket = socket;
		}

		private boolean running=true;
		
		/*
		 * (non-Javadoc)
		 * @see java.lang.Thread#run()
		 */
		public void run() {
			while (running) {
				DatagramPacket packet;
				byte[] bytes = new byte[MAX_UDP_SIZE+128];
				packet = new DatagramPacket(bytes, bytes.length);
				String request=null;
				try {
					socket.receive(packet);
					request=new String(packet.getData(),0,packet.getLength());
					
					StringTokenizer tokens=new StringTokenizer(request);
					processRequest(tokens,packet);
					
				} catch (IOException e) {
					if (running) {
						logger.debug("Error while receiving UDP data: "+
								e.getMessage());						
					}
				} catch (Exception nsee) {
					if (running) {
						logger.debug("Received invalid UDP data '"+
								request+"'");						
					}
				}
			}
		}
		
		/**
		 * Process a request.
		 * 
		 * @param requestTokens Request tokens.
		 * @param requestPacket
		 * @throws IOException
		 */
		private void processRequest(StringTokenizer requestTokens, 
				DatagramPacket requestPacket) throws IOException {
			
			String response=null;
			String command=requestTokens.nextToken();
			
			if (CMD_IFUUID.equals(command)) {
				
				String uuid=requestTokens.nextToken();
				if (uuid.equals(MiddlewareConfiguration.getHostUUID())) {
					processRequest(requestTokens,requestPacket);
				} else {
					response="Error: Mismatching host UUID.";
				}
				
			} else if (CMD_ADVERTISE.equals(command)) {
				
				final String uuid=requestTokens.nextToken();
				final String url=requestTokens.nextToken();
				final String instanceID=requestTokens.nextToken();
				final String serial=requestTokens.nextToken();

				String tmpUpdateStr=null;
				
				try {
					tmpUpdateStr=requestTokens.nextToken();
				} catch (NoSuchElementException e) {
				}

				final String updateStr=tmpUpdateStr;
				
				if (!MiddlewareConfiguration.getHostUUID().equals(uuid)) {
					
					lastAdvRcvdTS=System.currentTimeMillis();
					
					// do only process advertisments from other nodes
					new Thread() {
						public void run () {
							serviceDB.updateFromHost(
									instanceID, serial, uuid, url, updateStr);															
						}
					}.start();
				}
				
			} else if (CMD_GETATTRIBUTES.equals(command)) {
				
				String namespace=requestTokens.nextToken();
										
				try {
					Hashtable atts=infoService.getServiceAttributes(namespace);

					response="OK "+serializeAttributes(atts);

				} catch (Exception e) {
					response="Error: "+e.getMessage();
				}
				
			} else if (CMD_GETSTATUS.equals(command)) {
				
				try {
					
					int fromSerial=0;
					int toSerial=0;

					if (requestTokens.hasMoreTokens()) {
						fromSerial=Integer.parseInt(requestTokens.nextToken());
						toSerial=Integer.parseInt(requestTokens.nextToken());
					} 
					
					response="OK "+serializeStatusChanges(fromSerial,toSerial);
					
				} catch (Exception e) {
					e.printStackTrace();
					response="Error: "+e.getMessage();
				}
			}
			
			// send a response if this is desired
			if (response!=null) {
				
				if (response.length()>MAX_UDP_SIZE) {
					response="Error: Response too large for UDP.";
				}
				
				DatagramSocket socket=new DatagramSocket();
				DatagramPacket responsePacket=new DatagramPacket(
						response.getBytes(),response.length(),
						requestPacket.getAddress(),requestPacket.getPort());
				socket.send(responsePacket);
			}

		}
		
		/**
		 * Closes the thread.
		 *
		 */
		public void shutdown() {
			running=false;
			interrupt();
			socket.close();
		}
	}	
}
