/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap;

import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.List;
import java.util.Vector;

import org.ksoap2.serialization.SoapObject;
import org.ksoap2.serialization.SoapSerializationEnvelope;
import org.kxml2.io.KXmlParser;
import org.kxml2.io.KXmlSerializer;
import org.kxml2.kdom.Element;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlSerializer;

import EDU.oswego.cs.dl.util.concurrent.Semaphore;

import edu.wustl.mobilab.sliver.soap.Channel;
import edu.wustl.mobilab.sliver.soap.Fault;
import edu.wustl.mobilab.sliver.soap.MustUnderstandException;
import edu.wustl.mobilab.sliver.soap.Serialization;

/**
 * Local service invocation channel.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class LocalChannel extends Channel {

	private List data=new Vector();
	
	// semaphores for synchronization
	private Semaphore requestSem=new Semaphore(0);
	private Semaphore responseSem=new Semaphore(0);
	
	// is client of server?
	private boolean isClient=true;
	
	// pointer to other side of the channel
	private LocalChannel otherChannel=null;
	
	/**
	 * Constructor.
	 * 
	 */
	public LocalChannel() {
		super();
	}	
	
	/**
	 * Creates the service side channel.
	 * 
	 * @return Service side channel.
	 */
	protected LocalChannel createServiceSide() {
		
		LocalChannel ch=new LocalChannel();

		ch.requestSem=this.responseSem;
		ch.responseSem=this.requestSem;
		ch.isClient=false;
		ch.data=this.data;
		
		otherChannel=ch;
		ch.otherChannel=this;
		
		return ch;
	}
	
	/* (non-Javadoc)
	 * @see edu.wustl.mobilab.sliver.soap.Channel#closeImpl()
	 */
	protected void closeImpl() throws IOException {
	}

	/* (non-Javadoc)
	 * @see edu.wustl.mobilab.sliver.soap.Channel#receiveObject()
	 */
	public synchronized Object receiveObject() throws IOException, XmlPullParserException,
			MustUnderstandException {

		// wait until response is here
		try {
			responseSem.acquire();
		} catch (InterruptedException e) {
		}
		
		Object response= data.get(data.size()-1);
		
		SoapResponse.checkForFault(response);
		
		return response;
		
	}

	/* (non-Javadoc)
	 * @see edu.wustl.mobilab.sliver.soap.Channel#sendObject(java.lang.Object)
	 */
	public synchronized void sendObject(Object object) throws IOException {
		
		SoapObject request=null;
		
		try {
			request=(SoapObject)object;			
		} catch (ClassCastException cce) {
			throw new IOException("Only SoapObjects allowed.");
		}

		SoapObject requestCopy=null;

		if (object instanceof Fault) {
			requestCopy=request;
		} else {
			try {
				requestCopy=cloneObject(request);
			} catch (Exception e) {
				throw new IOException(e.getMessage());
			}			
		}
		
		
		if (isClient) {
			createServiceSide();
		}

		// copy outgoing headers to incoming of remote channel
		if (otherChannel!=null)
			otherChannel.incomingXMLHeaders=(Element[])outgoingHeaderElements.
			toXMLElements().toArray(new Element[]{});

		
		data.clear();

		if (isClient) {
			
			data.add(requestCopy);

			// call in background thread
			new Thread() {
				
				public void run() {
					requestSem.release();
					ServiceContainer.getLocalServiceContainer().
					getServer().newConnection(otherChannel);
				}
			}.start();
			
		} else {
			data.add(SoapResponse.retrieveResultObject(requestCopy));
			
			requestSem.release();
		}
	}

	/**
	 * Clones an object by serializing and deserializing it.
	 * 
	 * @param o Object.
	 * @param pi PropertyInfo object.
	 * @return Cloned object.
	 * @throws IOException If IO errors occured.
	 * @throws XmlPullParserException If the serializers/deserializers
	 * threw XML errors.
	 */
	private synchronized SoapObject cloneObject(SoapObject o) 
	throws IOException, XmlPullParserException {

		SoapSerializationEnvelope env=Serialization.newEnvelope();
		
		// serialize
		StringWriter sw=new StringWriter();
		XmlSerializer serializer = new KXmlSerializer();
		serializer.setOutput(sw);

		env.bodyOut=o;
		env.headerOut=null;
		
		env.write(serializer);
		sw.flush();
		
		
		// deserialize
		StringReader sr=new StringReader(sw.toString());
		XmlPullParser parser = new KXmlParser();
		
		parser.setInput(sr);
		parser.setFeature(XmlPullParser.FEATURE_PROCESS_NAMESPACES, true);
		env.parse(parser);
		
		return (SoapObject)env.bodyIn;
	}
}
