/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;

/**
 * ClassLoader for Web Services. Reads in new Jars and make containing classes
 * available.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class WSClassLoader extends ClassLoader {
	
    // we don't want duplicate urls
    protected HashSet addedURLs=new HashSet();
    
    // real loader
    protected InternalClassLoader loader=null;
                 
    // backup
    protected ClassLoader oldLoader=null;

    
    /**
     * Protected Constructor.
     *
     */
    protected WSClassLoader() {
		super();
	}

	/**
     * Constructor.
     * 
     * @param oldLoader Old ClassLoader which will be invoked if WSClassLoader
     * cannot find a class. 
     *
     */
    public WSClassLoader(ClassLoader oldLoader) {
        super();
        this.oldLoader=oldLoader;
        this.loader=
        	new WSClassLoader().new InternalClassLoader(new HashSet());
    }    
    
    /**
     * Constructor.
     * 
     * @param urls Array of URLs.
     * @param oldLoader Old ClassLoader which will be invoked if WSClassLoader
     * cannot find a class. 
     */
    public WSClassLoader(ClassLoader oldLoader, URL[] urls) {
        this(oldLoader);
        for (int i=0;i<urls.length;i++) {
            loader.addURL(urls[i]);
        }
    }

    /**
     * Adds a URL.
     * 
     * @param url URL.
     */
    public void addURL(URL url) {
        loader.addURL(url);        
    }

    /**
     * Deletes a URL.
     * 
     * @param url URL.
     * @return <code>true</code> if URL existed, otherwise <code>false</code>.
     */
    public boolean delURL(URL url) {
        if (addedURLs.contains(url)) {
            addedURLs.remove(url);
            loader=
            	new WSClassLoader(oldLoader).new InternalClassLoader(addedURLs);
            return true;
        }        
        return false;
    }
    
	/**
	 * @param arg0
	 * @return
	 * @throws ClassNotFoundException
	 * @see java.lang.ClassLoader#loadClass(java.lang.String)
	 */
	public Class loadClass(String arg0) throws ClassNotFoundException {
		try {
			return loader.loadClass(arg0);			
		} catch (ClassNotFoundException cnfe) {
			if (oldLoader!=null) {
				return oldLoader.loadClass(arg0);
			} else {
				throw cnfe;
			}
		}
	}

	/**
	 * @param arg0
	 * @return
	 * @see java.lang.ClassLoader#getResource(java.lang.String)
	 */
	public URL getResource(String arg0) {
		URL result=null;
		result=loader.getResource(arg0);
		if (result==null) {
			result=oldLoader.getResource(arg0);
		}
		return result;
	}
	
    /*
     * The real class loader.
     */
    protected class InternalClassLoader extends URLClassLoader {
        
        /**
         * Constructor.
         *
         */
        protected InternalClassLoader(HashSet urls) {
            super(new URL[]{});
            for (Iterator urlIter=urls.iterator(); urlIter.hasNext(); ) {
                addURL((URL)urlIter.next());
            }
        }    
        
        /* (non-Javadoc)
         * @see java.net.URLClassLoader#addURL(java.net.URL)
         */
        public void addURL(URL url) {
            if (!addedURLs.contains(url)) {
                super.addURL(url);   
                addedURLs.add(url);
            }
        }

        /**
         * Creates a class object from a byte array.
         * 
         * @param name Class name.
         * @param b Class bytes.
         * @return Class instance.
         */
        public Class classFromBytes(String name, byte[] b) {
        	
        	return super.defineClass(name, b, 0, b.length);
        }
    }

	/**
	 * 
	 * @see java.lang.ClassLoader#clearAssertionStatus()
	 */
	public void clearAssertionStatus() {
		loader.clearAssertionStatus();
	}

	/**
	 * @param arg0
	 * @return
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object arg0) {
		return loader.equals(arg0);
	}

	/**
	 * @param arg0
	 * @return
	 * @see java.net.URLClassLoader#findResource(java.lang.String)
	 */
	public URL findResource(String arg0) {
		return loader.findResource(arg0);
	}

	/**
	 * @param arg0
	 * @return
	 * @throws IOException
	 * @see java.net.URLClassLoader#findResources(java.lang.String)
	 */
	public Enumeration findResources(String arg0) throws IOException {
		return loader.findResources(arg0);
	}

	/**
	 * @param arg0
	 * @return
	 * @see java.lang.ClassLoader#getResourceAsStream(java.lang.String)
	 */
	public InputStream getResourceAsStream(String arg0) {
		return loader.getResourceAsStream(arg0);
	}

	/**
	 * @return
	 * @see java.net.URLClassLoader#getURLs()
	 */
	public URL[] getURLs() {
		return loader.getURLs();
	}

	/**
	 * @return
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return loader.hashCode();
	}

	/**
	 * @param arg0
	 * @param arg1
	 * @see java.lang.ClassLoader#setClassAssertionStatus(java.lang.String, boolean)
	 */
	public void setClassAssertionStatus(String arg0, boolean arg1) {
		loader.setClassAssertionStatus(arg0, arg1);
	}

	/**
	 * @param arg0
	 * @see java.lang.ClassLoader#setDefaultAssertionStatus(boolean)
	 */
	public void setDefaultAssertionStatus(boolean arg0) {
		loader.setDefaultAssertionStatus(arg0);
	}

	/**
	 * @param arg0
	 * @param arg1
	 * @see java.lang.ClassLoader#setPackageAssertionStatus(java.lang.String, boolean)
	 */
	public void setPackageAssertionStatus(String arg0, boolean arg1) {
		loader.setPackageAssertionStatus(arg0, arg1);
	}

	/**
	 * @return
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return loader.toString();
	}

	/**
	 * @param name
	 * @param b
	 * @return
	 * @see at.ac.tuwien.vitalab.middleware.soap.WSClassLoader.InternalClassLoader#classFromBytes(java.lang.String, byte[])
	 */
	public Class classFromBytes(String name, byte[] b) {
		return loader.classFromBytes(name, b);
	}
	
}
