/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap.header;

import java.util.Hashtable;
import java.util.NoSuchElementException;

import org.kxml2.kdom.Element;
import org.xmlpull.v1.XmlPullParserException;

/**
 * 
 * Abstract SOAP header element.
 * 
 * @author Lukasz Juszczyk
 *
 */
public abstract class AHeaderElement {

	protected String elementName="";
	protected String elementNamespace="";

	/**
	 * Constructor.
	 * 
	 * @param name Name of the element.
	 * @param namespace Namespace of the element.
	 */
	protected AHeaderElement(String name, String namespace) {
		elementName=name;
		elementNamespace=namespace;
	}
	
	/**
	 * Constructor, reading in from the SOAP headers.
	 * 
	 * @param headers XML header elements.
	 * @param name Name of the element.
	 * @param namespace Namespace of the element.
	 * @throws NoSuchElementException If the SOAP headers do not contain it.
	 * @throws XmlPullParserException If the header is malformed.
	 */
	protected AHeaderElement(Element[] headers, String name, String namespace) 
	throws NoSuchElementException, XmlPullParserException {
		this(name,namespace);
		this.load(headers);
	}

	/**
	 * Returns the key for the hashtable.
	 * 
	 * @return String key.
	 */
	protected String getKey() {
		return SoapHeaders.getKey(elementName,elementNamespace);
	}
	
	/**
	 * Reads in the data from the XML header element.
	 * 
	 * @param element XML element.
	 * @throws XmlPullParserException If it is malformed.
	 */
	public abstract void load(Element element) throws XmlPullParserException;
	
	/**
	 * Searches the SOAP headers for the corresponding XML header element
	 * and reads in the data from it.
	 * 
	 * @param elements XML elements.
	 * @throws NoSuchElementException If the element does not exist.
	 * @throws XmlPullParserException If it is malformed.
	 */
	public void load(Element[] elements) throws NoSuchElementException, 
	XmlPullParserException {
		if (elements!=null) {
			for (int i=0;i<elements.length;i++) {
				Element el=elements[i];
				if (elementName.equals(el.getName()) &&
						elementNamespace.equals(el.getNamespace())) {
					this.load(el);
					return;
				}
			}			
		}
		throw new NoSuchElementException("No such header element found.");
	}

	/**
	 * Returns the XML representation of this header element.
	 * 
	 * @return XML element.
	 */
	public abstract Element toXML();
	
	/**
	 * Returns the map of prefixes and namespaces which should be declared
	 * at the top header element.
	 * 
	 * @return Map.
	 */
	public Hashtable getNamespaceMap() {
		return new Hashtable();
	}
	
	/**
	 * Creates a new Element with the same name in the same namespace
	 * as the template.
	 * 
	 * @param src Source element.
	 * @return New element.
	 */
	protected Element createWithSameName(Element src) {
		Element dst=new Element();
		dst.setName(src.getName());
		dst.setNamespace(src.getNamespace());
		return dst;
	}
	
	/**
	 * Checks whether the elements have the same name and namespace.
	 * 
	 * @param a First element.
	 * @param b Second element.
	 * @return <code>true</code> if same, otherwise <code>false</code>.
	 */
	protected boolean hasSameName(Element a, Element b) {
		return (a.getName().equals(b.getName())) && 
		(a.getNamespace().equals(b.getNamespace()));
	}
}
