/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap.header;

import java.util.Hashtable;
import java.util.NoSuchElementException;

import org.kxml2.kdom.Element;
import org.xmlpull.v1.XmlPullParserException;

import at.ac.tuwien.vitalab.middleware.MiddlewareConfiguration;

/**
 * Asynchronous call header element.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class AsyncCallHeader extends AHeaderElement {

	protected String hostUUID=null;
	
	protected static Element asyncCallElement=new Element();
	protected static Element hostUUIDElement=new Element();
	
	static {
		asyncCallElement.setName("AsynchronousCall");
		asyncCallElement.setNamespace(SoapHeaders.MWA_NAMESPACE);
		
		hostUUIDElement.setName("ClientHostID");
		hostUUIDElement.setNamespace(SoapHeaders.MWA_NAMESPACE);
	}
	
	/**
	 * Constructor.
	 *
	 */
	public AsyncCallHeader() {
		super(asyncCallElement.getName(),asyncCallElement.getNamespace());
		
		hostUUID=MiddlewareConfiguration.getHostUUID();
	}

	/**
	 * Constructor, reading in the element from the SOAP headers.
	 * 
	 * @param headers XML header elements.
	 * @throws NoSuchElementException If the SOAP headers do not contain it.
	 * @throws XmlPullParserException If the header is malformed.
	 */
	public AsyncCallHeader(Element[] headers) 
	throws NoSuchElementException, XmlPullParserException {
		this();
		this.load(headers);
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.soap.header.AHeaderElement#load(org.kxml2.kdom.Element)
	 */
	public void load(Element el) throws XmlPullParserException {
		if (!this.hasSameName(asyncCallElement, el)) {
			throw new IllegalArgumentException(
					"Not a valid From element.");
		}

		hostUUID=null;
		
		for (int i=0;i<el.getChildCount();i++) {
			Object child=el.getChild(i);
			if (!(child instanceof Element)) {
				continue;
			}
			Element tmp=(Element)child;
			if (this.hasSameName(hostUUIDElement, tmp)) {
				hostUUID=tmp.getText(0);
				if (hostUUID==null) {
					throw new XmlPullParserException("Host ID missing");
				}
				if (hostUUID.startsWith("uuid:")) {
					hostUUID=hostUUID.substring(5);
				}
			}
		}
		
		if (hostUUID == null) {
			throw new XmlPullParserException("Child elements are missing");
		}
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.soap.header.AHeaderElement#toXML()
	 */
	public Element toXML() {

		Element asyncEl=this.createWithSameName(asyncCallElement);

		if (hostUUID!=null) {
			Element fromHostIDEl=this.createWithSameName(hostUUIDElement);
			if (hostUUID.startsWith("uuid:")) {
				fromHostIDEl.addChild(Element.TEXT,hostUUID);				
			} else {
				fromHostIDEl.addChild(Element.TEXT,"uuid:"+hostUUID);
			}
			asyncEl.addChild(Element.ELEMENT, fromHostIDEl);			
		}

//		fromEl.setAttribute("http://www.w3.org/2001/12/soap-envelope", 
//				"mustUnderstand", 
//				"1");
		
		return asyncEl;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.soap.header.AHeaderElement#getNamespaceMap()
	 */
	public Hashtable getNamespaceMap() {
		Hashtable result=new Hashtable();
		result.put("wma", SoapHeaders.MWA_NAMESPACE);
		return result;
	}

	/**
	 * Returns the Host ID.
	 * 
	 * @return Host ID.
	 */
	public String getHostID() {
		return hostUUID;
	}
}
