/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap.header;

import java.util.Hashtable;
import java.util.NoSuchElementException;

import org.kxml2.kdom.Element;
import org.xmlpull.v1.XmlPullParserException;

/**
 * RelatesTo header element.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class RelatesToHeader extends AHeaderElement {

	private String relatedMessageID=null;

	protected static Element relatedMessageIDElement=new Element();
	
	static {
		relatedMessageIDElement.setName("RelatesTo");
		relatedMessageIDElement.setNamespace(SoapHeaders.WSA_NAMESPACE);
	}
	
	/**
	 * Constructor.
	 *
	 */
	public RelatesToHeader() {
		super(relatedMessageIDElement.getName(),
				relatedMessageIDElement.getNamespace());
	}

	/**
	 * Constructor, reading in the element from the SOAP headers.
	 * 
	 * @param headers XML header elements.
	 * @throws NoSuchElementException If the SOAP headers do not contain it.
	 * @throws XmlPullParserException If the header is malformed.
	 */
	public RelatesToHeader(Element[] headers) 
	throws NoSuchElementException, XmlPullParserException {
		this();
		this.load(headers);
	}

	/**
	 * Constructor, relating to the passed MessageID element.
	 * 
	 * @param idElement MessageID element.
	 */
	public RelatesToHeader(MessageIDHeader idElement) {
		this();
		relateTo(idElement);
	}

	/**
	 * Relate to the passed MessageID element.
	 * 
	 * @param idElement MessageID element.
	 */
	public void relateTo(MessageIDHeader idElement) {
		relatedMessageID=idElement.getMessageID();
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.soap.AHeaderElement#getNamespaceMap()
	 */
	public Hashtable getNamespaceMap() {
		Hashtable result=new Hashtable();
		result.put("wsa", SoapHeaders.WSA_NAMESPACE);
		return result;
	}

	/*
	 * (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.soap.AHeaderElement#load(org.kxml2.kdom.Element)
	 */
	public void load(Element el) throws XmlPullParserException {
		if (!this.hasSameName(relatedMessageIDElement, el)) {
			throw new IllegalArgumentException(
					"Not a valid MessageID element.");
		}
		relatedMessageID=el.getText(0);
		if (relatedMessageID==null) {
			throw new XmlPullParserException("Message ID missing.");
		}
		if (relatedMessageID.startsWith("uuid:")) {
			relatedMessageID=relatedMessageID.substring(5);
		}
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.soap.AHeaderElement#toXML()
	 */
	public Element toXML() {
		
		Element messageIDEl=this.createWithSameName(relatedMessageIDElement);
		
		if (relatedMessageID.startsWith("uuid:")) {
			messageIDEl.addChild(Element.TEXT,relatedMessageID);				
		} else {
			messageIDEl.addChild(Element.TEXT,"uuid:"+relatedMessageID);
		}
		
		return messageIDEl;
	}
	
	/**
	 * Returns the related message ID.
	 * 
	 * @return Related message ID.
	 */
	public String getRelatedMessageID() {
		return relatedMessageID;
	}
	
}
