/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap.header;

import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.kxml2.kdom.Element;

/**
 * Soap Headers.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class SoapHeaders {

	public final static String WSA_NAMESPACE=
		"http://schemas.xmlsoap.org/ws/2004/08/addressing";
	
	public final static String MWA_NAMESPACE=
		"http://www.vitalab.tuwien.ac.at/RESCUE/addressing";
	
	// hashtable containing AHeaderElement instances
	private Hashtable headerElements=new Hashtable();
	
	/**
	 * Constructor.
	 *
	 */
	public SoapHeaders() {
		// mandatory headers
		this.addHeader(new MessageIDHeader());
	}

	/**
	 * Returns the map of prefixes and namespaces.
	 * 
	 * @return Map.
	 */
	public Hashtable getNamespaceMap() {
		Hashtable result=new Hashtable();
		
		Iterator headerIter=headerElements.values().iterator();
		
		while (headerIter.hasNext()) {
			
			AHeaderElement header=(AHeaderElement)headerIter.next();
			Hashtable nsTable=header.getNamespaceMap();
			for (Iterator prefixIter=nsTable.keySet().iterator();
			prefixIter.hasNext();) {
				String prefix=prefixIter.next().toString();
				String namespace=nsTable.get(prefix).toString();
				String tmpPrefix=prefix;
				int i=0;
				while (result.containsKey(tmpPrefix) && 
						!result.get(tmpPrefix).equals(namespace)) {
					tmpPrefix=prefix+(i++);
				}
				result.put(tmpPrefix, namespace);
			}
		}
		
		return result;
	}
	
	/**
	 * Adds a new header.
	 * 
	 * @param header Header. 
	 */
	public void addHeader(AHeaderElement header) {
		headerElements.put(header.getKey(), header);
	}
	
	/**
	 * Adds header elements. 
	 * 
	 * @param headers Header elements.
	 */
	public void addHeaders(SoapHeaders headers) {
		if (headers!=null) {
			addHeaders(headers.headerElements.values());			
		}
	}
	
	/** 
	 * Adds header elements. 
	 * 
	 * @param headers Header elements.
	 */
	public void addHeaders(Collection headers) {
		if (headers==null) {
			return;
		}
		Iterator headerIter=headers.iterator();
		while (headerIter.hasNext()) {
			AHeaderElement header=(AHeaderElement)headerIter.next();
			addHeader(header);
		}
	}
	
	/**
	 * Returns all headers.
	 * 
	 * @return List of headers.
	 */
	public Vector getHeaders() {
		return new Vector(headerElements.values());
	}
	
	/**
	 * Returns a header.
	 * 
	 * @param header Header from which name and namespace are taken.
	 * @return Header.
	 */
	public AHeaderElement getHeader(AHeaderElement header) {
		return (AHeaderElement)headerElements.get(header.getKey());
	}

	/**
	 * Returns a header.
	 * 
	 * @param name Name.
	 * @param namespace Namespace.
	 * @return Header.
	 */
	public AHeaderElement getHeader(String name, String namespace) {
		return (AHeaderElement)headerElements.get(getKey(name,namespace));
	}
	
	/**
	 * Deletes a header.
	 * 
	 * @param header Header from which name and namespace are taken.
	 * @return The previously existing header or <code>null</code> if nothing
	 * was found.
	 */
	public AHeaderElement deleteHeader(AHeaderElement header) {
		return (AHeaderElement)headerElements.remove(header.getKey());
	}

	/**
	 * Deletes a header.
	 * 
	 * @param name Name.
	 * @param namespace Namespace.
	 * @return The previously existing header or <code>null</code> if nothing
	 * was found.
	 */
	public AHeaderElement deleteHeader(String name, String namespace) {
		return (AHeaderElement)headerElements.remove(getKey(name,namespace));
	}
	
	/**
	 * Returns the key for the passed element.
	 * 
	 * @param e Element.
	 * @return Key String.
	 */
	protected static String getKey(Element e) {
		return getKey(e.getName(),e.getNamespace());
	}

	/**
	 * Returns the key for the passed name and namespace.
	 * 
	 * @param name Name.
	 * @param ns Namespace.
	 * @return Key String.
	 */
	protected static String getKey(String name, String ns) {
		return name+":"+(ns!=null?ns:"");
	}

	/**
	 * Returns the XML elements of all headers.
	 * 
	 * @return XML elements.
	 */
	public Collection toXMLElements() {
		
		Vector result=new Vector();
		
		for (Iterator headerIter=headerElements.values().iterator();
		headerIter.hasNext();) {
			result.add(((AHeaderElement)headerIter.next()).toXML());
		}
		
		return result;
	}
}