/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.util;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Properties;
import java.util.Set;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;



/**
 * Utility for handling configuration files.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class Configuration {

	protected String confFile=null;
	
	protected static Logger logger=LoggerFactory.getLogger(Configuration.class);
	
	// key = path to config file
	// value = configuration properties
	protected static Hashtable configTable=new Hashtable();
	
	// key = hash out of config file name, config entry name, and type name
	// value = value of config entrys
	protected static HashMap configEntries=new HashMap();
	
	// we don't want to tell more than once that a file was not found
	protected static Set notFoundFileNames=new HashSet();
	
	/**
	 * Constructor.
	 * 
	 * @param confFile Configuration source file.
	 */
	public Configuration(String confFile) {
		super();
		this.confFile = confFile;
	}
	
	/**
	 * Returns the name of the configuration file 
	 * 
	 * @return File name.
	 */
	public String getConfigurationFileName() {
		return confFile;
	}

	/**
	 * Gets a configuration value from the properties.
	 * 
	 * @param name Name of the value.
	 * @param type Expected Java type.
	 * @param def Default value.
	 * @return Value from the configuration or default one if none exists.
	 */
	public Object getValue(String name, Class type, Object def) {

		Object value=this.getValue(name, type);
		
		Integer hash=new Integer((confFile+"&"+name+"&"+
				(type==null?"":type.getClass().getName())).hashCode());
		
		
		if (value!=null) {
			return value;
		}

		logger.debug("Assuming default configuration value for '"+
				name+"': "+def);
		
		configEntries.put(hash, def);
		return def;
	}

	
	/**
	 * Gets a configuration value from the properties.
	 * 
	 * @param name Name of the value.
	 * @param type Expected Java type.
	 * @return Value from the configuration or <code>null</code> if none exists.
	 */
	public Object getValue(String name, Class type) {

		Object value=null;
		
		Integer hash=new Integer((confFile+"&"+name+"&"+
				(type==null?"":type.getClass().getName())).hashCode());
		
		value=configEntries.get(hash);
		
		if (value!=null) {
			return value;
		}

		Properties config=null;
		
		if ((config=(Properties)configTable.get(confFile))==null){
			try {
				config=new Properties();
				config.load(new FileInputStream(confFile));
				configTable.put(confFile, config);
				logger.info("Reading config file '"+confFile+"'.");
			} catch (FileNotFoundException e) {
				if (!notFoundFileNames.contains(confFile)) {
					logger.warning("Config file '"+confFile+"' not found.");
					notFoundFileNames.add(confFile);					
				}
			} catch (IOException e) {
				logger.error("Cannot read config file '"+confFile+
						"': "+e.getMessage());
			}			
		}

		value=config.get(name);

		if (value!=null) {
			try {
				if (type==null || !type.equals(String.class)) {
					// try the constructor of the type with a string as argument
					Constructor constructor=
						type.getConstructor(new Class[]{String.class});
					value=constructor.newInstance(
							new Object[]{value.toString()});					
				}
				logger.debug("Configuration value of '"+name+"' is "+value);
				configEntries.put(hash, value);
				return value;
			} catch (Exception e){
				if (!configEntries.containsKey(hash)) {
					logger.warning("Configuration entry '"+name+"' has an " +
							"invalid "+"value '"+value+"'.");
				}
				value=null;
			}
		}

		configEntries.put(hash, value);
		return value;
	}
	
	/**
	 * Saves a corrected configuration value into the cache.
	 * 
	 * @param name Name of the value.
	 * @param type Expected Java type.
	 * @param value New value.
	 */
	public void saveCorrectedValue(String name, 
			Class type, Object value) {
		
		Integer hash=new Integer((confFile+"&"+name+"&"+
				(type==null?"":type.getClass().getName())).hashCode());
		
		configEntries.put(hash, value);
	}
}
