/*
    This file is part of RESCUE.

    RESCUE is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RESCUE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with RESCUE. If not, see <http://www.gnu.org/licenses/>.
 */
package edu.wustl.mobilab.sliver.log;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.text.DateFormat;
import java.util.Date;
import java.util.Properties;

/**
 * A base class for simple {@link Logger}s that writes messages to a stream
 * (e.g., stderr or a file).
 * 
 * @author Greg Hackmann
 * @author Lukasz Juszczyk
 */
public class Logger
{
	
	/**
	 * The logger's name.
	 */
	private final String name;
	
	private static boolean initialized=false;
	
	private static boolean debugAllowed=false;
	private static PrintStream printer=null;
	
	private static void init() {
		
		if (initialized) {
			return;
		}
		
		try {
			FileInputStream fis=new FileInputStream("conf/log.conf");
			
			Properties props=new Properties();
			
			props.load(fis);

			fis.close();

			if (props.get("debug").toString().equalsIgnoreCase("on")) {
				debugAllowed=true;
			}
			
			String dest=props.getProperty("destination","-");
			
			if (dest.equals("-") || dest.equalsIgnoreCase("stdout")) {
				printer=new PrintStream(System.out,true);
				printer.println("Starting logger.");
			} else {
				try {
					printer=new PrintStream(
							new FileOutputStream(dest,true),true);
					printer.println("Starting logger.");					
				} catch (IOException ioe) {
					printer=new PrintStream(System.out,true);
					printer.println("Logger falling back to stdout.");					
				}
			}
			
			printer.println("Debugging output "+
					(debugAllowed?"":"not ")+"allowed.");
			
		} catch (IOException e) {
			debugAllowed=true;
			printer=new PrintStream(System.out);
		} finally {
			initialized=true;
		}
	}
	
	/**
	 * Creates a new Logger.
	 * 
	 * @param name the logger's name
	 */
	protected Logger(String name)
	{
		init();
		this.name = name;
	}

	public void info(String message)
	{
		write("[info ]: " + message);
	}

	public void error(String message)
	{
		write("[error]: " + message);
	}

	public void warning(String message)
	{
		write("[warn ]: " + message);
	}
	
	public void debug(String message)
	{
		if (debugAllowed)
			write("[debug]: " + message);
	}
	
	/**
	 * Writes the specified logging message to the stream.
	 * 
	 * @param message the message to write
	 */
	protected synchronized void write(String message)
	{
		DateFormat time=DateFormat.getTimeInstance(DateFormat.LONG);
		printer.println(time.format(new Date())+" "+message);
	}

	public static boolean isDebugAllowed() {
		return debugAllowed;
	}	

}
